// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
   @author Juergen Zimmer University of Edinburgh
   @version $Date: 2006/08/24 13:40:31 $ $Revision: 1.16 $
*/
public class DCTP extends ATP {
  
  public String DCTP_HOME = getSystemProperty("dctp.home", "DCTP_HOME");
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String[] dctp_envs = new String[] {"DCTP_HOME="+DCTP_HOME};
  public String dctpExecutable = findSystemExecutable(DCTP_HOME, "dctp-10.21p", "Dctp-10.21p.exe");
  public String limitedRunExecutable = findSystemExecutable(TPTP_HOME+"/SystemExecution", "TreeLimitedRun", "TreeLimitedRun.exe");

  private static Logger logger = Logger.getLogger("DCTP");
  private static String logFileName = getSystemProperty("dctp.log.file", "DCTP_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
    
  private static String sep = File.separator;
    
  private static final int defaultNice = 15;
  
  public DCTP()
  {
    super();
    name = "DCTP_10_21p";
    acceptedConcept = "TstpProblem";
    resultConcept = "FOBrFPResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "dctp_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  /** 
   * @param problemFile a file containing a TSTP proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      outputFile = File.createTempFile("dctp", ".out");
      
      //      writeFile(origFile, problem);
      
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        
        /*  first transform the problem into TPTP CNF form without include statements */
        String toSyntax= "oldtptp";
        String transform = "add_equality";
        String newProb;

        inputFile = tstp2X.tstp2X(problemFile, toSyntax, transform);

        if (inputFile == null) 
          return new String[] {"Unknown", null, "The tptp2X utility could not generate a problem file for " + this.name + "."};
        //        logger.info("new problem: " + newProb);
        
        /*  write the new problem description in the input file */
        //        writeFile(inputFile,newProb);
        
        /*  create the shell command to call the prover and execute it */
        //        String eOptions = " --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out --cpu-limit=" + time + " ";
        String dctpOptions = options;
        //        String eOptions = " -s --print-statistics -xAuto -tAuto --tstp-in --tstp-out -l4 --memory-limit=100 --cpu-limit=" + time + " ";        

        String command = dctpExecutable + " " +  dctpOptions + " " +
          inputFile.getAbsolutePath() + " " + time + " > " + outputFile.getAbsolutePath() + " 2>&1";
                  
        logger.info("[DCTP] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.info("[DCTP] starting prover...");
        this.start(command, dctp_envs, defaultNice, time);
        
        /*  read the output written by the prover call */
        //        logger.info("[DCTP] reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
        String output = this.readFile(outputFile);
        
        //logger.info("[DCTP] output: " + output);
        //         logger.info("[DCTP] transforming to TSTP...");
        //         String tstp = x2tstp.x2tstp(output, "E---0.82");
        //         logger.info("[DCTP] TSTP..." + tstp);
        //         return tstp;
        
        /*  split the output in lines and examine them to extract the status */
        String[] split = output.split("\n");
        StringBuffer buf = new StringBuffer();
        String status = "Unknown";
        
        for(int i=0; i<split.length; i++){
          logger.info("[line] " + split[i]);
          if (split[i].length() > 10){
            if (split[i].indexOf("PROOF found") > -1){
              //              System.out.println("found status ");
              status = "Unsatisfiable";
              i = split.length;
            } else 
              if (split[i].indexOf("MODEL found") > -1){
                status = "Satisfiable";
                i = split.length;
              } else 
                if (split[i].indexOf("has status FAIL") > -1){
                  status = "GaveUp";
                  i = split.length;
                }
            
          }
        }
        
        /*  create and return the resulting String array */
        //        System.out.println("output " + resArray);
        return new String[] {status, null, output};
      } else{
        throw new Exception("Input language "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[DCTP] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[DCTP] Exception " + e); }
    }
    
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    DCTP dctp = new DCTP();
    OWLJDom owlJDom = OWLJDom.getInstance();
    TSTPCreator creator = new TSTPCreator();
    String timeInd;
    //timeInd = owlJDom.createTimeIndi(600,600);
    timeInd = owlJDom.createTimeIndi(20,20);
    //      timeInd = dctp.readFile(msHome + "/examples/owl/fault-time.owl");
    String servF1;
    System.out.println("path: " +System.getProperty("java.library.path"));

    String ind;
    //    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/PUZ/PUZ001+1.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GEO/GEO079-1.p");
    //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/TOP/TOP020+1.p");
    //ind = creator.tstp2OWLProblem("/home/jzimmer/casc/Active/Problems/EPS/NoTransform/tptp/SYN/SYN816-1+short+ran.p");  
    //        ind = creator.tstp2OWLProblem("/home/jzimmer/casc/Active/Problems/EPS/NoTransform/tptp/SYN/SYN831-1+short+ran.p");
    //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG001-2.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRA/GRA001-1.p");
    //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP043+1.p");
            ///       ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SET/SET062+3.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
    ind  = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN455+1.p");

     
           System.out.println(ind);
      
    try {
      System.out.println(timeInd);
      String result = null;
      //         for (int i=0; i< 700; i++) 
      //         {
      //             System.out.println(" call " + i);            
      //             e = new DCTP();
      result = dctp.proveProblem(ind, timeInd);
      //dctp.writeFile(msHome + "/examples/owl/dctp-proof2.owl", result);
      //result = e.proveProblemXml(ind, timeInd);
      //System.out.println(result);
        
//       String tptpProb = readFile("/home/jzimmer/test.p");
//       //tptpProb = readFile("/home/jzimmer/out");
//       result = dctp.proveTPTPProblem(tptpProb, new Integer(60));
        
      //        logger.info("result1: " + result);
        
      //         result = dctp.proveProblem(ind);
        
      //        logger.info("result2: " + result);
      System.out.println(result);
    }
    catch(Exception ex){
      ex.printStackTrace();
      System.out.println(ex);
    }
  }
  
  
} // end DCTP.java


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
