// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:

package org.mathweb.broker;

import java.io.*;
import java.util.*;

import salvo.jesus.graph.*;
import salvo.jesus.graph.visual.*;
import salvo.jesus.graph.visual.layout.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.*;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.parser.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.basicplanner.*;
import org.mathweb.reasoning.partialorderplanner.*;
//import org.mathweb.reasoning.regressionplanner.*;

import org.mathweb.graph.*;
import org.mathweb.broker.ServiceEntry;

import org.jdom.*;

// import java.awt.*;
// import java.awt.event.*;
// import java.applet.*;
// import javax.swing.*;
// import javax.swing.tree.*;

/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2006/08/24 13:41:18 $ $Revision: 1.33 $
*/

class ServicePlanner extends BasicService 
{
  private PartialOrderPlanner planner;
  private PartialOrderPlan plan;

  private Hashtable serviceHash;
  private Hashtable nodeHash;
  private Hashtable edgeHash;

  private OntoReasoner reasoner;
  private String withGUI = getSystemProperty("service.planner.gui.flag", "SERVICE_PLANNER_GUI_FLAG");
  private static Logger logger = Logger.getLogger("ServicePlanner");
  private String logFileName = getSystemProperty("planner.log.file", "PLANNER_LOG_FILE");
  private String prodigyExec = getSystemProperty("prodigy.exec", "PRODIGY_EXEC");
  private String prodigyDump = getSystemProperty("prodigy.dump", "PRODIGY_DUMP");
  private String prodigyTypes = getSystemProperty("prodigy.types", "PRODIGY_TYPES");
  private String prodigyTimeout = getSystemProperty("prodigy.timeout", "PRODIGY_TIMEOUT");
  private String prodigyDepthBound = getSystemProperty("prodigy.depth.bound", "PRODIGY_DEPTH_BOUND");
  private String ACL_HOME = getSystemProperty("acl.home", "ACL_HOME");
  private String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME};
  private String aclExecutable = findSystemExecutable(ACL_HOME, "lisp", "alisp");
  private String lispLogFileName = getSystemProperty("acl.log.file", "ACL_LOG_FILE");
  private static final int defaultNice = 15;

  private OWLJDom owlJDom = OWLJDom.getInstance();

  private class MyComp implements Comparator
  {
    public int compare(Object o1, Object o2)
    {
      if ((o1 instanceof ServicePlan) && (o2 instanceof ServicePlan)){
        ServicePlan plan1 = (ServicePlan) o1;
        ServicePlan plan2 = (ServicePlan) o2;
        Graph g1 = plan1.getGraph();
        Graph g2 = plan2.getGraph();
        int v1 = g1.getVerticesCount();
        int v2 = g2.getVerticesCount();
        //        System.out.println("comp: " + v1 + " " + v2);
        
        if (v1 < v2) return -1;
        else if (v1 > v2) return 1;
      }
      return 0;
    }
  }
  
  public ServicePlanner()
  {
    super();
    reasoner =  PellReasoner.getInstance();

    logger.setLevel(Level.DEBUG);
    //    logger.setLevel(Level.OFF);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }

  public Vertex getStartNode()
  {
    return (Vertex) nodeHash.get("Source");
  }
  
  public Vertex getEndNode()
  {
    return (Vertex) nodeHash.get("Sink");
  }
    
  
  /** 
   * @param query A reasoning query to be answered.
   * @param registry A service registry containing all available services 
   & @return A string containing an XML representation of PRODIGY plans
  */
  public List findProdigyPlans(Query query, ServiceRegistry registry, int maxLength)
  {
    List result = new ArrayList();
    reasoner =  PellReasoner.getNewReasoner();
    StringBuffer description = new StringBuffer();
    int counter = 1;
    serviceHash = new Hashtable();
    nodeHash = new Hashtable();
    edgeHash = new Hashtable();
    
    List queryInputs = query.getProblem().getInputs();
    List queryOutputs = query.getProblem().getOutputs();
    
    /*  First check whether all the objects bound to inputs of the query are valid OWL individuals */
    for(int i=0; i<queryInputs.size(); i++){
      Input in = (Input) queryInputs.get(i);
      reasoner.tellIndividual(in.getContents());
    }
    if (!reasoner.isValid()) return null;
   
    description.append("(create-problem-space 'mathserv :current t)\n\n");
    //     description.append(reasoner.getProdigyClassTree());
    String typeDecls = readFile(prodigyTypes);
    description.append(typeDecls);
    description.append("\n");
    
    ArrayList services = registry.getAllEntries(); // list of reasoning services
    for (int i =0; i< services.size(); i++){
      ServiceEntry e = (ServiceEntry) services.get(i);
      ReasoningService s = (ReasoningService) e.getService();
      serviceHash.put(s.getURI().substring(7).toUpperCase(), s);
      //      System.out.println("s: " + s.getName().toUpperCase());
      if (maxLength > 0) 
        description.append(s.toProdigyOperator(counter, false));
      else 
        description.append(s.toProdigyOperator(counter, true));

      counter += ReasoningService.maxObjs;
      description.append("\n");
    }
    description.append(query.toProdigyProblem());
    //  System.out.println("descr: " + description);

    File inputFile = null;
    File outputFile = null;
    
    try{
      inputFile = File.createTempFile("prodigy", ".in");
      outputFile = File.createTempFile("prodigy", ".out");
      writeFile(inputFile, description.toString());
      writeFile(msHome + "/etc/data/broker/test.lisp", description.toString());
      int timeout = Integer.parseInt(prodigyTimeout);
      int depthBound = Integer.parseInt(prodigyDepthBound);

      if (maxLength > 0) 
        depthBound = 8 * maxLength;
      
      
      String command = prodigyExec + " " + aclExecutable + " " + prodigyDump + 
        " " + inputFile.getAbsolutePath() + " " + timeout + " " + depthBound + 
        " " +outputFile.getAbsolutePath()  +
        " >" + lispLogFileName + "-prodigy.log";
      System.out.println("command: " + command);
    
      String[] commands = {"sh", "-c", command};
      logger.debug("[ServicePlanner] starting prodigy...");
      this.start(command, acl_envs, defaultNice, timeout + 10);
      String plansXML = readFile(outputFile);
      System.out.println("plans: " + plansXML);
      
      writeFile(msHome + "/etc/data/broker/test.xml", plansXML);
      
      Element root = owlJDom.getRootElement(plansXML);
      //      System.out.println("root: " + root);
      List plans = root.getChildren("plan");
      //      System.out.println("plans: " + plans);
      
      for(int i = 0; i< plans.size(); i++){
        Element plE = (Element) plans.get(i);
        ServicePlan plan = new ServicePlan(null, plE);
        result.add(plan);
      }
    } catch(Exception e){
      e.printStackTrace();
      logger.debug("[ServicePlanner] Exception " + e);
      return null;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[ServicePlanner] Exception " + e); }
    }
    return result;
  }
  /* prodigy-mw ${lisp} prodigy-i386_linux.dxl /home/jzimmer/import/mathserve/thirdParty/prodigy/working/domains/mathserv/domain5.lisp 3.0 /home/jzimmer/test/prodigy5.out
   */

  private VertexImpl findStartVertex(DirectedGraph graph)  {
    VertexImpl source = null;
    for(Iterator it = graph.getVerticesIterator(); it.hasNext();){
      VertexImpl v = (VertexImpl) it.next();
      String label = v.getLabel();
      if ((label.length() == 6) && (label.substring(0,6).equals("Source"))) {
        source = v; 
      }
    }
    return source;
  }
  
  private int distance(VertexImpl from, VertexImpl to, DirectedGraph graph)  {
    return 1;
  }

  private boolean equalModObj(VertexImpl v1, VertexImpl v2)
  {
    Object o1 = v1.getObject();
    Object o2 = v1.getObject();
    return (o1 == o2);
  }

  private VertexImpl containsModObj(VertexImpl v, List vcs)
  {
    Object obj = v.getObject();
    for(int i = 0; i< vcs.size(); i++){
      VertexImpl ver = (VertexImpl) vcs.get(i);
      Object o = ver.getObject();
      if (o == obj) return ver;
    }
    return null;
  }

  /**
   * alist - blist
   */
  private List listDifferenceMod(List alist, List blist){
    List result = new ArrayList();
    for(int i=0; i< alist.size(); i++){
      VertexImpl v = (VertexImpl) alist.get(i);
      if (containsModObj(v, blist) == null)
        result.add(v);
    }
    return result;
  }


  private List listDifference(List alist, List blist){
    List result = new ArrayList();
    for(int i=0; i< alist.size(); i++){
      VertexImpl v = (VertexImpl) alist.get(i);
      if (!blist.contains(v))
        result.add(v);
    }
    return result;
  }
  
  /**
   * Create an ND proof for a TSTP FOF problem given a FO-ATP result.
   * 
   * @param agraph A directed graph with a source vertex 'Start' and a sink 'Finish'
   * @param bgraph Another directed graph with a source vertex 'Start' and a sink 'Finish'
   * @return true, if agraph is the same graph as bgraph.
  */
  public boolean equal(DirectedGraph agraph, DirectedGraph bgraph)  {   
    if ((agraph != null) && (bgraph != null)){
      ArrayList avcs = new ArrayList(agraph.getVertexSet());
      ArrayList bvcs = new ArrayList(bgraph.getVertexSet());
      
      List diff = listDifference(bvcs, avcs);
      //List reversediff = listDifference(avcs, bvcs);
      if (!diff.isEmpty()) return false ;
      //    if (!reversediff.isEmpty()) return false;
      
      List ins = null;
      List outs = null;
      List seen = new ArrayList();
      
      for(int i = 0; i< avcs.size(); i++){
        VertexImpl av = (VertexImpl) avcs.get(i);
        if (av!=null){
          try{ 
            List a_outs = agraph.getOutgoingAdjacentVertices(av);
            if (bvcs.contains(av)){
              List b_outs = bgraph.getOutgoingAdjacentVertices(av);
              if ((a_outs!=null) && (b_outs!=null)) {
                diff = listDifference(a_outs, b_outs);     
                if (!diff.isEmpty()) return false ;
              }
            }
            
          } catch(Exception ex){
            ex.printStackTrace();
            return false;
          }
        }
        
      }
      
      return true;
    }
    return false;
  }
  
  
  /**
   * Translates the XML representation of PRODIGY plans into a Graph representation for the Broker GUI.
   * 
   * @param query A reasoning Query. 
   * @param plansXML A string containing the XML representation of PRODIGY plans for the given query. 
   * @return A List containing DirectedGraphs. One graph for each prodigy plan. 
  */
  public void plansComputeGraphs(Query query, List plans) { 
    
    List queryInputs = query.getProblem().getInputs();
    List queryOutputs = query.getProblem().getOutputs();
    try{
      //      ArrayList graphs = new ArrayList();
      
      VertexImpl startVertex = new VertexImpl("Source");
      VertexImpl endVertex = new VertexImpl("Sink");
      nodeHash.put("Source", startVertex);
      nodeHash.put("Sink", endVertex);

      for(int i=0; i<plans.size(); i++){
        ServicePlan sPlan = (ServicePlan) plans.get(i);
        Element plan = sPlan.getProdigyPlan();
        //        System.out.println("planE: " + plan);
        List ops = plan.getChildren("operator");

        //      for(int i=0; i<1; i++){
        Hashtable inputsMap = new Hashtable();
        Hashtable outputsMap = new Hashtable();
        Hashtable outputArgsMap = new Hashtable();
        
        DirectedGraph graph = new DirectedGraphImpl();
        graph.add(startVertex);
        graph.add(endVertex);
        
        for(int m=0; m<queryInputs.size(); m++){
          Input qin = (Input) queryInputs.get(m);
//           String qinName = qin.getName().toUpperCase();
//           String paramName = query.getName().toUpperCase() + "#" + qinName;
          String qinName = qin.getName();
          String paramName = query.getName().toUpperCase() + "#" + qinName.toUpperCase();
          outputsMap.put(paramName, startVertex);
          outputArgsMap.put(paramName, qinName);
        }
        
        // for each instantiated operator in the plan look at its arguments (args)
        for(int k=0; k<ops.size(); k++){
          Element op = (Element) ops.get(k);
          String opName = op.getAttributeValue("name").trim();
          //          System.out.println("op: " + opName);
          ArrayList args = new ArrayList(op.getChildren("arg"));
          //          System.out.println("args: " + args);

          // get the object names of the arguments 
          for(int l=0; l<args.size(); l++){
            Element arg = (Element) args.get(l);
            //            System.out.println("arg: " + arg.getAttributeValue("name"));
            args.set(l, arg.getAttributeValue("name"));
          }
          
          // retrieve the reasoning service corresponding to the operator
          ReasoningService serv = (ReasoningService) serviceHash.get(opName);
          //          System.out.println("serv: " + serv.getName());
          String servName = serv.getName();
 
//           String[] split = servName.split("#");
//           if (split.length > 1){ servName = split[1];}
          VertexImpl v = null;
          
          if (nodeHash.containsKey(servName))
            v = (VertexImpl) nodeHash.get(servName);
          else {
            // create a new node for this service
            v = new VertexImpl(servName);
            v.setObject(serv);
            nodeHash.put(servName, v);
          }
          
          // TODO: Is this dangerous? If a service is called more than once we can't differentiate the occurences anymore!!!
          graph.add(v);
          
          // now do something with the args...
          List inputs = serv.getProblem().getInputs();
          for(int m=0; m<inputs.size(); m++){
            Input in = (Input) inputs.get(m);
            String iName = in.getName();
            String argObj = (String) args.get(m);
            //            inputsMap.put(servName + "#" + iName, argObj);
	    
            // lookup the source node and the output parameter name feeding
            // this input parameter via the common object.
            VertexImpl source = (VertexImpl) outputsMap.get(argObj);
            String oName = (String) outputArgsMap.get(argObj);

            //            System.out.println("source output param: " + oName);
            
            if (source != null){
//               Edge e = new DirectedEdgeImpl(source, v);
              Annotation ann = new Annotation(oName, iName);
              AnnotatedEdge e = new AnnotatedEdge(source, v, ann);          
              String label;
              e.setFollowVertexLabel(false);
              //               label = ann.toString();
              label = oName + "::" + iName;
              e.setLabel(label.toLowerCase());
              graph.addEdge(e);	      
            }
          }

          // put the output of this node in the outputsMap table.
          // with the object in the planning operator as a key and the node as a value
          int offset = inputs.size();
          List outputs = serv.getProblem().getOutputs();
          for(int m=0; m<outputs.size(); m++){
            Output out = (Output) outputs.get(m);
            String oName = out.getName();
            String outArg = (String) args.get(offset + m);
            outputsMap.put(outArg, v);
            outputArgsMap.put(outArg, oName);
          }
        }

        // last node: connect the last nodes in the graph with the outputs of the End node.
        for(int n=0; n<queryOutputs.size(); n++){
          Output out = (Output) queryOutputs.get(n);
          String oName = out.getName();
          String qoutName = query.getName().toUpperCase() + "#" + oName.toUpperCase();
          VertexImpl source = (VertexImpl) outputsMap.get(qoutName);
          String sourceArgName = (String) outputArgsMap.get(qoutName);
          if (source != null){
            //            Edge e = new DirectedEdgeImpl(source, endVertex);
            //            System.out.println("edge: " + sourceArgName + " " + oName);
            Annotation ann = new Annotation(sourceArgName, oName);
            AnnotatedEdge e = new AnnotatedEdge(source, endVertex, ann);
            e.setFollowVertexLabel(false);
            String label = ann.toString();
            ////e.setLabel(oName + "::" + out.getType());
            //   e.setLabel(sourceArgName + "::" + oName);
            e.setLabel(label);
            graph.addEdge(e);	      
          }
        }
        sPlan.setGraph(graph);
        
//         //    if (!reasoner.isValid()) return null;
        
//         graphs.add(graph);
        //        
      }
      MyComp comparator = new MyComp();
      
      // sort the graphs according to size (number of nodes)
      Collections.sort(plans, comparator);
      
    } catch(Exception e){
      e.printStackTrace();
      logger.debug("[ServicePlanner] Exception " + e);
    }
  }
  

  /**
   * Create an ND proof for a TSTP FOF problem given a FO-ATP result.
   * 
   * @param agraph A directed graph with a source vertex 'Start' and a sink 'Finish'
   * @param bgraph Another directed graph with a source vertex 'Start' and a sink 'Finish'
   * @return true, if agraph subsumes bgraph, i.e. bgraph looks exactly like agraph except
   that it has one or more verteces in between two adjacent verteces of agraph. <br>
   Example: agraph contains A->C and bgraph contains A->B->D->C instead. <br>
   B and D are redundant steps.
   NOTE: THIS ALGO IS FAST BUT _WRONG_!!! 
  */
  public boolean subsumes(DirectedGraph agraph, DirectedGraph bgraph)
  {
    //    System.out.println("subsumes: " + agraph + ":" + bgraph);
    ArrayList avcs = new ArrayList(agraph.getVertexSet());
    ArrayList bvcs = new ArrayList(bgraph.getVertexSet());
    //    List diff = listDifferenceMod(bvcs, avcs);

    /* diff contains all the nodes that are in bgraph but not in agraph */
    List diff = listDifference(bvcs, avcs);
    /* diff contains all the nodes that are in agraph but not in bgraph */
    List reversediff = listDifference(avcs, bvcs);

    if (diff.isEmpty()) return false;
    //if (diff.size() > 1) return false;
    if (!reversediff.isEmpty()) return false;
    
    List ins = null;
    List outs = null;
    List seen = new ArrayList();
    
    for(int i = 0; i<diff.size(); i++){
      VertexImpl v = (VertexImpl) diff.get(i);
      //      System.out.println("extra node: " + v);
      if (!seen.contains(v)) {
        
        /* ins: vertices that deliver intput to extra node */
        ins = bgraph.getIncomingAdjacentVertices(v);
        /* outs: vertices that take output of extra node */
        outs = bgraph.getOutgoingAdjacentVertices(v);
        
        boolean apath = false; 
        boolean end = false;
        
        do{
          /* is there exactly one vertex feeding and one being fed by extra node? */
          if ((ins.size() == 1 && (outs.size() == 1))){
            VertexImpl b_in = (VertexImpl) ins.get(0);
            VertexImpl b_out = (VertexImpl) outs.get(0);
            VertexImpl a_in;            
            VertexImpl a_out;
            
//             VertexImpl a_in = containsModObj(b_in, avcs);
            /* does agraph contain the predecessor vertex of bgraph? */
            if (avcs.contains(b_in)) a_in = b_in;
            else a_in = null;
//             VertexImpl a_out = containsModObj(b_out, avcs);
            /* does agraph contain the successor vertex of bgraph? */
            if (avcs.contains(b_out)) a_out = b_out;
            else a_out = null;
            
            //            System.out.println(" nodes: " + b_in + " " + b_out + " " +a_in + " " +a_out);
            if(a_in != null) 
              if (a_out != null){
                /* found corresponding vertices in agraph */
                end = true;
                /* now check whether they are directly adjacent in agraph */
                Edge e = agraph.getEdge(a_in, a_out);
                if (e != null)
                  apath = true; 
                else 
                apath = false;
              } else {
                seen.add(b_out);
                outs = bgraph.getOutgoingAdjacentVertices(b_out);
              }
            else {
              seen.add(b_in);
              ins = bgraph.getIncomingAdjacentVertices(b_in);
            }
          }
          else 
            /*  there is one vertex feeding extra node and two vertices being fed */
            if ((ins.size() == 1 && (outs.size() == 2))){
              VertexImpl b_in = (VertexImpl) ins.get(0);
              VertexImpl b_out1 = (VertexImpl) outs.get(0);
              VertexImpl b_out2 = (VertexImpl) outs.get(1);
              VertexImpl a_in;            
              VertexImpl a_out1;
              VertexImpl a_out2;
              /* does agraph contain the predecessor vertex of bgraph? */
              if (avcs.contains(b_in)) a_in = b_in;
              else a_in = null;
              /* does agraph contain the 2 successor verteces of bgraph? */
              if (avcs.contains(b_out1)) a_out1 = b_out1;
              else a_out1 = null;
              if (avcs.contains(b_out2)) a_out2 = b_out2;
              else a_out2 = null;
              
              //              System.out.println(" node triplet: " + b_in + " " + b_out1 + 
              //                                 " " + b_out2 + " " +a_in + " " + a_out1 + " " + a_out2);
              
              if(a_in != null) 
                if ((a_out1 != null) && (a_out2 != null)){
                  /* found corresponding vertices in agraph */
                  end = true;
                  /* now check whether they are directly adjacent in agraph */
                  Edge e1 = agraph.getEdge(a_in, a_out1);
                  Edge e2 = agraph.getEdge(a_in, a_out2);
                  if ((e1 != null) && (e1 != null))
                    apath = true; 
                  else 
                    apath = false;
                } else {
                  seen.add(b_out1);
                  seen.add(b_out2);
                  outs = bgraph.getOutgoingAdjacentVertices(b_out1);
                  outs.addAll(bgraph.getOutgoingAdjacentVertices(b_out2));
                }
              else {
                seen.add(b_in);
                ins = bgraph.getIncomingAdjacentVertices(b_in);
              }
            } else {end = true; apath = false; }
        } while (!end);
        if (!apath) 
          return false;
      }
    }
    return true;
  }
  
    
  /**
   * Translates the XML representation of PRODIGY plans into a Graph representation for the Broker GUI.
   * 
   * @param query A reasoning Query. 
   * @param plansXML A string containing the XML representation of PRODIGY plans for the given query. 
   * @return A string representing the DTGolog domain for containing all services occuring in one of the PRODIGY plans, 
   *    and a procedure with the query name.
  */
  public String plans2DTGologDomain(Query query, List plans, ServiceRegistry registry) {
    StringBuffer result = new StringBuffer();
    
    try{
      ReasoningProblem problem = query.getProblem();
      List inputs = problem.getInputs();
      List outputs = problem.getOutputs();
      ArrayList params = new ArrayList();
      params.addAll(inputs);
      params.addAll(outputs);
        
      result.append("proc(" + ReasoningService.actionString(query.getURI(), params) + ", \n   ");
      //       Element root = owlJDom.getRootElement(plansXML);
      //       System.out.println("root: " + root);
      //       List plans = root.getChildren("plan");
      //       System.out.println("plans: " + plans);
      List prefix = commonPrefix(plans);
      List prefixElements = new ArrayList();
      
      /*  first write the steps of the common prefix .... */
      for(int i=0; i < prefix.size(); i++){
        ArrayList prefixList = (ArrayList) prefix.get(i);
        for(int j=0; j < prefixList.size(); j++){
          Element prefixElem = (Element) prefixList.get(j);
          //          System.out.println("prefix node: " + prefixElem);
          prefixElements.add(prefixElem);
          if (j == 0){
            String prefixElemName = prefixElem.getAttributeValue("name").trim();
            //            System.out.println("prefixElem name: " + prefixElemName);
            ArrayList args = new ArrayList(prefixElem.getChildren("arg"));
            //            System.out.println("args: " + args);
            
            //           // get the object names of the arguments 
            for(int l=0; l<args.size(); l++){
              Element arg = (Element) args.get(l);
              String name = arg.getAttributeValue("name");
              String[] split = name.split("#");
              if (split.length > 1){ name = split[1];}
            
              //              System.out.println("arg: " + name);
              args.set(l, name);
            }
            ReasoningService serv = (ReasoningService) serviceHash.get(prefixElemName);
            String servStr = ReasoningService.actionStringWithArgnames(serv.getURI(), args);
            
            if (i > 0) result.append("   ");
            result.append(servStr);
          }
        }
        result.append(" :\n");
      }
      
      /*  ... then write a nondeterministic choice for rest of all the plans */
      StringBuffer suffixStr = new StringBuffer();
      for(int i=0; i<plans.size(); i++){
        //         Hashtable inputsMap = new Hashtable();
        //         Hashtable outputsMap = new Hashtable();
        //         Hashtable outputArgsMap = new Hashtable();
      
        ServicePlan sPlan = (ServicePlan) plans.get(i);
        Element plan = sPlan.getProdigyPlan();
        //        System.out.println("planE: " + plan);
        List ops = plan.getChildren("operator");
      
        int stepsLeft = 0;
        StringBuffer planStr = new StringBuffer();
        
        for(int k=0; k<ops.size(); k++){
          Element op = (Element) ops.get(k);
          if (!prefixElements.contains(op)){
            stepsLeft += 1;
            String opName = op.getAttributeValue("name").trim();
            //System.out.println("insertig op: " + opName);
            ArrayList args = new ArrayList(op.getChildren("arg"));
            //           // get the object names of the arguments 
            for(int l=0; l<args.size(); l++){
              Element arg = (Element) args.get(l);
              String name = arg.getAttributeValue("name");
              String[] split = name.split("#");
              if (split.length > 1){ name = split[1];}
              
              //System.out.println("arg: " + name);
              args.set(l, name);
            }
        
            /* retrieve the reasoning service corresponding to the operator */
            ReasoningService serv = (ReasoningService) serviceHash.get(opName);
            String servStr = ReasoningService.actionStringWithArgnames(serv.getURI(), args);
            
            /* are we beyond the common prefix? */
            if (k > prefix.size()) 
              planStr.append("   ");
            planStr.append(servStr);

            /* are there still more steps in the plans' step sequence? */
            if (k < ops.size() - 1)
              planStr.append(" :\n");
          }
        }
        /* are there any steps more than the common prefix? */
        if (stepsLeft > 0) { 
          /* if it's more than one step, put () around them */
          if (stepsLeft > 1) 
            suffixStr.append("(");
          suffixStr.append(planStr);
          if (stepsLeft > 1) 
            suffixStr.append(")");
          
          /* are there more plans to come? */
          if (i < plans.size() - 1)
            suffixStr.append(" #\n  ");
          else 
            suffixStr.append(")");        
        }
      } // end for(plans)
      
      /* are there any more steps by any of the plans? */
      if (suffixStr.length() > 0) {
        /* the prefix was not empy? */
        if (prefix.size() > 0)
          result.append(" (");
        else 
          result.append("\n (");
        
        result.append(suffixStr);
        result.append(")");
      }  
      /* end the procedure declaration */
      result.append(".\n\n");
      
    } catch(Exception e){
      e.printStackTrace();
      logger.debug(e);
    }
    ArrayList services = registry.getAllServices();
     result.append(ReasoningService.computeDTGologDomain(services));
    return result.toString();
  }

  public String plan2ServiceURIs(ServicePlan sp){
    Element pp = sp.getProdigyPlan();
    Element op = pp.getChild("operator");
    String opName = op.getAttributeValue("name").trim();

    ReasoningService serv = (ReasoningService) serviceHash.get(opName);
    return serv.getURI();
  }
  
  public String gologXml2Math(String policyXml)   {    
    StringBuffer buffer = new StringBuffer();
    try{
      Element root = owlJDom.getRootElement(policyXml);
      buffer = gologXml2Math(root, "");
    } catch(Exception ex){
    }
    return buffer.toString();
  }

  
  /**
     Merging Plans:
     ==============
start: the full set of all plans. 
1) build subsets of plans for all types of root nodes (services) reachable directly from Start node
 (take node with distance 1, where distance max lenghts of paths to that node)
2) in each subset with cardinality > 1, try do merge(subset) plans as far as possible

merge(subset):
--------------
1. if all nodes of all plans are the same, simply collapse them to one node in the new graph.
2. go one level deeper (look into graph algo's that start in source node and explore graph in levels 
   a level $n$  is the set of all nodes of certain distance $n$ from current node.
3. if all nodes (services) on certain level are of the same classification (eg. FO-ATP), then  merge them together to a Choice node. 
4 the paths in the choice node are just the remainding paths of the single graphs.

Problems:
---------
- how do we map the I/O parameters properly?
- for the remainding paths in the graph we also have to get the dataflow right...


    Execution of merged Plans:
    ==========================
- At Choice nodes, the broker has to use analysis services (or just built-in) and the specialisations of the services to put a rating on the nodes. Then it chooses the service with the highest rating 
- For this, ratings have to be numerical values of expected success on a problem class.
- for the analysis we need to see which parameters correspond (looking at
  the types (OWL classes) of the parameters).

Actually!:
----------
- We only need to see whether all edges have the same type, and they must,
  because the preceeding service is all the same for all plans, so we just
  find the input of a certain type and look for the specialisation value on
  this input in the Profile.

  */    

  /**
   * Translates DirectedGraphs into the DOT representation of GraphViz
   * 
   * @param query A directed Graph
   * @return A string containing the DOT representation of the input graph.
  */
  public static String graph2Dot(DirectedGraph g) 
  {
    StringBuffer dot = new StringBuffer();
    dot.append("digraph g {\n");
    //fontnames == Times Helvetica Courier Symbol Times-Italic Times-Bold   don't work!
    String fontname = "courier"; 
    VertexImpl v1, v2;
    try{
      Iterator it = g.getVerticesIterator();
      while(it.hasNext()){
        Object obj = it.next();
        //        System.out.println("obj: " + obj.getClass().getName());
        VertexImpl v = (VertexImpl) obj;        
        String nodeLabel = v.getLabel();
        //         if (!(nodeLabel.equals("Sink") || nodeLabel.equals("Source"))){
        //           ReasoningService serv = (ReasoningService) v.getObject();
        //            nodeLabel = serv.getName();
          
        
        //         String label = v.getLabel();
        //         String[] parts = label.split("/");
        //         int last = parts.length -1;
        //        String nodeLabel = v.getLabel();
        
        //         if (last > 0)
          //           nodeLabel = parts[last];
        //        System.out.println("node: " + nodeLabel);
        
        dot.append("node" + nodeLabel + "[shape=ellipse label=\"" + nodeLabel + "\" fontname=\""+fontname+"\" fontsize=8];\n");
      }
      Set edges = g.getEdgeSet();
      it = edges.iterator();
      while(it.hasNext()){
        Object obj = it.next();
        //        System.out.println(obj + " type: " + obj.getClass().getName());
        if (obj instanceof AnnotatedEdge){
          AnnotatedEdge e = (AnnotatedEdge) obj;
          
          //          String label = e.getLabel();
          Object obj1 = e.getVertexA();          
          Object obj2 = e.getVertexB();          
          //          System.out.println("obj1: " + obj1.getClass().getName());
          //          System.out.println("obj2: " + obj2.getClass().getName());
          v1 = (VertexImpl) obj1 ;
          v2 = (VertexImpl) obj2;
          
          Annotation ann = e.getAnnotation();
          String edgeLabel = "(" + ann.getSource() + ", " +  ann.getTarget() + ")";
          
          //         String[] split = edgeLabel.split("::");
          //         if (split.length > 1){ edgeLabel = split[1];}
          
          //        dot.append("\"node" +  v1.getLabel() + "\" -> \"node" + v2.getLabel() + "\" [label=\""+e.getLabel()+"\" fontname=\"courier\" fontsize=8];\n");
//           String[] parts = v1.getLabel().split("/");
//           int last = parts.length -1;
          String nodeLabel1 = v1.getLabel();
//           if (last > 0)
//             nodeLabel1 = parts[last];
//           parts = v2.getLabel().split("/");
//           last = parts.length -1;
          String nodeLabel2 = v2.getLabel();
//           if (last > 0)
//             nodeLabel2 = parts[last];
          
//           System.out.println("edge: " + nodeLabel1 + " --> " + nodeLabel2);
          
          dot.append("\"node" +  nodeLabel1 + "\" -> \"node" + nodeLabel2 + "\" [label=\""+edgeLabel+"\" fontname=\"courier\" fontsize=8];\n");
          
        }
      }
      
      dot.append("}\n");
      logger.debug("dot graph " + dot);
    } catch(Exception e){
      e.printStackTrace();
      logger.debug(e);
    }
    return dot.toString();
    
  }


  
    
  /**
   * Translates the XML representation of a GOLOG procedure into a math-style syntax for the Broker GUI.
   * 
   * @param policyRoot The JDom root element of the policy XML document. 
   * @return A string representing the Golog procedure.
  */
  public StringBuffer gologXml2Math(Element policyNode, String indent)   {    
    //    logger.info("executeRec called with " + node);
    //    System.out.println("given time: " + givenTime);
    StringBuffer buffer = new StringBuffer();
    
    String name = policyNode.getName();
    
    if (name.equals("program")){ 
      List sequents = policyNode.getChildren();
      Element head = (Element) sequents.get(0);
      Element body = (Element) sequents.get(1);
      buffer.append("proc");
      buffer.append(gologXml2Math(head, " "));
      buffer.append("\n");
      buffer.append(gologXml2Math(body, "  "));
      buffer.append("end\n");
      return buffer;
    }
    
    if (name.equals("head") || name.equals("body")){ 
      List sequents = policyNode.getChildren();
      Element seq = (Element) sequents.get(0);
      buffer.append(gologXml2Math(seq, " "));
      return buffer;
    }

    if (name.equals("stop")){ 
      buffer.append(indent);
      buffer.append("stop\n");
      return buffer;
    }
    
    if (name.equals("sequence")){
      List sequents = policyNode.getChildren();
      for (int i = 0; i< sequents.size(); i++){
        Element seq = (Element) sequents.get(i);
        String seqName = seq.getName();
        if (! (seqName.equals("senseEffect") || seqName.equals("test"))){
          buffer.append(gologXml2Math(seq, indent));
          if (i < (sequents.size() -1)) 
            buffer.append(" ;");
          if (! seqName.equals("if-then-else"))
            buffer.append("\n");
          //           else 
          //             buffer.append("\n");
        }
      }
      return buffer;
    }
    
    /**        SENSE-EFFECT     */
    else if (name.equals("senseEffect")){
      return buffer;
    }
    
    /**        IF-THEN-ELSE      */
    else if (name.equals("if-then-else")){
      List children = policyNode.getChildren();
      Element cond = (Element) children.get(0);
      Element thenPart = (Element) children.get(1);
      Element elsePart = (Element) children.get(2);
      buffer.append(indent);
      buffer.append("if ");
      buffer.append(gologXml2Math(cond, ""));
      buffer.append(" then\n");
      buffer.append(gologXml2Math(thenPart, indent + "  "));
      buffer.append(indent);
      buffer.append("else\n");
      buffer.append(gologXml2Math(elsePart, indent + "  "));
      buffer.append(indent);
      buffer.append("endIf\n");
      return buffer;
    }

//     /**          BIND         */
//     else if (name.equals("bind")){
//       String varName = policyNode.getAttributeValue("name");
//       String toVar = policyNode.getAttributeValue("to");
//       System.out.println("Explicitely binding variable " + varName + " to " + toVar + "!");
//       if (!this.bindings.containsKey(varName)){
//         if (this.bindings.containsKey(toVar)){
//           this.bindings.put(varName, this.bindings.get(toVar));
//           this.jdomBindings.put(varName, this.jdomBindings.get(toVar));
//           this.variableBindings.put(varName, this.variableBindings.get(toVar));
//         }
//       } else {
//         System.out.println("Variable " + varName + " is already bound. Stopping execution!");
//         return;
//       }
//     }

    /**          CHOICE         */
    else if (name.equals("choice")){
      List choices = policyNode.getChildren();
      for (int i = 0; i< choices.size(); i++){
        Element seq = (Element) choices.get(i);
        String seqName = seq.getName();
        if (!seqName.equals("senseEffect")){
          buffer.append(gologXml2Math(seq, indent + "  "));
          if (i < (choices.size() -1)) 
            buffer.append("| \n");
          else 
            buffer.append("\n");
        }
      }
      return buffer;
    }
    
    if (name.equals("appl")){
      String functor = policyNode.getAttributeValue("func");
      List params = policyNode.getChildren();
      buffer.append(indent);
      buffer.append(functor);
      buffer.append("(");
      for (int i = 0; i< params.size(); i++){
        Element par = (Element) params.get(i);
        String parName = par.getAttributeValue("name");
        buffer.append(parName);
        if (i < (params.size() -1)) 
          buffer.append(", ");
        else 
          buffer.append(")");
      }
    }
    return buffer;
    
  }

  /**
   * @param plans A list of ServicePlan instances. 
   * @return A list all JDom Elements containing steps of the prefix common to all ServicePlans given. 
   */
  public List commonPrefix(List plans) {
    List result = new ArrayList();
    if (!plans.isEmpty()){
      int counter = 0;
      boolean havePrefix = true;
      while(havePrefix){
        Element step = null;
        String stepName = null;
        ArrayList stepArgs = null;
        List localMatch = new ArrayList();
        /* go through all plans and first check whether the first step is common, then the second, etc. */ 
        for(int i=0; i<plans.size(); i++){
          ServicePlan sPlan = (ServicePlan) plans.get(i);
          Element plan = sPlan.getProdigyPlan();
          List ops = plan.getChildren("operator");
          if (ops.size() > counter){
            /* Are we looking at the first plan? */ 
            if (i == 0){
              step = (Element) ops.get(counter);
              stepName = step.getAttributeValue("name").trim();
              System.out.println("step name: " + stepName);
              stepArgs = new ArrayList(step.getChildren("arg"));
              localMatch.add(step);
            } else {
              Element otherStep = (Element) ops.get(counter);
              String opName = otherStep.getAttributeValue("name").trim();
              //              System.out.println("other step name: " + opName);
              ArrayList args = new ArrayList(otherStep.getChildren("arg"));
              if ((opName.equals(stepName)) && (args.size() == stepArgs.size())){

                /* Compare the argments of the steps by name */ 
                for(int l=0; l<args.size(); l++){
                  Element arg = (Element) args.get(l);
                  Element stepArg = (Element) stepArgs.get(l);
                  String argName = arg.getAttributeValue("name");
                  //                  System.out.println("other step arg: " + argName);
                  String stepArgName = stepArg.getAttributeValue("name");
                  if (!argName.equals(stepArgName))
                    havePrefix = false;
                }
                if (havePrefix) 
                  localMatch.add(otherStep);
              } else 
                havePrefix = false;
            }
          } else havePrefix = false;
        }
        if (havePrefix) 
          result.add(localMatch);
        counter++;
      } // end while (havePrefix)
    }
    return result;
  }

  
          
          //            if (source != null){

//               Edge e = new DirectedEdgeImpl(source, v);
//   {
//     System.out.println("2golog: " + query + " " + graphs);
    
//     String procHead = query.toGologProcedureHead(); 
//     StringBuffer procBody = new StringBuffer();
//     for (int g = 0; g< graphs.size(); g++){
//       DirectedGraph graph = (DirectedGraph) graphs.get(g);
//       List currentPolicyNodes = new ArrayList();
//       VertexImpl start = findStartVertex(graph);
//       if (start != null) 
//         currentPolicyNodes.add(start);
//       while(! currentPolicyNodes.isEmpty()){
//         List newCurrent = new ArrayList();
//         for(int n = 0; n< currentNodes.size(); n++) {
//           VertexImpl node = (VertexImpl) currentNodes.get(n);
//           System.out.println("node " + node);
//           List adjacent = graph.getOutgoingAdjacentVertices(start);
//           newCurrent.addAll(adjacent);
//           for(int v = 0 ; v< adjacent.size(); v++){
//             VertexImpl next  = (VertexImpl) adjacent.get(v);
//             System.out.println("  adjacent " + next);
//              AnnotatedEdge edge = (AnnotatedEdge) graph.getEdge(node, next);
//             Annotation ann = edge.getAnnotation();
//             System.out.println("  edge with annotation " + ann);
//  }
//         }
//         currentNodes = newCurrent;
//       }
//     }
//     return procHead + procBody.toString() + ").";
//   }
  
        /**
graph2Dot() creates Strings like the following which can be used 
as input for the graphviz dot tool:

digraph g {
node [shape = record,height=.1];
node0[label = "<f0> |<f1> G|<f2> "];
node1[label = "<f0> |<f1> E|<f2> "];
node2[label = "<f0> |<f1> B|<f2> "];

"node0":f2 -> "node4":f1;
"node0":f0 -> "node1":f1;
"node1":f0 -> "node2":f1;
"node1":f2 -> "node3":f1;
"node2":f2 -> "node8":f1;
"node2":f0 -> "node7":f1;
"node4":f2 -> "node6":f1;
"node4":f0 -> "node5":f1;
}
      */
  

  public static void main(String argv[]) throws Exception {
    System.out.println("null == null" + (null == null));
    String policy = BasicService.readFile("/home/jzimmer/policy.xml");
    ServicePlanner pl = new ServicePlanner();
    String math =  pl.gologXml2Math(policy);
    System.out.println(math);
  }
    

}

  
  

    
    
    //     out1 = graph1.getOutgoingAdjacentVertices(source1);
//     out2 = graph1.getOutgoingAdjacentVertices(source2);
    
//     try{
//       while (!end){
//         if (out1.size() != out2.size())
//           return false;
//         List objs1 = new ArrayList();
//         List newOut1 = new ArrayList();
//         List newOut2 = new ArrayList();
        
//         for(int i = 0; i<out1.size(); i++){
//           VertexImpl v = (VertexImpl) out1.get(i);
//           newOut1.addAll(graph1.getOutgoingAdjacentVertices(v));
//           objs1.add(v.getObject());
//         }
        
//         for(int j = 0; j<out2.size(); j++){
//           VertexImpl v = (VertexImpl) out2.get(j);
//           Object obj = v.getObject();
//           // if one of the vertices is not in graph1 
//           if (!objs1.contains(obj)){
//             // get the next nodes in this branch of graph2
//             List next  = graph2.getOutgoingAdjacentVertices(v);
//             if (next.size() != 1){
//               return false;
//             }
//             else {
//               v = (VertexImpl) next.get(0);
//               obj = v.getObject();
//               if (!objs1.contains(obj)){
//                 newOut2.addAll(graph2.getOutgoingAdjacentVertices(v));
//                 subsumes = true;
//                 else
//                   return false;
//               }
//             }
//           } else {
//             newOut2.addAll(graph2.getOutgoingAdjacentVertices(v));
//           }
//         }
//         if ((newOut1.size() == 1) && (newOut2.size()==1))
    

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


//         ReasoningService start = new ReasoningService();
//         start.setName("Source");
//         ReasoningProblem startProb = new ReasoningProblem();
//         startProb.outputs = queryInputs;
//         startProb.postconds = query.getProblem().getPreconds();
//         start.setProblem(startProb);

//         ReasoningService end = new ReasoningService();
//         start.setName("Finish");
//         ReasoningProblem endProb = new ReasoningProblem();
//         endProb.inputs = queryOutputs;
//         endProb.preconds = query.getProblem().getEffects();
//         end.setProblem(endProb);
        
// hack to avoid Otterfier loops
//     description.append("(Control-Rule reject-wrong-binding-for-www.mathweb.org/owl/trans/OtterfierService.owl2");
//     description.append("  (if (and");
//     description.append("        (applicable-operator (www.mathweb.org/owl/trans/OtterfierService.owl <old_result>  <new_result>))");
//     description.append("        (or");
//     description.append("         (exact-type-of-object <old_result> OWL-FOBrFPResult))");
//     description.append("         ))");
//     description.append("  (then sub-goal))");

//   public ArrayList findPopPlans(Query query, ServiceRegistry registry)
//   {
//     ArrayList graphs = new ArrayList();
    
//     DirectedGraph graph = null;
//     graph = findPlanPOP(query, registry);
//     while(graph != null) {
//       graphs.add(graph);
//       graph = generateNextPOP();
      
//     }
//     return graphs;

//   }
  
//   public DirectedGraph findPlanPOP(Query query, ServiceRegistry registry)
//   {
//     //    reasoner =  PellReasoner.getNewReasoner();
//     //    logger.debug("query to match: " + query.toString());
    
//     List inputs = query.getProblem().getInputs();
//     for(int i=0; i<inputs.size(); i++){
//       Input in = (Input) inputs.get(i);
//       reasoner.tellIndividual(in.getContents());
//     }
//     if (!reasoner.isValid()) return null;

//     StringBuffer planningProb = new StringBuffer();
//     ArrayList services = registry.getAllEntries(); // list of reasoning services
//     //    logger.debug("all services: " + services.toString());
    
//     Vector actions = new Vector(); // vector of planning operators for LogicSystem
    
//     serviceHash = new Hashtable();
//     for (int i =0; i< services.size(); i++){
//       //logger.debug("service " + services.get(i).toString());
//       ServiceEntry e = (ServiceEntry) services.get(i);
//       ReasoningService s = (ReasoningService) e.getService();
//       serviceHash.put(s.getName(), s);
//       Action a = s.toPlanningOperator();
//       if (i<10) 
//         System.out.println("operator: " + a.toLongString());
//       //        logger.debug("operator: " + a.toLongString());
//       actions.add(a);
//     }
    
// //     actions.addAll(reasoner.rulesToActions());
    
//     for (int i =0; i< actions.size(); i++){
//       planningProb.append(((Action) actions.get(i)).toProbString());
//     }
    
//     LogicSystem sys = new LogicSystem(actions, reasoner);
//     State initial = query.getInitialState();
//     State goal = query.getGoalState();
//     logger.debug("init state: " + initial.toString());
//     logger.debug("goal state: " + goal.toString());
//     planningProb.append("initial {\n");
//     planningProb.append(initial.toString());
//     planningProb.append("}\n");
//     planningProb.append("goal {\n");
//     planningProb.append(goal.toString());
//     planningProb.append("}\n");
//     writeFile(msHome +"/etc/data/broker/prob", planningProb.toString());
    
//     planner = new PartialOrderPlanner(sys, initial, goal);
//     planner.setMaxPlans(5);
//     plan = null;
//     if (withGUI.equals("yes")) {
//       POPGUI popgui = new POPGUI(planner);
//       plan = popgui.generatePlan();
//     } else 
//       plan = planner.generatePlan();
    
//     logger.debug("plan " + plan);
//     HashSet acts = plan.getActions();
//     Iterator it = acts.iterator();
//     while (it.hasNext()) logger.debug("act: " + it.next());
    
//     if (!(plan == null)){
//       DirectedGraph g = plan.toGraph(serviceHash);
//       logger.debug("graph " + g);
//       return g;
//     }
//     else
//       return null;
    
//     //     RegressionPlanner planner = new RegressionPlanner(sys);
//     //     LinearPlan plan = planner.generatePlan(initial, goal);
    
//     //     BasicPlanner planner = new BasicPlanner(sys);
//     //     Plan plan = planner.generatePlan(initial, goal);
    
//     //     logger.debug(" plan " + plan);
//   }
//   public DirectedGraph generateNextPOP()
//   {
//     plan = planner.generateNextPlan();
//     logger.debug("plan " + plan);
//     if (!(plan == null)){
//       DirectedGraph g1 = plan.toGraph(serviceHash);
//       logger.debug("graph " + g1);
//       return g1;
//     }else return null;
//   }
  
