 // -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.core;

import java.io.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.CommandThread;
import org.mathweb.util.BasicProperties;
import org.mathweb.util.ExpandedProperties;

public class BasicService 
{
  //  private static boolean verbose=true;
  private static boolean verbose=false;
  public static String msHome = System.getProperty("mathserv.home");

  protected static final String sep = File.separator;

  private static String defaultPropFileName = System.getProperty("os.name") + 
    "-" +System.getProperty("os.arch") + ".config";
  protected static BasicProperties props = getProperties(); 
  
  public static String cpuLimitedRun = getSystemProperty("cpu.limited.run", "CPU_LIMITED_RUN");
  
  private String[] presentedArgs = null;
  public File tempDir;
  
  /**
   * Sets the agents properties.
   * @param aProperties The properties to be used.
   */
//   public void setProperties(BasicProperties aProperties) {
//     properties = aProperties;
//   }
  
  public static BasicProperties getProperties() {
    ExpandedProperties exprops = new ExpandedProperties(); 
    String propFileName = defaultPropFileName;
    FileInputStream propertyStream = null;
    try {
      if (msHome == null)
        msHome = System.getProperty("catalina.home") + sep + ".." + sep+ "..";
      if (msHome == null){
        System.out.println("Could not fine MathServe Home!!!");
        System.exit(-1);
      }
      
      propFileName = msHome + sep + "conf" + sep + propFileName;
      //      System.out.println("Property file name: "+propFileName);
      
      propertyStream = new FileInputStream(propFileName);
      if (propertyStream != null) {
	exprops.load(propertyStream);
      }
    } catch (IOException ioe) {
      System.err.println("Error reading property file:" + propFileName);
      System.exit(-1);
    } catch (Exception any) {
      any.printStackTrace();
      // ignore - likely class not in package.
    } finally{
      if (propertyStream != null){
        //System.out.println("BasicService: closing property file stream");
        try{ propertyStream.close(); }
        catch(Exception e){
        }
      }
    }
    //    System.out.println("Props: " + props.toStringArray());
    return exprops;
  }

  /**
   * Called to set the agents arguments so we use it to
   * initialize our properties. It calls getProperties to
   * get the starting properties, copies them into a new instance
   * of ExpandedProperties, calls its parseArgs method of handle
   * the presented arguments, and then replaces our starting
   * properties with this result. This ensures that empty setting
   * in the arguments will overlay those in starting.
   * arguments.
   * @param args The arguments passed to the agent.
   */
//   public void setArguments(String[] args) {
//     presentedArgs = args;
//     ExpandedProperties newProperties = new ExpandedProperties();
//     newProperties.copyProperties(getProperties());
//     newProperties.parseArgs(args);
//     setProperties(newProperties);
//   }

  /**
   * Fetch the String[] which was given to setArguments.
   * @return The value given to setArguments - may be null;
   */
  public Object[] getArguments() {
    return presentedArgs;
  }

  public static FileAppender newAppender(String logFileName){
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    return appender;
  }
  
  public static String getSystemProperty(String property, String alternative) {
    String q= null;
    q = System.getProperty(property);
    if ( q != null ) {
      return q;
    }
    q = System.getProperty(alternative);
    if ( q != null ) {
      return q;
    } 
    q = props.getProperty(property);
    if ( q != null ) {
      return q;
    }
    q = props.getProperty(alternative);
    if ( q != null ) {
      return q;
    }
    else {
      throw new IllegalStateException ("The properties " +property+ " and " +alternative+"  is not set or has wrong type. This class refuses now to load.");
    }
  }
  
  public static String findSystemExecutable(String system_home, String cand1, String cand2) {
    File candidate = new File(system_home + sep + cand1);
    if ( candidate.isFile() ) return candidate.getAbsolutePath();
    candidate = new File(system_home + sep + cand2);
    if ( candidate.isFile() ) return candidate.getAbsolutePath();
    else throw new IllegalStateException("The system executable cannot be found. Is "+system_home+" properly defined, it should contain an executable file called \""+cand1+"\" or \""+cand2+"\".");
  }

  /*************************** Basic I/O methods *********************** */

  /** 
   * This method creates a new temporary directory with a given name prefix.
   * @param prefix the prefix for the new temporary dir
   */
  public void mkTempDir(String prefix) {
    //    System.out.println("mktempdir("+prefix+")");
    try {
      tempDir = File.createTempFile(prefix,"");
      tempDir.delete();
      tempDir.mkdir();
    } catch (IOException e) {
      System.err.println("Temporary directory with prefix " + prefix + "could not be created. Choosing another name.");
       try { 
        tempDir = File.createTempFile("","");
        tempDir.delete();
        tempDir.mkdir();
      } catch (IOException ioe) { 
        tempDir = null;
      } 
    }
  }
    
  /** 
   * This method removes a previously created temporary directory.
   * @param prefix the prefix for the new temporary dir
   */
  public void removeTempDir() {
    if (tempDir != null) {
      tempDir.delete();
      tempDir = null;
    }
  }
  
  /** 
   * Reads the contens of a file in a java.lang.String.
   * @param fileName the name of the file to read.
   * @return The contents of the file.
   */
  public static String readFile(String fileName){
    FileInputStream isFile  = null;
    try {
      isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    } finally {
      if (isFile != null)
        try{ isFile.close(); } catch (Exception e) {}
    }
  }
  

  /** 
   * Reads the contens of a file in a java.lang.String.
   * @param file the java.io.File to read from.
   * @return The contents of the file.
   */
  public String readFile(File file){
    FileInputStream isFile = null;
    try {
      isFile = new FileInputStream(file);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    } finally {
      if (isFile != null)
        try{ isFile.close(); } catch (Exception e) {}
    }
  }
  
  /** 
   * Writes a string into a file
   * @param fileName the name of the file to write into.
   * @param content the string to be written.
   * @return True if the file as written successfully, false otherwise.
   */
  public boolean writeFile(String fileName, String content){
    FileOutputStream outFile = null;
    try {
      outFile = new FileOutputStream(fileName) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    } finally {
      if (outFile != null)
        try{ outFile.close(); } catch (Exception e) {}
    }
  }
  
  /** 
   * Writes a string into a file
   * @param file the java.io.File to write into.
   * @param content the string to be written.
   * @return True if the file as written successfully, false otherwise.
   */
  public boolean writeFile(File file, String content){
    FileOutputStream outFile = null;
    try {
      outFile = new FileOutputStream(file) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    } finally {
      if (outFile != null)
        try{ outFile.close(); } catch (Exception e) {}
    }
  }
      
  private static void storeTime(String key){
    return;
  }
  
  /*************************** External Process Handling *********************** */

  /** 
   * runs a shell command in a new CommandThread. 
   * @param command the command string.
   * @param envs a string array with environments settings, i.e. strings of the form "VAR=value".
   * @return The newly created org.mathweb.core.CommandThread.
   */
  public CommandThread runThreaded(String command, String[] envs){
    CommandThread t = new CommandThread(command, envs);
    t.start();
    return t;
  }

  public CommandThread startThreaded(String command, String[] envs, int nice, int maxCPU, boolean wait){
    CommandThread t = new CommandThread(command, envs, nice, maxCPU, wait);
    t.start();
    return t;
  }

  /** 
   * runs a shell command with a CPU time limit.
   * @param command the command string.
   * @param envs a string array with environments settings, i.e. strings of the form "VAR=value".
   * @param nice the nice value for the command, i.e. an integer between -20 (highest priority) to 19 (lowest). However, the nice value is currently ignored.
   * @param maxCPU The CPU time limit for the command in seconds.
   */
  protected void start(String command, String[] envs, int nice, int maxCPU) {
    String new_command;
    boolean result;
    
    if (maxCPU==0) {
      new_command = command;
    }
    else {
      //      new_command = cpuLimitedRun + " " + maxCPU + " " + command;
      new_command = cpuLimitedRun + " " + maxCPU + " " + maxCPU + " " + command;
    }
    String[] commands = {"/bin/sh", "-c", new_command};
    //    System.out.println("executing: "+ new_command);
    storeTime("start1");
    result = exec(commands, envs, false, true);
    //    System.out.println("Result: " + result);
    storeTime("end1");
  }
  

  /** 
   * runs a shell command.
   * @param commands a string array containing all the different parts of the command, ie. {"/bin/sh", "-c", "echo foo"}
   * @param envs a string array with environments settings, i.e. strings of the form "VAR=value".
   * @param printResults if true, the output of the command execution is saved in a buffer. However the buffer is currently not returned.
   * @param wait if true, the method waits until the execution of the command has terminated.
   * @return true, if everything worked fine, false otherwise.
   */
  public static boolean exec(String[] commands, String[] envs, boolean printResults, boolean wait) 
  {
    Process p = null;
    try {
      // Start running command, returning immediately.
      p = java.lang.Runtime.getRuntime().exec(commands,envs);
            
      // Print the output. Since we read until
      // there is no more input, this causes us
      // to wait until the process is completed
      if(printResults) {
        BufferedInputStream buffer = new BufferedInputStream(p.getInputStream());
        BufferedReader  commandResult = new BufferedReader(new InputStreamReader(buffer));
        String s = null;
        try {
          while ((s = commandResult.readLine()) != null)
            //            System.out.println("Output: " + s);
          commandResult.close();
          if (p.exitValue() != 0) {
            if (verbose)
              printError(commands + " -- p.exitValue() != 0");
            return(false);
          }
          // Ignore read errors; they mean process is done
        } catch (Exception e) {}
                
        // If you don't print the results, then you
        // need to call waitFor to stop until the process
        // is completed
      } else if (wait) {
        try {
          //          System.out.println(" command completed ");
          int returnVal = p.waitFor();
          if (returnVal != 0) {
            if (verbose)
              printError(commands);
            return(false);
          }
        } catch (Exception e) {
          e.printStackTrace();
          if (verbose)
            printError(commands, e);
          return(false);
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
      if (verbose)
        printError(commands, e);
      return(false);
    } finally{
      if (p != null) {
        try{ InputStream is = p.getErrorStream();
        //        System.out.println(" exec() error stream " + is);
        is.close();
        } catch (Exception e){      e.printStackTrace();
 }
        try{ InputStream is = p.getInputStream();
        //        System.out.println(" exec() input stream " + is);
        is.close();
        } catch (Exception e){      e.printStackTrace();
 }
        try{ OutputStream os = p.getOutputStream();
        //        System.out.println(" exec() output stream " + os);
        os.close();
        } catch (Exception e){      e.printStackTrace();
 }
        p.destroy();
      }
    }
    return(true);
  }
  
  /** 
   * Prints an error message.
   * @param command  the command that produced the error.
   * @param e th Exception raised.
   */
  protected static void printError(String command,
                                   Exception e) {
    System.out.println("Error doing exec(" +
                       command + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  /** 
   * Prints an error message.
   * @param commands  the command array that produced the error.
   * @param e th Exception raised.
   */
  protected static void printError(String[] commands,
                                   Exception e) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  /** 
   * Prints an error message.
   * @param commands  the command array that produced the error.
   */
  protected static void printError(String[] commands) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + ").");
    System.out.println("Return value not 0 ");
  }

  /** 
   * Prints an error message.
   * @param command  the command that produced the error.
   */
  protected static void printError(String command) {
    System.out.println("Error doing exec(" +
                       command + ").");
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
    
  /** 
   * Prints an error message.
   * @param command  the command that produced the error.
   * @param value the return value of the command.
   */
  protected static void printError(String command, int value) {
    System.out.println("Error executing '" +
                       command + " " + value + "'.");
  }

  public void startup()
  {
    return;
  }

  public void shutdown()
  {
    return;
  }
  
  /** 
   * Suspends the current thread for a given time.
   * @param millis the number of milliseconds to wait. 
   */
  public static void waitMillis(long millis)
  {
    long t1 = System.currentTimeMillis();
    while ((System.currentTimeMillis() - t1) < millis);
  }
  
}

