// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
 http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.dp;

import java.io.*;
import java.util.*;
import java.net.URI;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;

import org.mathweb.atp.TimeoutException;
import org.mathweb.util.SmtCreator;


/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2006/08/23 10:45:55 $ $Revision: 1.1 $
*/
public class DecProc extends BasicService {

  /* The default timeout for decision procedures */
  private int defaultTime = 60;
    
  protected OWLJDom owlJDom;
  protected OntoReasoner reasoner;
  protected String name;

  private static Hashtable instances = new Hashtable();
  private static SmtCreator smtCreator = new SmtCreator();

  private Random random = new Random(System.currentTimeMillis());
  
  /* The MathServ ontology class that is accepted by the service as an input */
  public String acceptedConcept = "SmtProblem";
  
  /* The MathServ ontology class of the OWL individual that is returned by the service */
  public String resultConcept = "DecProcResult";
  
  /* The ID of the calculus individual in the MathServ ontology representing the calculus of the decider. */
  public String methodConcept = "Nelson-Oppen";

  private static Logger logger = Logger.getLogger("DecProc");
  private static String logFileName = getSystemProperty("atp.log.file", "DecProc_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);


  
  protected static int maxInputLength = 100000;
  protected static int maxOutputLength = 1000000;
  protected static int maxProofObjLength = 80000;
  
  public DecProc()
  {
    super();
    owlJDom = OWLJDom.getInstance();
    reasoner = PellReasoner.getNewReasoner();
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    Enumeration enu = logger.getAllAppenders();
//     for( ; enum.hasMoreElements() ;) {
//       System.out.println("appender: " +enum.nextElement());
//     }
    if (!enu.hasMoreElements()) logger.addAppender(appender);
  }

  
  /** 
   * @param problemFile a file containing an SMT decision problem.
   * @param language the language the proving problem is in. Should be SMT.
   * @param time the time limit for the theorem decision procedure in seconds.
   * @param options additional command line options for the decision procedure command if needed.
   * @return A String array of size 3 where the first entry is the status determined by the decision procedure. 
   *  The second entry can contain the complete output of the decision procedure.
   */
  public String[] callDecProc(File problemFile, String language, int time, String options) throws Exception
  {
    return new String[] {"Failure", null, "Please override method DecProc.callDecProc."};
  }

  /** 
   * @param decisionProblem a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param timeRes the time resource given for the problem, i.e. an individual of the OWL class TimeResource as an XML string.
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
  public String decideProblem(String decisionProblem, String timeRes) {
    return decideProblemWithOptions(decisionProblem, timeRes, "");
  }

  /** 
   * @param tptpProblem a first-order proving problem in TPTP format.
   * @param timeRes the time resource given for the problem
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
  public String decideSmtroblem(String smtProblem, Integer timeResInt) {
    return decideSmtProblemWithOptions(smtProblem, timeResInt, "");
  }
  
  /** 
   * @param tptpProblem a first-order proving problem in Smt format.
   * @param timeRes the time resource given for the problem.
   * @param options additional command line arguments for the underlying decision procedure. 
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
  public String decideSmtProblemWithOptions(String smtProblem, Integer timeResInt, String options)
  {
    int timeRes = timeResInt.intValue();
    String timeInd = owlJDom.createTimeIndi(timeRes, timeRes);
    String probID = "SmtProblem_" + System.currentTimeMillis();
    String owlProblem = smtCreator.smtString2OWLProblem(smtProblem, probID);
    System.out.println("owl: " + owlProblem);
    
    if (owlProblem != null) 
      return decideProblemWithOptions(owlProblem, timeInd, options);
    else 
      return owlJDom.createFailure("Could not transform Smt problem into an OWL individual of class SmtProblem!");
  }
  
  /** 
   * @param omdocProblem a first-order proving problem in OMDoc format using hol CDs.
   * @param options additional command line arguments for the underlying decision procedure. 
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
//   public String decideOmdocProblem(String omdocProblem, Integer timeResInt)
//   {
//     return decideOmdocProblemWithOptions(omdocProblem, timeResInt, "");
//   }

  /** 
   * @param omdocProblem a first-order proving problem in OMDoc format using hol CDs.
   * @param timeRes the time resource given for the problem.
   * @param options additional command line arguments for the underlying decision procedure. 
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
//   public String decideOmdocProblemWithOptions(String omdocProblem, Integer timeResInt, String options)
//   {

//     String smtProblem = x2tstp.omdoc2TSTP(omdocProblem);
//     System.out.println("Smt: "+ smtProblem);
//     return decideSmtProblemWithOptions(smtProblem, timeResInt, options);
//   }

  /** 
   * @param decisionProblem a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param timeRes the time resource given for the problem, i.e. an individual of the OWL class TimeResource as an XML string.
   * @param options additional command line options for the decision procedure. The method assumes that the decision procedure can understand the options given.
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class.
   */
  public String decideProblemWithOptions(String decisionProblem, String timeRes, String options) {
    /*  initialise local variables */
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    String probID = null;
    boolean valid = false; 
    int time = 0;
    File problemFile = null;
    
    //     System.out.println(decisionProblem);
    
    try{
      /*  parse the XML document and get the JDOM root elements */
      Element probRoot = null;
      
      /*  If the file is too big, use the modified SAX parser: */
      if ((decisionProblem != null) && (decisionProblem.length() > maxInputLength)){
        //      System.out.println("monster problem " + probID);
        //        return owlJDom.createFailure(probID, startTime, cpuStart, "Problem (of size " + decisionProblem.length() + " bytes) is too big to be handled by MathServe at the moment.");
        Document probDoc = owlJDom.parseLargeDocument(decisionProblem);
        StringWriter writer = new StringWriter();
        new XMLOutputter().output(probDoc,writer);
        /* The new XML document with huge problem descriptions written in a file */
        decisionProblem = writer.toString(); 
        System.out.println(writer.toString());
        probRoot = probDoc.getRootElement();
      } else {
        probRoot = owlJDom.getRootElement(decisionProblem);
      }
      
      String xmlBase = probRoot.getAttributeValue("base", Namespace.XML_NAMESPACE);
      Element timeRoot = owlJDom.getRootElement(timeRes);
      
      logger.info("... parsing, took me " + (System.currentTimeMillis() - startTime) + " msecs");
      logger.debug("[atp] : " + probRoot + " " + owlJDom);
      //        List cs = probRoot.getChildren();
      //        for(int i=0; i<cs.size();i++) logger.debug(cs.get(i));

      /*  look for potential rdf:IDs of the individual we are looking for. */
      List children = probRoot.getChildren();
      Element rightChild = null;
      for(int i=0; i<children.size(); i++){
        Element c = (Element) children.get(i);
        String potentialID = owlJDom.valueOf(owlJDom.RDF_ID, c);
        if (potentialID != null) {
          probID = xmlBase + potentialID;
          rightChild = c;
          break;
        }
      }
      
      if (probID == null) 
        return owlJDom.createFailure(xmlBase, startTime, cpuStart, "Problem has to be of class " + acceptedConcept + "!");
      logger.info("[DecProc] prob ID: " + probID);
      
      /*  tell the individual to the ontology reasoner to be able to infer further properties. */
      long indiTime = System.currentTimeMillis();
      valid = reasoner.tellIndividual(decisionProblem);
      logger.info("...telling individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
      //        startTime = System.currentTimeMillis();
      //        probID = reasoner.nameOfLastIndividual(acceptedConcept);
      //        logger.info("...looking for ID took me " + (System.currentTimeMillis() - startTime) + " msecs");
       
      //valid = reasoner.isValid();
      //       valid = true;
      //       logger.debug("valid OWL?: " + valid);
      //       logger.debug("right class?: " +reasoner.hasClass(probID,this.acceptedConcept) );
       
      // has the individual got the right type?
      if (valid){
        URI probURI = URI.create(probID);
        /*  get the  formal description of the proving problem. */
        Element formal = (Element) OWLJDom.PROVING_PROBLEM_FORMAL_DESCRIPTIONS.selectSingleNode(rightChild);
        logger.info("formal: " + formal);
        Attribute resource = formal.getAttribute("resource", OWLJDom.RDF_NS);
        if (resource != null){
          problemFile = new File(new URI(resource.getValue()));
          logger.info("formal file: " + problemFile.getAbsolutePath());
        } else {
          String content = formal.getText();
          logger.info("formal: " + content);
          problemFile = File.createTempFile("smt", ""); 
          logger.info("file: " + problemFile.getAbsolutePath());
          writeFile(problemFile, content);
        }
                  
        /*  big problems have been written to a file whose name is in the href attribute. */
        //        String problem = reasoner.valueOf("formalDescription", probURI);
        //        logger.info("[DecProc] formal: " + problem);
        
        /*  ask the reasoner for the language the proving problem is written in. */
        String language = reasoner.valueOf("language", probURI);
        logger.info("[DecProc] lang: " + language);


        time = 0;
        Double num = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(timeRoot);
        if (!num.isNaN()) 
          time = num.intValue();

        /* if no time given, try to extract it from the proving problem */
        if (time == 0) { 
          String timeIndi = reasoner.valueOf("timeRes", probURI);
          logger.debug(timeIndi);
          if (timeIndi == null) 
            time = this.defaultTime;
          else {
            String timeStr = reasoner.valueOf("wallClockTime", timeIndi);
            time =  Integer.parseInt(timeStr);
          }
        } 
        logger.info("[DecProc] time: " + time);
        
        String logic = reasoner.valueOf("inLogic", probURI);
        logger.info("[DecProc] in logic: " + logic);
         
        /*-----------------------------------------------------------*/
        /*  Call the DecProc specific method in the subclass of DecProc.java */
        /*-----------------------------------------------------------*/
        String[] result = this.callDecProc(problemFile, language, time,  options);
        //         System.out.println(result);
         
        //TODO: extract result status
        String status = result[0];
        
        if (status == null) status = "Unknown";
        logger.info("[DecProc] status : " + status);
        
        if ((status.equals("Unknown")) && 
            checkForTimeout(result[1], time))
          status="Timeout";
        
	//logger.info("[atp] result: " + result);
        
        /*  Call the final result, i.e. an OWL individual with all the result info. */
        return createResult(probID, status, result[1], startTime, cpuStart);
	
        
      } else {
        return owlJDom.createFailure(probID, startTime, cpuStart, "Problem has to be valid RDF description of an individual of OWL class " +  acceptedConcept + "!");
      }
    } catch(TimeoutException ex){
      return createResult(probID, "Timeout", null, startTime, cpuStart);
    } catch(Exception ex){
      ex.printStackTrace();
      return owlJDom.createFailure(probID, startTime,  cpuStart,"Error: "+ex.getMessage());
    } finally {
      if (problemFile != null) problemFile.delete();
      if ((valid)  && (probID !=null)){
        long indiTime = System.currentTimeMillis();
        reasoner.untellIndividual(probID);
        logger.info("deleting individual took me " + (System.currentTimeMillis() - indiTime) + " msecs");
        System.gc();
      }
    }
  }
  
  /** 
   * @param result the output of a  decision procedure.
   * @return The status in the <a href="http://www.mathweb.org/owl/status.owl">SZS Status Ontology</a> 
   *        which unambiguously indentifies the result of the decision procedure.
   */
  protected String extractStatusFrom(String result){
    String[] lines = result.split("\n");
    for(int i=0; i<lines.length; i++){
      if (lines[i].length()>15){
        //        System.out.println(lines[i].substring(0,15));
        if (lines[i].substring(0,15).equals("% Result     : ")){
          String rest = lines[i].substring(15).trim();
          String[] split = rest.split(" ");
          if (split.length > 1)
            return split[0];
          else 
            return rest;
        }
      }
    }
    return null;
  }
  
  /** 
   * Parses the line  "FINAL WATCH: 20.0 CPU 21.0 WC" output by TreeLimitedRun. 
   * @param output the output of a decision procedure call with TreeLimitedRun.
   * @param givenTime the time in seconds given to TreeLimitedRun.
   * @return Returns true, if the decision procedure timed out, false otherwise.
   */
  protected boolean checkForTimeout(String output, int givenTime)
  {
    float timeFloat = (float) givenTime;
    String[] lines = output.split("\n");
    for(int i=0; i<lines.length; i++){
      if ((lines[i].length()>12) && (lines[i].startsWith("FINAL WATCH:"))) {
        String[] lineParts = lines[i].split(" ");
        try { 
          Float cpuFloat = new Float(lineParts[2].trim());
          Float wcFloat = new Float(lineParts[4].trim());
          //          System.out.println("cpu : " + cpuFloat);
          float diff = Math.abs(cpuFloat.floatValue() - timeFloat);
          //          System.out.println("diff : " + diff);
          
          if (diff <= 0.2)
            return true;
          else 
            return false;
        } catch(NumberFormatException e){
          e.printStackTrace();
        }
      }
    }
    return false;
  }
  
  /** 
   * @param probID  a proving problem in RDF, i.e. an individual of the OWL class ProvingProblem as an XML string.
   * @param stat a valid status from the  <a href="http://www.mathweb.org/owl/status.owl">SZS Status Ontology</a> which unambiguously indentifies the result of the decision procedure.
   * @param startTime the start wall-clock time measured at the beginning of the decision procedure call.
   * @param cpuStart the start CPU time measured at the beginning of the decision procedure call.
   * @param output the formal (proof) output of the theorem decision procedure.  
   * @param xml a boolean indicating whether the output is an XML string or not. 
   * @return an individual of the OWL class specified in the private string <i>resultConcept</i> of the class. 
   */
  public String createResult(String probID, String stat, String output, long startTime, long cpuStart) {
    try{
      /*  create the XML root element rdf:RDF */
      Element rdf = new Element("RDF", OWLJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      rdf.setAttribute("base", OWLJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);

      /*  create the FoAtpResult element */
      Element dec_proc_result = new Element(resultConcept,OWLJDom.MW_NS);
      //      String resultID = OWLJDom.MW_NS_STR+name+ "_" +System.currentTimeMillis();
      String resultID = name + "_"  + System.currentTimeMillis();
      dec_proc_result.setAttribute("ID", resultID, OWLJDom.RDF_NS);
      Element result_for = new Element("resultFor",  OWLJDom.MW_NS);
      result_for.setAttribute("resource", probID, OWLJDom.RDF_NS);
      dec_proc_result.addContent(result_for);
      Element status = new Element("status",OWLJDom.MW_NS);
      status.setAttribute("resource", OWLJDom.STAT_NS_STR+stat, OWLJDom.RDF_NS);
      dec_proc_result.addContent(status);
      Element system = new Element("system",OWLJDom.MW_NS);
      system.addContent(name);
      dec_proc_result.addContent(system);
      
      
      /*  Add the TimeResource telling the cpu and wc time used up */
      Element time = new Element("time",  OWLJDom.MW_NS);
      Element timeres = OWLJDom.createTimeResource(startTime, cpuStart);
      time.addContent(timeres);
      dec_proc_result.addContent(time);
      
      if (output != null) {
        Element outputEl = new Element("output",  OWLJDom.MW_NS);
        outputEl.addContent(output);
        dec_proc_result.addContent(outputEl);
      }
      
      rdf.addContent(dec_proc_result); 
      Element newTime = (Element) timeres.clone();
      rdf.addContent(newTime); 
      return owlJDom.newXMLDocument(rdf);
    }
    catch(Exception e){
      e.printStackTrace();
      return owlJDom.createFailure(probID, startTime, cpuStart, e.getMessage());
    }
  }
  
  
    public void shutdown()
    {
      return;
    }
    
} // end DecProc.java


      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

