// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.dp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.SmtCreator;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/08/23 10:45:55 $ $Revision: 1.1 $
*/ 
public class Yices extends DecProc {
  
  public String YICES_HOME = getSystemProperty("yices.home", "YICES_HOME");
  public String[] Yices_envs = new String[] {"YICES_HOME="+YICES_HOME};
  public String YicesExecutable = findSystemExecutable(YICES_HOME, "yices", "Yices.exe");
  private static String axiomsUsed = "Formulae used in the proof :";
  
  private static String sep = File.separator;
  private static String YicesOptions = "";
    
  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Yices");
  private static String logFileName = getSystemProperty("yices.log.file", "YICES_LOG_FILE"); 
  private static FileAppender appender = newAppender(logFileName);
  
  public Yices()
  {
    super();
    name = "Yices_1_0";
    acceptedConcept = "SmtProblem";
    resultConcept = "DecProcResult";
    methodConcept = "Nelson-Oppen";
    
    
    logger.setLevel(Level.ALL);
    //logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  /** 
   * @param problemFile a file containing an SMT decision problem.
   * @param language the language the proving problem is in. Should be SMT.
   * @param time the time limit for the theorem decision procedure in seconds.
   * @param options additional command line options for the decision procedure command if needed.
   * @return A String array of size 3 where the first entry is the status determined by the decision procedure. 
   *  The second entry can contain the complete output of the decision procedure.
   */
  public String[] callDecProc(File problemFile, String language, int time, String options) throws Exception
  {
    File outputFile = null;    
    try{
      outputFile = File.createTempFile("Yices", ".out");
      
      if (language.equals(OWLJDom.MW_NS_STR+"SMT")) {
        
        String smt = readFile(problemFile);
        System.out.println(smt);
//        System.out.println("path " + problemFile.getPath());
        
        String command = YicesExecutable + " " + options + " " +  YicesOptions  + " " +
          problemFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[YICES] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[YICES] starting prover...");
        this.start(command, Yices_envs, defaultNice, time);
        //      this.exec(commands, Yices_envs, false, true);
        logger.debug("reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > DecProc.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
        
        String output = this.readFile(outputFile);
        System.out.println(output);
        
        String status; 
	if (output.indexOf("unsat") > -1)  status = "Unsatisfiable";
	else if (output.indexOf("sat") > -1)  status = "Satisfiable";
	else status = "Unknown";
	
        return new String[] {status, output};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Yices] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Yices] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Yices e = new Yices();
      OWLJDom owlJDom = OWLJDom.getInstance();
      SmtCreator creator = new SmtCreator();
      String timeInd = owlJDom.createTimeIndi(20,20);
      {
        String servF1;
        String tptpProb;
        String result;
        String ind; 
	ind = creator.smt2OWLProblem(msHome + "/thirdParty/SMT/QF_UF/NEQ/NEQ016_size5.smt");
        
	result = e.decideProblem(ind, timeInd);
	System.out.println("result: " + result);
	
      }
  }
  
  

} // end Yices.java



/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
