// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.owls;

import java.io.*;
import java.net.URI;
import java.net.URL;
import java.util.*;


import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mindswap.owl.OWLType;
import org.mindswap.owl.OWLCache;
import org.mindswap.owl.OWLReader;
import org.mindswap.owl.OWLOntology;
//import impl.owl.OWLCacheImpl;
import org.mindswap.owl.OWLErrorHandler;
import org.mindswap.owl.OWLDataProperty;
import org.mindswap.owl.OWLDataValue;
import org.mindswap.owls.OWLSFactory;

import org.mindswap.owls.OWLSListFactory;
import org.mindswap.owl.OWLFactory;
import org.mindswap.query.ValueMap;
import org.mindswap.owl.OWLKnowledgeBase;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.grounding.AtomicGrounding;
import org.mindswap.owls.grounding.WSDLAtomicGrounding;
import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.ParameterList;
import org.mindswap.owls.process.InputList;
import org.mindswap.owls.process.OutputList;
import org.mindswap.owls.process.ConditionList;
import org.mindswap.owls.process.ResultList;
import org.mindswap.owls.process.Result;
import org.mindswap.owls.process.Parameter;
import org.mindswap.owls.process.Input;
import org.mindswap.owls.process.Output;
import org.mindswap.owls.process.Condition;
import org.mindswap.owls.profile.Profile;
import org.mindswap.owls.profile.ServiceCategory;
import org.mindswap.owls.profile.ServiceParameter;
import org.mindswap.owls.service.Service;
import org.mindswap.owls.process.execution.ProcessExecutionEngine;

import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.util.ModelLoader;

import org.mathweb.core.*;
import org.mathweb.onto.*;

import org.mathweb.util.Java2WSDL;
import org.mathweb.util.TSTPCreator;
import org.mathweb.reasoning.logic.*;
import org.mathweb.serv.TaxonomyEntry;
import org.mathweb.serv.ReasoningService;
import org.mathweb.serv.PreCondition;
import org.mathweb.serv.Effect;
import org.mathweb.serv.StochasticEffect;
import org.mathweb.serv.ReasoningProblem;
import org.mathweb.serv.Query;

public class OWLSParser extends BasicService
{
  public static final Namespace RDF_NS = Namespace.getNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
  public static final Namespace RDFS_NS = Namespace.getNamespace("rdfs","http://www.w3.org/2000/01/rdf-schema#");
  public static final Namespace OM_NS = Namespace.getNamespace("om","http://www.openmath.org/OpenMath");
  public static final Namespace OMDOC_NS = Namespace.getNamespace("omdoc","http://www.openmath.org/omdoc");
  public static final Namespace OWL_NS = Namespace.getNamespace("owl","http://www.w3.org/2002/07/owl#");
  public static final Namespace OWLX_NS = Namespace.getNamespace("owlx","http://www.w3.org/2003/05/owl-xml#");
  public static final Namespace RULEML_NS = Namespace.getNamespace("ruleml","http://www.ruleml.org/inspec#");
  public static final Namespace SWRL_NS = Namespace.getNamespace("swrl","http://www.daml.org/services/owl-s/1.1/generic/swrlx.owl#");
  public static final Namespace WSDL_NS = Namespace.getNamespace("wsdl","http://schemas.xmlsoap.org/wsdl/");
  public static final Namespace WSDL_SOAP_NS = Namespace.getNamespace("wsdlsoap","http://schemas.xmlsoap.org/wsdl/soap/");
  public static final Namespace PROFILE_NS = Namespace.getNamespace("profile" , "http://www.daml.org/services/owl-s/1.1/Profile.owl#");
  public static final Namespace PROCESS_NS = Namespace.getNamespace("process" , "http://www.daml.org/services/owl-s/1.1/Process.owl#");
  
  /**   service descriptions  */

  public static final XPath serviceLocation = newXPath("./wsdl:service/wsdl:port/wsdlsoap:address/@location");
  public static final XPath wsdlName = newXPath("./wsdl:service/@name");
  //  public static final XPath serviceOperation = newXPath("./wsdl:operation/@name");
  

  /**   queries  */
  public static final XPath queryidx = newXPath("./process:AtomicProcess/@rdf:ID");
  public static final XPath queryProfileidx = newXPath("./profile:Profile/@rdf:ID");

  /**     I/O parameters    */
  public static final XPath namex = newXPath("./@name");
  public static final XPath contentsx = newXPath("./rdf:RDF");

  public static final XPath atomClassx = newXPath("./owlx:Class/@owlx:name");
  public static final XPath atomVarx = newXPath("./ruleml:var");
  public static final XPath atomArg1x = newXPath("./swrl:argument1/@rdf:resource");
  public static final XPath atomArg2x = newXPath("./swrl:argument2/@rdf:resource");

  public static final XPath atomPropx = newXPath("./swrl:propertyPredicate/@rdf:resource");

  public static final XPath bodyx = newXPath("./swrl:body");
  public static final XPath headx = newXPath("./swrl:head/*");
  public static final XPath listFirstx = newXPath("./swrl:AtomList/swrl:first/*");
  public static final XPath listRestx = newXPath("./swrl:AtomList/swrl:rest/*");
  
  private ArrayList condNames;
  
  protected SAXBuilder builder;
  protected static boolean expandEntities = false;
  private static Logger logger = Logger.getLogger("OWLSParser");

// new for 1.1.0
  private OWLKnowledgeBase kb;
  OWLCache cache;
  
  private String owlsDir = getSystemProperty("mathserv.owls.dir", "MATHSERV_OWLS_DIR");
  private String wsdlFileDir = getSystemProperty("mathserv.wsdl.directory", "MATHSERV_WSDL_DIRECTORY");

  private static URI rdfID = URI.create("http://www.w3.org/1999/02/22-rdf-syntax-ns#ID");
  private static URI exprLang = URI.create("http://www.daml.org/services/owl-s/1.1/generic/Expression.owl#expressionLanguage");
  private static URI exprBody = URI.create("http://www.daml.org/services/owl-s/1.1/generic/Expression.owl#expressionBody");
  private static URI servNameURI = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceName");
  private static URI serviceParameter = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceParameter");
  private static URI sParameter = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#sParameter");
  private static URI onParameter = URI.create("http://www.mathweb.org/owl/mathweb.owl#onParameter");
  private static URI inClass = URI.create("http://www.mathweb.org/owl/mathweb.owl#inClass");
  private static URI serviceName = URI.create("http://www.daml.org/services/owl-s/1.1/Profile.owl#serviceName");
    
  private static URI paramValue = URI.create("http://www.daml.org/services/owl-s/1.1/Process.owl#parameterValue");
  
  private Hashtable currentParameters;
  private String  currentURI;
  
  private OWLJDom owlJDom = OWLJDom.getInstance();

  public OWLSParser()
  {
    builder = new SAXBuilder();
    builder.setExpandEntities(expandEntities);
    Format format = Format.getPrettyFormat();
    format.setLineSeparator("\n");
    
    OWLCache cache = new OWLCacheImpl();
    //    System.out.println("owls dir: " + owlsDir);
    // set directory that contains service.idx
    cache.setLocalCacheDirectory(owlsDir);
    cache.setForced(true);
    
// new for 1.1.0 
    kb = OWLFactory.createKB();
    //    kb = OWLSFactory.createKB();
    OWLReader reader = kb.getReader();
    reader.setCache(cache);
    //    reader.setErrorHandler(new Listener());

//     owlsReader = OWLFactory.createOWLSReader("1.1");
    
    logger.setLevel(Level.ALL);
  }
  
  public OWLKnowledgeBase getKB()
  {
    return kb;
  }
  
  private CachedHashCodeObject getLiteralForAppl(Element applE){
    String functor = applE.getAttributeValue("func");
    String[] split = functor.split("#");
    if (split.length > 1)
      functor = split[1];
    List argEs = applE.getChildren();
    if (argEs.size() == 2){
      Element argE0 = (Element) argEs.get(0);
      Element argE1 = (Element) argEs.get(1);
      Argument arg0, arg1;
      if (argE0.getName().equals("var"))
        arg0 = new Variable(argE0.getAttributeValue("name"));
      else 
        arg0 = new Constant(argE0.getAttributeValue("name"));
      if (argE1.getName().equals("var"))
        arg1 = new Variable(argE1.getAttributeValue("name"));
      else 
        arg1 = new Constant(argE1.getAttributeValue("name"));
      
      Argument[] args = {arg0, arg1};
      return new Literal(new Predicate(functor), args, true);
    } else return null;
  }
  

  public Query parseAbstractQuery(String queryXML)
  {
    Query query = new Query();
    ReasoningProblem queryProb = new ReasoningProblem();
    
    try{
      Element root = owlJDom.getRootElement(queryXML); 
      System.out.println("root: " + root.getName());
     
      if (root.getName().equals("query")){
        query.uri = root.getAttributeValue("id");
        query.name = query.uri.substring(7);
        List inputEs = root.getChildren("input");
        List outputEs = root.getChildren("output");
        List precondEs = root.getChildren("precondition");
        List effectEs = root.getChildren("effect");
        List inputs = new ArrayList();
        List outputs = new ArrayList();
        List preconds = new ArrayList();
        List effects = new ArrayList();
        Hashtable paramNames = new Hashtable();
        
        for(int i=0; i < inputEs.size(); i++) {
          Element inE = (Element) inputEs.get(i);
          System.out.println("in: " + inE);
          String inID = inE.getAttributeValue("id");
          List children = inE.getChildren();
          if (children.size() > 0){
            Element inIndi = (Element) children.get(0);
            String type = inIndi.getName();
            System.out.println("query in " + inID + " of type " + type);
            String contents = null;
            if (inIndi.getChildren().size() >0){
              Element newIndi = (Element) inIndi.clone();
              Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
              rdf.setAttribute("base", OWLJDom.MW_NS_STR, Namespace.XML_NAMESPACE);
              rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
              rdf.addContent(newIndi);
              contents = owlJDom.newXMLDocument(rdf);
            }
            paramNames.put(inID, inE);
            inputs.add(new org.mathweb.serv.Input(inID, type, contents));
          }
        }

        for(int i=0; i < outputEs.size(); i++) {
          Element outE = (Element) outputEs.get(i);
          String outID = outE.getAttributeValue("id");
          System.out.println("out: " + outID);
          List children = outE.getChildren();
          if (children.size() > 0){
            Element outIndi = (Element) children.get(0);
            String type = outIndi.getName();
            System.out.println("query in " + outID + " of type " + type);
            outputs.add(new org.mathweb.serv.Output(outID, type));
          }
          paramNames.put(outID, outE);
        }
        
        for(int i=0; i < precondEs.size(); i++) {
          Element precondE = (Element) precondEs.get(i);
          String preID = precondE.getAttributeValue("id");
          List applEs = precondE.getChildren("appl");
          for (int k=0; k< applEs.size(); k++){
            Element applE = (Element) applEs.get(k);
            CachedHashCodeObject lit = getLiteralForAppl(applE);
            preconds.add(new PreCondition(preID, lit));
            System.out.println("appl: " + applE);
          }
        }

        for(int i=0; i < effectEs.size(); i++) {
          Element effectE = (Element) effectEs.get(i);
          String preID = effectE.getAttributeValue("id");
          List applEs = effectE.getChildren("appl");
          for (int k=0; k< applEs.size(); k++){
            Element applE = (Element) applEs.get(k);
            CachedHashCodeObject lit = getLiteralForAppl(applE);
            effects.add(new Effect(preID, lit));
          }
        }
        
        queryProb.inputs = inputs;
        queryProb.outputs = outputs;
        queryProb.preconds = preconds;
        queryProb.effects = effects;
        queryProb.name = query.uri.substring(7) + "_Problem";
        query.problem = queryProb;
        
        return query;
      } else return null;
    } catch(Exception e){
      e.printStackTrace();
      System.out.println(e.getMessage());

      return null;
    }
  }

  public org.mathweb.serv.Query parseQueryFile(String inputFileName) {
    return parseQuery(readFile(inputFileName));
    
  }
  public org.mathweb.serv.Query parseQuery(String description) {
    org.mathweb.serv.Query query = new org.mathweb.serv.Query();

    File inputFile = null;
    try{ 
      inputFile = File.createTempFile("owls-query", ".owl");
      String inputFileName = inputFile.getAbsolutePath();
      URI uri =  URI.create("file://" + inputFileName);
      writeFile(inputFile, description);
      
//       Model model = ModelLoader.loadModel("file:" + inputFileName);
//        Element root = owlJDom.getRootElement(description);
//       String queryURI = root.getAttributeValue("base", Namespace.XML_NAMESPACE);
//       String queryID = queryidx.valueOf(root);
//       //      System.out.println("processuri: " + queryURI + "#" + queryID);
//       Resource processInfo = model.getResource(queryURI + "#" + queryID);
      //      System.out.println("processInfo: " + processInfo);
      
      Service owlserv = kb.readService(uri);
      Process process = owlserv.getProcess();
      URI procuri = owlserv.getURI();
      ////       String queryURI = procuri.getScheme() + procuri.getUserInfo() + procuri.getAuthority();
      String queryURI = "http://" + procuri.getAuthority() + procuri.getPath();
      //      System.out.println("process: " + queryURI); 

      query.setOwlsService(owlserv);
      
      currentParameters = new Hashtable();
      currentURI        = queryURI;

      ReasoningProblem prob = getProblem(process.getInputs(),
                                         process.getOutputs(),
                                         process.getConditions(),
                                         process.getResults(),
                                         process.getLocals(),
                                         new ArrayList(),
                                         process.toString());
      query.setProblem(prob);
      query.name = queryURI.toString().substring(7);
      query.setURI(queryURI);

      kb.unload(uri);
      return query;
    } catch (Exception e){
      System.err.println("Query could not be parsed!");
      e.printStackTrace();
      return null;
    } finally{
      if (inputFile != null) inputFile.delete();
    }
  }

  /** 
   * @param description a string containing an OWL-S description
   * @return A MathServe ReasoningService.
   */
  public ReasoningService parseService(String description) {
    return parseService(description, null);
  }
  
  /** 
   * @param description a string containing an OWL-S description
   * @param newWsdl, a new WSDL description of the underlying WS if necessary, 
   *              if newWsdl is null, the WSDL binding in the OWL-S description is used.
   * @return A MathServe ReasoningService.
   */
  public ReasoningService parseService(String description, String newWsdl) {
    File inputFile = null;
    try{
      inputFile = File.createTempFile("owls-service", ".owl");
      writeFile(inputFile, description);
      return parseServiceFile(inputFile.getAbsolutePath(), newWsdl);
    } catch(Exception e){
      logger.debug("[OWLS] Exception " + e);
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
      } catch(Exception e){ logger.debug("[OWLS] Exception " + e); }
    }
    return null;
  }

  /** 
   * @param descriptionFileName the name of a file containing an OWL-S description
   * @return A MathServe ReasoningService.
   */
  public ReasoningService parseServiceFile(String descriptionFileName) {
    return parseServiceFile(descriptionFileName, null);
  }
  
  /** 
   * @param descriptionFileName the name of a file containing an OWL-S description
   * @param newWsdl, a new WSDL description of the underlying WS if necessary, 
   *              if newWsdl is null, the WSDL binding in the OWL-S description is used.

   * @return A MathServe ReasoningService.
   */
  public ReasoningService parseServiceFile(String descriptionFileName, String newWsdl) {
    ReasoningService serv = new ReasoningService();
    Service owlserv = null;
    URI uri =  URI.create("file://" + descriptionFileName);
    
    try{
      //      System.out.println("filename " + new URI("file://" + descriptionFileName));
      //      Service owlserv = owlsReader.read(new URI("file://" + descriptionFileName));
      //      System.out.println("service " + owlserv);

      owlserv = kb.readService(uri);
      Process process = owlserv.getProcess();
      Profile profile = owlserv.getProfile();
      List profileParams = profile.getServiceParameters();

      //      URI servURI = owlserv.getFileURI();
      URI procuri = owlserv.getURI();
      String servURI = "http://" + procuri.getAuthority() + procuri.getPath();
      //      String servURI = owlserv.getURI().toString();
      
      
      serv.name = profile.getServiceName();
      //      serv.name = servURI.toString().substring(7);
      //      serv.name = profile.getProperty(nameProp).getLexicalValue();
//       OWLDataProperty nameProp = kb.getDataProperty(serviceName);
//       serv.name = profile.getProperty(nameProp).getLexicalValue();

      serv.setURI(servURI);

      
      serv.setOwlsService(owlserv);
      
      ServiceCategory servCat = profile.getCategory();
      TaxonomyEntry tax = new TaxonomyEntry(servCat.getTaxonomy(), servCat.getCode());
      serv.taxonomy = tax;
      
      currentParameters = new Hashtable();
      currentURI        = servURI;
      
      ReasoningProblem prob = getProblem(process.getInputs(),
                                         process.getOutputs(),
                                         process.getConditions(),
                                         process.getResults(),
                                         process.getLocals(),
                                         profileParams,
                                         process.toString());
      serv.problem = prob;
      
      //      System.out.println("service: " + serv);
      //      System.out.println("process: " + process.getLabel());
      
      if (newWsdl != null){
        String wsdlFileName = wsdlFileDir + sep + serv.name + ".xml";
        writeFile(wsdlFileName, newWsdl);
        Grounding gr = owlserv.getGrounding();
        List grs = gr.getAtomicGroundings();
        for(int i=0; i< grs.size(); i++){
          AtomicGrounding ag = (AtomicGrounding) grs.get(i);
          if (ag instanceof WSDLAtomicGrounding){
            WSDLAtomicGrounding wag = (WSDLAtomicGrounding) ag;
            wag.setWSDL(URI.create("file:" + wsdlFileName));
          }
        }
        serv.setWsdlXML(newWsdl);
      }

      //      kb.unload(uri);
      return serv;
    } catch(Exception e){
      System.out.println(" error " + e.getMessage());
      e.printStackTrace();
      return null;
    }
  }

  private ReasoningProblem getProblem(InputList owlins, 
                                      OutputList owlouts, 
                                      ConditionList owlpres, 
                                      ResultList owlress,
                                      ParameterList locals, 
                                      List profileParams, 
                                      String probName
                                      )      
  {
//     System.out.println("ins: " + owlins);
//     System.out.println("outs: " + owlouts);
//     System.out.println("preconds: " + owlpres);
//     System.out.println("results: " + owlress);
    //    owleffs,
    ReasoningProblem prob = new ReasoningProblem();
    
    ArrayList inputs = new ArrayList(owlins);
    ArrayList outputs = new ArrayList(owlouts);
    ArrayList preconds = new ArrayList(owlpres);
    ArrayList effects = new ArrayList();
    ArrayList stochEffects = new ArrayList();

    HashMap locs = new HashMap();
    // new for 1.1.0 
    ConditionList owleffs = OWLSListFactory.createConditionList();
    //    ConditionList owleffs = OWLSFactory.createConditionList();
    
    
    for(int i=0; i<owlress.size(); i++){
      Result res = owlress.resultAt(i);
      Object eff = res.getEffect();
      //      System.out.println("result effect: " + eff);
      owleffs.add(eff);
//       ConditionList clist = res.getEffects();
//       owleffs.addAll(clist);
    }
    
    for(int i=0; i<locals.size(); i++) {
      Parameter par = locals.parameterAt(i);
      String name = par.getLocalName();
      OWLType type = par.getParamType();
      String typeStr = type.getURI().getFragment();
      currentParameters.put(currentURI + name, typeStr);
      locs.put(name, typeStr);
    }
    
    for(int i=0; i<inputs.size(); i++){
      Input in = owlins.inputAt(i);
      String name = in.getLocalName();
      String contents = null; 
      OWLDataProperty valueProp = kb.getDataProperty(paramValue);
      //      System.out.println("valueProp: " + valueProp);
      if ((valueProp != null) && (in.hasProperty(valueProp))) {
        contents = in.getProperty(valueProp).getLexicalValue();
      }
      
      //System.out.println("contents: " + contents);
//       try{
//         Element cont = owlJDom.getRootElement(contents);
//         contents = owlJDom.newXMLDocument((Element) cont.clone());
//       } catch(Exception e){
//         e.printStackTrace();
//         contents = null;
//       }
//       System.out.println("contents: " + contents);
      
      OWLType type = in.getParamType();
      String typeStr = type.getURI().getFragment();
      org.mathweb.serv.Input newIn = new org.mathweb.serv.Input(name, typeStr, contents);
      currentParameters.put(currentURI + name, newIn);
      inputs.set(i, newIn);
      //      System.out.println(newIn);
    }

    /* TODO: THIS IS A BIG HACK!!! It simply puts all inputs with type TimeResource at the 
       end of the list. We have to fix that. */
    int endIndex = inputs.size()-1;
    for(int i=0; i<inputs.size(); i++){
      org.mathweb.serv.Input in = (org.mathweb.serv.Input) inputs.get(i);
      String typeStr = in.getType();
      if (typeStr.equals("TimeResource")){
        Object lastIn = inputs.get(endIndex);
        inputs.set(endIndex, in);
        inputs.set(i, lastIn);
        break;
      }
    }
    
    for(int i=0; i<outputs.size(); i++){
      Output out = owlouts.outputAt(i);
      String name = out.getLocalName();

      OWLType type = out.getParamType();
      String typeStr = type.getURI().getFragment();
      //      System.out.println("output: " + name + "::" + type);
      org.mathweb.serv.Output newOut = new org.mathweb.serv.Output(name, typeStr);
      currentParameters.put(currentURI + name, newOut);
      outputs.set(i, newOut);
    }
    
    condNames = new ArrayList();
    preconds = parseConds(owlpres);
    for(int i=0; i<preconds.size(); i++){
      CachedHashCodeObject s = (CachedHashCodeObject) preconds.get(i);
      //      System.out.println("pre: " + s);
      org.mathweb.serv.PreCondition p = new org.mathweb.serv.PreCondition(); 
      p.condition = s; 
      p.setName((String) condNames.get(i));
      preconds.set(i,p);
    }
    
    condNames = new ArrayList();
    effects = parseConds(owleffs);
    for(int i=0; i<effects.size(); i++){
      CachedHashCodeObject s = (CachedHashCodeObject) effects.get(i);
      //      System.out.println("eff: " + s);
      org.mathweb.serv.Effect p = new org.mathweb.serv.Effect(); 
      p.condition = s; 
      p.setName((String) condNames.get(i));
      effects.set(i,p);
    }

    
    if ((profileParams != null) && (profileParams.size() > 0))
      stochEffects = parseStochEffects(profileParams);
//     for(int k=0; k<stochEffects.size(); k++){
//       System.out.println("stochastic : " +stochEffects.get(k));
//     }
    
    prob.name = probName;
    prob.inputs = inputs;
    prob.outputs = outputs;
    prob.setPreconds(preconds);
    prob.setEffects(effects);
    prob.setStochEffects(stochEffects);
    prob.locals = locs;
    
    return prob;
  }

  public ArrayList parseStochEffects(List profileParams){
    //    System.out.println("prop: " + kb.getDataProperty(sParameter));
    OWLDataProperty sProp = kb.getDataProperty(sParameter);
    ArrayList result = new ArrayList();
    Iterator iter = profileParams.iterator();
    try{
      
      while(iter.hasNext()){
        ServiceParameter par = (ServiceParameter) iter.next();
        
        //        System.out.println("par " + par);
        //        System.out.println("has prop: " + par.hasProperty(sProp));
        OWLDataValue indi = par.getProperty(sProp);
        Element root =  null;
        if (indi != null){
          //        System.out.println(indi.getClass().getName());
          String perform = indi.getLexicalValue();
          root = owlJDom.getRootElement(perform);
          //          System.out.println("first document: " + root);
        }
        
        String file = null;
        if (root != null) 
          file = root.getAttributeValue("file", owlJDom.PERF_NS);
        
        if (file != null) {
          URL url = new URL(file);
          InputStream inStream = url.openStream();
          Reader reader =  new InputStreamReader(inStream);
          Document doc = builder.build(reader);
          //          System.out.println(root + " " + file + " " + doc); 
          root = doc.getRootElement();
        }
        if (root != null) {
          List performs = root.getChildren("Perform", owlJDom.PERF_NS);
          for(int i=0; i < performs.size(); i++){
            Element perf = (Element) performs.get(i);
            Element cond = (Element) perf.getChild("Condition",owlJDom.PERF_NS);
            Element eff = (Element) perf.getChild("Effect", owlJDom.PERF_NS);
            Element prob = (Element) perf.getChild("Probability", owlJDom.PERF_NS);
            Element cost = (Element) perf.getChild("AverageCost", owlJDom.PERF_NS);
            Element reward = (Element) perf.getChild("Reward", owlJDom.PERF_NS);
            //            System.out.println("parts: " + perf + " " + cond + " " +eff + " " +prob);
            
            Float probFloat = new Float(prob.getTextTrim());
            Integer costInt = new Integer(cost.getTextTrim());
            Integer rewardInt = new Integer(reward.getTextTrim());
            
            CachedHashCodeObject condAtom = parseCond((Element) cond.getChildren().get(0));
            org.mathweb.serv.PreCondition condObj = new org.mathweb.serv.PreCondition(); 
            condObj.condition = condAtom; 
            condObj.setName("foo");

            CachedHashCodeObject effAtom = parseCond((Element) eff.getChildren().get(0));          
            org.mathweb.serv.Effect effObj = new org.mathweb.serv.Effect(); 
            effObj.condition = effAtom; 
            effObj.setName("bar");
            
            StochasticEffect stochEff = new StochasticEffect(condObj, effObj, probFloat, costInt, rewardInt);
            result.add(stochEff);
          }
        } // if(root...
      }
      
    }catch(Exception e){
      e.printStackTrace();
    }
    return result;
  }
  

  /**  every pre-condition or post-condition becomes a Cached */
  public ArrayList parseConds(ConditionList pc)
  {
    ArrayList roots = new ArrayList();
    if (!pc.isEmpty())
      for(int i=0; i<pc.size(); i++){ 
        // pre-condition element
        try {
          Condition cond = pc.conditionAt(i);
          String expr = cond.getProperty(kb.getDataProperty(exprBody)).getLexicalValue();
          //          System.out.println("condition: " + expr + expr.getClass().getName());
          StringReader reader = new StringReader(expr);
          Document sDoc = builder.build(reader);
          Element root = sDoc.getRootElement();
          //          System.out.println("root: " + root);
          roots.add(root);
        }catch(Exception e){
          e.printStackTrace();
        }
      }
    return parseRDFConds(roots);
  }

//   private ReasoningProblem parseProblem(Element problem)      {
//     ReasoningProblem prob = new ReasoningProblem();
//     try{
//       prob.name = namex.valueOf(problem);
//       ArrayList inputs = new ArrayList(inputsx.selectNodes(problem));
//       ArrayList outputs = new ArrayList(outputsx.selectNodes(problem));
//       ArrayList preconds = new ArrayList(precondsx.selectNodes(problem));
//       ArrayList postconds = new ArrayList(postcondsx.selectNodes(problem));
//       System.out.println("problem: " +problem);
//       System.out.println("inputs: " +inputs);
      
//       String name = null;
//       String sig = null;
      
//       for(int i=0; i<inputs.size(); i++){
// 	Element e = (Element) inputs.get(i);
// 	name = namex.valueOf(e);
// 	sig = signaturex.valueOf(e);	
// 	// try to split the namespace and the OWL class name from signature
// 	String[] st= sig.split("#");
// 	String cn;
// 	//System.out.println("split: " + st.length + " " +st[0]  + " " +st[1] );
// 	if (st.length>1) cn = st[1];
// 	else cn = sig;
// 	// get the content of the input (in case we are in a query queries)
// 	Element cont = (Element) contentsx.selectSingleNode(e);
// 	System.out.println("input " + name+ "::" + cn+ "= " +cont);
// 	String doc;
// 	if (cont != null) {
// 	  doc = owlJDom.newXMLDocument((Element) cont.clone());
// 	} else doc = "";
// 	Input in = new Input(name, cn, doc);
// 	inputs.set(i,in);
//       }
      
//       for(int i=0; i<outputs.size(); i++){
// 	Element e = (Element) outputs.get(i);
// 	name = namex.valueOf(e);
// 	sig = signaturex.valueOf(e);
// 	String[] st= sig.split("#");
// 	String cn;
// 	if (st.length>1) cn = st[1];
// 	else cn = sig;
// 	Output in = new Output(name,cn);
// 	outputs.set(i,in);
//       }
//       preconds = parseConds(preconds);
//       for(int i=0; i<preconds.size(); i++){
// 	CachedHashCodeObject s = (CachedHashCodeObject) preconds.get(i);
// 	PreCondition p = new PreCondition(); 
// 	p.condition = s; 
// 	preconds.set(i,p);
//       }
//       postconds = parseConds(postconds);
//       for(int i=0; i<postconds.size(); i++){
// 	CachedHashCodeObject s = (CachedHashCodeObject) postconds.get(i);
// 	Effect p = new Effect(); 
// 	p.condition = s; 
// 	postconds.set(i,p);
//       }
//       prob.inputs=inputs;
//       prob.outputs=outputs;
//       prob.setPreconds(preconds);
//       prob.setEffects(postconds);
//       return prob;
//     } catch(Exception e){
//       e.printStackTrace();
//       return prob;
//     }
//   }
  

  /**  
       @pc a list of root elements, i.e. rdf:RDF elements
  */
  public ArrayList parseRDFConds(ArrayList pc)
  {
    ArrayList allConds = new ArrayList();
    
    if (!pc.isEmpty())
      for(int i=0; i<pc.size(); i++){ 
        // pre-condition element
        Element cond = (Element) pc.get(i);
        ArrayList cs = new ArrayList(cond.getChildren());
        for(int k=0; k<cs.size(); k++){
          cond = (Element) cs.get(k);
          condNames.add(cond.getAttributeValue("ID", RDF_NS));
          //          System.out.println("cond-root: " + cond + " " + cond.getName());
          allConds.add(parseCond(cond));
        }
      }
    return allConds;
  }
  
  public ArrayList parseAllConds(ArrayList pc)
  {
    for(int i=0; i<pc.size(); i++){ 
      Element cond = (Element) pc.get(i);
      pc.set(i, parseCond(cond));
    }
    return pc;
  }
  
  private CachedHashCodeObject parseCond(Element cond)  {
    if (cond.getName().equals("ClassAtom")){
      return parseClassAtom(cond);
    }
    else if (cond.getName().equals("IndividualPropertyAtom")){
      return parseIndProperty(cond);
    }
    else if (cond.getName().equals("SameIndividualAtom")){
      return parseSameInd(cond);
    }
    else if (cond.getName().equals("Assertion")){
      return parseCond((Element) (cond.getChildren().get(0)));
    }
    else if (cond.getName().equals("And")){
      return parseConjunct(cond);
    }
    else if (cond.getName().equals("Imp")){
      return parseImplication(cond);
    }
    else if (cond.getName().equals("Or")){
      return parseDisjunct(cond);
    }
    Argument[] args = {(new Constant("a")), (new Constant("b"))};
    return new Literal(new Predicate("Condition of type " + cond.getName() + " not known!"), args, true);
  }
  
  
  public CachedHashCodeObject parseConjunct(Element cond){
    ArrayList cs = new ArrayList(cond.getChildren());
    ArrayList literals = parseAllConds(cs);
    Conjunct con = new Conjunct(literals);
    return con;
  }

  public CachedHashCodeObject parseDisjunct(Element cond){
    ArrayList cs = new ArrayList(cond.getChildren());
    ArrayList literals = parseAllConds(cs);
    Disjunct con = new Disjunct(literals);
    return con;
  }
    
  public CachedHashCodeObject parseImplication(Element cond){
    try{
      
      Element body =  (Element) bodyx.selectSingleNode(cond);
      Element head =  (Element) headx.selectSingleNode(cond);
      //      System.out.println("body: " + body);
      //      System.out.println("head: " + head);
      ArrayList bodys = new ArrayList();
      boolean end = false;
      
      while (!end){
        Element first = (Element) listFirstx.selectSingleNode(body);
        //        System.out.println("first: " + first);
        bodys.add(first);
        Element rest =  (Element) listRestx.selectSingleNode(body);
        //        System.out.println("rest: " + rest);
        if (rest == null) end = true;
        else {
          body = rest;
        }
      }
      ArrayList literals = parseAllConds(bodys);
      //      System.out.println("lits: " + literals);
      Literal lit = (Literal) parseCond(head);
      
      return new Implication(lit ,literals);    
    }catch(Exception e){
      e.printStackTrace();
      return null;
    }
  }
  
    
  public CachedHashCodeObject parseClassAtom(Element cond){
    try{
      String cn = atomClassx.valueOf(cond);
      String vn = ((Element)atomVarx.selectSingleNode(cond)).getText();
      Variable var = new Variable(vn.toLowerCase());
    
      //      System.out.println("classAtom: " + cond + " " + cn);
      String[] st= cn.split("#");
      String name, ns;
      //      System.out.println("split: " + st.length + " " +st[0]  + " " +st[1] );
      if (st.length>1){  name = st[1]; ns=st[0];      }
      else {name = cn; ns = OWLJDom.MW_NS.getURI();  }

      Constant classSym = new Constant(name);
      Predicate type = new Predicate("type"); 
      //RDF_NS.getURI());
      Argument[] args = new Argument[2];
      args[0] = var;
      args[1] = classSym;
      return new Literal(type, args, true);
    }catch (org.jdom.JDOMException e) {
      return null;
    }
  }

  public CachedHashCodeObject parseSameInd(Element cond){
    try{
      Argument[] args = new Argument[2];
      
      String vn0 = atomArg1x.valueOf(cond);
      String par0 = vn0;
      String vn1 = atomArg2x.valueOf(cond);
      String par1 = vn1;
      //      System.out.println("args: " + vn0 + " " + vn1);
      
      String[] st = vn0.split("#");
      if (st.length>1) {
        par0 = st[1];
      }
      st= vn1.split("#");
      if (st.length>1) 
        par1 = st[1];
      
      if(currentParameters.containsKey(currentURI + par0)){
        args[0] = new Variable(par0);
      } else {
        args[0] = new Constant(vn0);
      }
      
      if(currentParameters.containsKey(currentURI + par1)){
        args[1] = new Variable(par1);
      } else {
        args[1] = new Constant(vn1);
      }
      
      Predicate equalP = new Predicate("equal"); 
      return new Literal(equalP, args, true);
    }catch (org.jdom.JDOMException e) {
      e.printStackTrace();
      return null;
    }
  }
  
  public CachedHashCodeObject parseIndProperty(Element cond){
    try{
      //      System.out.println("indProperty: " + cond);
//       System.out.println("propPred: " + propPred);
//       Element e = (Element) propPred;
//       System.out.println("propPred: " + propPred + " " + e.getAttribute("resource", RDF_NS)); 
//       List ats = e.getAttributes();
//       for(int i=0; i< ats.size(); i++){
// 	Attribute a = (Attribute) ats.get(i);
// 	System.out.println("att: " + ats.get(i) + a.getName() + " " + a.getNamespace());
//       }
      Argument[] args = new Argument[2];
      
      String prop = atomPropx.valueOf(cond);
      //      System.out.println("prop: " + prop);
      String[] st= prop.split("#");
      String name;
      if (st.length>1) name = st[1];
      else name = prop;
      Predicate propSym = new Predicate(name);

      String vn0 = atomArg1x.valueOf(cond);
      String par0 = vn0;
      String vn1 = atomArg2x.valueOf(cond);
      String par1 = vn1;
      //      System.out.println("args: " + vn0 + " " + vn1);
      
      st = vn0.split("#");
      if (st.length>1) 
        par0 = st[1];
      st= vn1.split("#");
      if (st.length>1) 
        par1 = st[1];
      
      /* service parameters and locals become variables whose name does not contain the service URI 
       */
      if(currentParameters.containsKey(currentURI + par0)){
        args[0] = new Variable(par0);
      } else 
        /* every thing else becomes a constant with the full URI as a name */
        args[0] = new Constant(vn0);
      
      if(currentParameters.containsKey(currentURI + par1)){
        args[1] = new Variable(par1);
      } else 
        args[1] = new Constant(vn1);
      
      //      System.out.println("args: " + args[0] + " " + args[1]);
      
      return new Literal(propSym, args, true);
    }catch (org.jdom.JDOMException e) {
      return null;
    }
  }
  
  public static XPath newXPath(String x)
  {
    try{
      XPath xpath =  XPath.newInstance(x);
      xpath.addNamespace(OWLJDom.MW_NS);
      xpath.addNamespace(RDF_NS);
      xpath.addNamespace(RDFS_NS);
      xpath.addNamespace(OMDOC_NS);
      xpath.addNamespace(OWL_NS);
      xpath.addNamespace(OWLX_NS);
      //      xpath.addNamespace(OM_NS);
      //      xpath.addNamespace(RULEML_NS);
      xpath.addNamespace(SWRL_NS);
      xpath.addNamespace(WSDL_NS);
      xpath.addNamespace(WSDL_SOAP_NS);
      xpath.addNamespace(PROFILE_NS);
      xpath.addNamespace(PROCESS_NS);
      return xpath;
    } catch(Exception e){
      return null;
    }
  }

  public static void main(String argv[]) throws Exception {
    OWLSParser c = new OWLSParser();
    String msHome = System.getProperty("mathserv.home");
    int counter = 1;
    List services = new ArrayList();
    
    //    Query q = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/nd-query1.owl");
       Query q = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/anyResult.owl");
    
    
//     services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/TrampNDService.owl"));
//     services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/TrampCNFService.owl"));
//    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/OtterService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/DctpService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/EService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/SpassService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/ParadoxService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/VampireService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/WaldmeisterService.owl"));
    services.add(c.parseServiceFile(msHome + "/etc/services/owls/1.1/TSTPAnalyserService.owl"));   
    
    // example with Java Grounding
//     s[8] = c.parseServiceFile(msHome + "/thirdParty/owl-s-svn/trunk/src/JGroundingTest.owl");   
//     Hashtable map = new Hashtable();
//     map.put("FirstParam", new Integer(100));
//     map.put("SecParam", new Double(100.00));
//     ValueMap resMap = s[8].invokeOWLS(map);
//     System.out.println("result: " + resMap);
    
    String domain = ReasoningService.computeDTGologDomain(services);    
    System.out.println(domain);
    
      //       System.out.println("action: " + s[i].toPlanningOperator().toLongString());
      //     }
    //      org.mathweb.serv.Query  query;
//      //     query = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/nd-query1.owl");
//      query = c.parseQueryFile(msHome + "/examples/queries/owls/1.1/result-query1.owl");
//    System.out.println("query " + q.toProdigyProblem());
    
  }

}

 
      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

