// -*- indent-tab-smode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;

import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.TrampThreadXmlRpcHandler;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/03/30 16:12:31 $ $Revision: 1.2 $
*/
public class TrampThread extends Thread {

  /** Lisp executables and environments */
  private static BasicService core = new BasicService();
  private String ACL_HOME = core.getSystemProperty("acl.home", "ACL_HOME");
  private String aclExecutable = core.findSystemExecutable(ACL_HOME, "lisp", "alisp");
  
  private String trampDump = core.getSystemProperty("tramp.dump", "TRAMP_DUMP");
  private String trampExecutable = core.getSystemProperty("tramp.exec", "TRAMP_EXEC");
  private String trampXtermExecutable = core.getSystemProperty("tramp.exec.xterm", "TRAMP_EXEC_XTERM");
  private String trampEmacs = core.getSystemProperty("tramp.emacs.exec", "TRAMP_EMACS_EXEC");
  private String trampIlisp = core.getSystemProperty("tramp.ilisp", "TRAMP_ILISP");
  private String DISPLAY = core.getSystemProperty("tramp.display", "TRAMP_DISPLAY");
  private String emacsGeometry = core.getSystemProperty("tramp.emacs.geometry", "TRAMP_EMACS_GEOMETRY");
  private String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME, "DISPLAY="+DISPLAY};


  private static Logger logger = Logger.getLogger("TrampThread");
  private static String logFileName = core.getSystemProperty("tramp.log.file", "TRAMP_LOG_FILE");

  private static String lispLogFileName = core.getSystemProperty("acl.log.file", "ACL_LOG_FILE");

  /** For communication with Lisp process */
  private static String sep = File.separator;
  private int serverPort = 0;
  private int defaultPort = 12345;
  private CommandThread lispThread;
  
  private WebServer xmlRpcServer;
  private String lispServerURL = "";
  private org.apache.xmlrpc.XmlRpcClient lispServerConnection;
  
  private static final int defaultNice = 15;
  private Throwable exception;
  private boolean started = false;
  private boolean ready = false;
  private int cnfCounter = 1;
  private int ndCounter = 1;
  private static int logfileCounter = 1;
  private static int lisplogfileCounter = 1;
  private String user = "foo";
  
  public TrampThread(String usr)
  {
    core.mkTempDir("tramp");
    //    owlJDom = OWLJDom.getInstance();
    DISPLAY = DISPLAY.replace('"', '\'');
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    FileAppender appender = BasicService.newAppender(logFileName + usr + (logfileCounter++) + ".log");
    logger.removeAllAppenders();
    logger.addAppender(appender);

    xmlRpcServer = null;
    lispServerConnection = null;
    started = false;
    ready = false;
    user = usr;
    //    startTramp();
  }
  
  public void run()
  {
    startTramp();
  }
  
  
  public String[] callTrampClausifier(String postProb) throws Exception {
    String cnfProb = null;
    File probFile = null;
    try{
      String tempDirPath = core.tempDir.getAbsolutePath();
      String probFileName = tempDirPath + sep + "problem" + cnfCounter + ".post";
      cnfCounter ++;
      probFile = new File(probFileName);
      core.writeFile(probFile, postProb);
      Vector params = new Vector();
      params.addElement(probFileName);
      
      callLispCommand("read-problem", params);
      params = new Vector();
      Object result = callLispCommand("clausify", params);
      System.out.println("result: " + result + result.getClass().getName());
      Vector resVec = (Vector) result;
      
      cnfProb = (String) resVec.get(0);
      String delta = (String) resVec.get(1);
      logger.debug("[TrampThread] result " + cnfProb); 
      //        logger.debug("[Tramp] delta relation: " + delta);
      return new String[] {cnfProb, delta};
    } catch (XmlRpcException e) {
      logger.debug("TrampThread error " + e.code + ":" + e.getMessage());
      throw new Exception("Tramp error (code " + e.code + "): " + e.getMessage());
    } finally{
      try{ 
        if (probFile != null) probFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
  }
  
  public String callTrampFof2ND(String postProblem, String tstpProof) throws Exception
  {
    File postFile = null;
    File tstpFile = null;
    File outputFile = null;
    try{
      String tempDirPath = core.tempDir.getAbsolutePath();
      String postFileName = tempDirPath + sep + "tramp"+ ndCounter + ".post";
      String tstpFileName = tempDirPath + sep + "tramptstp"+ ndCounter + ".tstp";
      String outputFileName = tempDirPath + sep + "tramp"+ ndCounter + ".twega"; 
      ndCounter++;
      postFile = new File(postFileName);
      tstpFile = new File(tstpFileName);
      outputFile = new File(outputFileName);
      core.writeFile(postFileName,postProblem);
      core.writeFile(tstpFileName,tstpProof);

      // first read in the POST problem:
      Vector params = new Vector();
      params.addElement(postFileName);
      callLispCommand("read-problem", params);

      // then call the proof transformation command:
      params = new Vector();
      params.addElement(tstpFileName);
      String nd = (String) callLispCommand("transform-tstp-proof-from-file", params);
      logger.debug("Tramp ND: " + nd);
      callLispCommand("show-pds", params);
      return nd;
    } catch (XmlRpcException e) {
      logger.debug("Tramp error " + e.code + ":" + e.getMessage());
      throw new Exception("Tramp error (code " + e.code + "): " + e.getMessage());
    } finally{
      try{ 
//         if (postFile != null) postFile.delete();
//         if (tstpFile != null) tstpFile.delete();
	if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
  }
  
  
  private void startXmlRpcServer() throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = defaultPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        logger.debug("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ( "$default", new TrampThreadXmlRpcHandler(this));
        success = true;
      }catch(Exception e){
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        logger.debug("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  public TrampThread startTrampEmacs(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String ilisp = core.readFile(trampIlisp);
        String tempDirPath = core.tempDir.getAbsolutePath();
        String inputFileName = tempDirPath + sep + "tramp.el";
        String emacsStartup = ilisp + "\n" +
          "(setq loui-program " + 
          "\"" + trampExecutable + " " + aclExecutable + " " + trampDump + " " + serverURL + "\") ";
        core.writeFile(inputFileName, emacsStartup);
        
        if (serverPort != 0) {
          String command2 = trampEmacs + " " +
            "-bg white -fg black -debug-init -display " + DISPLAY + " -name \"Tramp Service (Emacs)\" " +
            "-geometry " + emacsGeometry + " -font 8x13" +
            " -l " + inputFileName + " -f loui";
          logger.debug("tramp command: " + command2);
          lispThread = core.runThreaded(command2, acl_envs);
          logger.debug("after execution " + lispThread);
          started= true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Tramp:" +e);
        logger.fatal("Something went wrong when starting Tramp:" +e);
      }
    }
    return this;
  }

  public TrampThread startTrampXterm(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String tempDirPath = core.tempDir.getAbsolutePath();
        
        if (serverPort != 0) {
          //          String echoCommand =  "echo ${DISPLAY} > /home/jzimmer/display";
          //          lispThread = runThreaded(echoCommand, acl_envs);
          String command2 =  "\"" + trampXtermExecutable + " " + aclExecutable + " " + trampDump + " " + serverURL + " " + DISPLAY + "\"";
          System.out.println("tramp xterm command: " + command2);
          //          lispThread = runThreaded(command2, acl_envs);
                   String[] commands = {"/bin/sh", "-c", command2};
                   core.exec(commands, acl_envs, false, true);
          
          System.out.println("after execution " + lispThread);
          started = true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Tramp:" +e);
        logger.fatal("Something went wrong when starting Tramp:" +e);
      }
    }
    return this;
  }

  public TrampThread startTramp() {
      if (!started){
        try {
          startXmlRpcServer();
          String serverURL = "http://localhost:" + serverPort;
          String tempDirPath = core.tempDir.getAbsolutePath();
          String outputFileName = tempDirPath + sep + "tramp" + ".out";
          
          if (serverPort != 0) {
            String command2 = aclExecutable + " -I " + trampDump + 
              " -qq -e \"(progn (ags::rpc~start-server '(\\\"" + serverURL + 
              "\\\" \\\"initLispServer\\\" \\\""+ user + "\\\")) " +
              "(setf excl::*restart-app-function* (lambda () (omega::tramp~top))))\" > " + lispLogFileName + user + (lisplogfileCounter++) + ".log 2>&1";
            logger.debug("tramp command: " + command2);
            lispThread = core.runThreaded(command2, acl_envs);
          }
          started = true;
          return this;
        }
        catch(Exception e) {
          System.err.println("Something went wrong when starting Tramp:" +e);
          logger.fatal("Something went wrong when starting Tramp:" +e);
        }
      }
      return this;
  }
 
  public boolean isStarted() { return this.started; }
  public boolean isReady() { return this.ready; }
  public void setReady() { this.ready = true; }
  
  public void setLispServerConnection(String url)
  {
    try{
      lispServerURL = url;
      lispServerConnection = new org.apache.xmlrpc.XmlRpcClient(url);
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
  }
  
  public org.apache.xmlrpc.XmlRpcClient getLispServerConnection()
  {
    return lispServerConnection;
  }
  

  public Object callLispCommand(String command, Vector params) throws XmlRpcException
  {
    try{
      Object result = lispServerConnection.execute(command, params);
      //      logger.debug("Result: " + result + " " + result.getClass());
      return result;
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    } catch(IOException e) {
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
    return null;
  }
  
  /**  Shuts down the underlying lisp process as cleanly as possible. 
   */
  public void shutdown()
  {
    try { 
      Vector params = new Vector();
      if (lispServerConnection != null) {
        logger.debug("shutting down lisp process!");
        callLispCommand("shutdown", params);
	lispServerConnection = null;
        logger.debug("waiting 3 secs for Lisp process to terminate...");
        synchronized(this){
          this.wait(3000);
        }
        logger.debug("...done.");
        xmlRpcServer.removeHandler("$default");
        xmlRpcServer.shutdown();
        lispThread.shutdown();
        logger.debug("Lisp process terminated!");
      }
    } catch (XmlRpcException e) {
      logger.debug("Trouble with RPC call " + e);
      e.printStackTrace();
    } catch (InterruptedException e) {
      logger.debug("Trouble with RPC call " + e);
      e.printStackTrace();
    } finally {
      core.removeTempDir();
    }
  }
   
  public  void test() {
    while(! this.ready) {
      try {
	//      synchronized(this.ready){
	System.out.println("not ready");
	synchronized(this){
          this.wait(1000);
        }
      } catch(Exception ex){
	ex.printStackTrace();
      }
    }
    try{
      String postProb = core.readFile(BasicService.msHome + "/examples/post/agathas-murderer.post");
      String[] cnf;
      for (int i=0; i< 20; i++){
	cnf = callTrampClausifier(postProb);
	;;      logger.debug("[Tramp] generated cnf: " + cnf);
      }
      
      
//       nd = ndForFOF(fofProb, tstpProof);
//       logger.debug("[Tramp] generated ND proof: " + nd);
      
      //       Vector params = new Vector();
      //       params.addElement(probFile);
      //       callLispCommand("read-problem", params);
      //       params = new Vector();
      //       Object result = callLispCommand("clausify", params);
      //       String prob = (String) result;
      //        String delta = (String) result.get(1);
      //OtterService o = new OtterService();
      //String n = o.callProver(prob, "TSTP", 300, "");
      //      logger.debug("[Tramp] result " + prob); 
//       logger.debug("[Tramp] final result2: " + nd);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    TrampThread e = new TrampThread("test");
    //e.startTrampEmacs();
    //    e.startTrampXterm();
    e.startTramp();
    e.test();
    e.shutdown();
    System.exit(0);
  }
  
} // end Tramp.java

      
