(* ========================================================================= *)
(* Positions and related data structures                                     *)
(* ========================================================================= *)


(** Module Position implements term positons and related data structures.
    @author Frank
    @since 19-05-06*)


(** The type of relative positions. *)
type relpos = Abstr | Func | Arg


type position = relpos list
(** A term position is specified by a list of relative positions relpos. This list is
    interpreted as a path from the top term to the subterm at the given position, starting
    with the first list element at the top term.*)


(** A positiontable is a datastructure to record information related to term positions. It
    can be interpreted either top-down or bottom-up. Top-down interpretation is used to
    record occurrences of subterms (e.g. all free variables of a term or all occurrences
    of  bound variables with the position relative to their binder). Bottom-up
    interpretation is used to record occurrences of terms within superterms (e.g. all
    superterms in which a term occurs as a subterm at a given position).*)
type 'a positiontable =
    Empty
  | Table of 'a positiontable *
             'a positiontable *
             'a positiontable *
             'a * bool


(** {6 Positiontable construction} *)

val new_positiontable : unit -> 'a positiontable
(** Create a new positiontable *)


val insert_at_pos : 'a positiontable  -> position -> 'a -> 'a positiontable
(** Insert a new subterm element at a given position into a positiontable (top-down).
    @param table a referenced table
    @param pos the position
    @param v the value to be inserted*)


(** {6 Positiontable lookup} *)

val data_at_pos : 'a positiontable -> position -> 'a
(** Lookup the data at a given position *)



(*
val all_occurrences : 'a positiontable ref -> 'a list
(** Lookup all entries in a positiontable (top-down/ bottom-up)
    @param table a referenced table*)


val subterms_at : 'a positiontable ref -> position -> 'a list
(** Lookup all subterm entries {e at} a given position (top-down)
    @param table a referenced table
    @param pos the position *)


val subterms_below : 'a positiontable ref -> position -> 'a list
(** Lookup all subterm entries {e below} a given position (top-down)
    @param table a referenced table
    @param pos the position *)

    
val subterms_above : 'a positiontable ref -> position -> 'a list
(** Lookup all subterm entries {e above} a given position (top-down)
    @param table a referenced table
    @param pos the position *)


val insert_superterm_at : 'a positiontable ref -> position -> 'a -> unit
(** Insert a new superterm element at a given position into a positiontable (bottom-up).
    @param table a referenced table
    @param pos the position
    @param v the value to be inserted*)


val superterms_at : 'a positiontable ref -> position -> 'a list
(** Lookup all superterm entries {e at} a given position (bottom-up)
    @param table a referenced table
    @param pos the position *)


val superterms_no_closer_than : 'a positiontable ref -> position -> 'a list
(** Lookup all superterm entries {e no closer than} a given position (bottom-up)
    @param table a referenced table
    @param pos the position *)


val superterms_no_farther_than : 'a positiontable ref -> position -> 'a list
(** Lookup all superterm entries {e no farther than} a given position (bottom-up)
    @param table a referenced table
    @param pos the position *)
*)