(* ========================================================================= *)
(* The Signature                                                             *)
(* ========================================================================= *)

open Hol_type
open Term
open List

type ('a,'b) hashtbl = ('a,'b) Hashtbl.t

type signature = {
          basetypes : typ list;
    mutable typevars  : (string, typ) hashtbl;
    mutable tvarcount : int ;
    fixedlogs : (string, term) hashtbl;
(*                name,  symbol, definition, type  *)
    defineds  : (string, term * term * typ) hashtbl;
    uis       : (string, typ) hashtbl;
  }

type symbol_class = FixedBT | Typevar | FixedLogicalSymbol | DefinedSymbol | UISymbol


let bt_i = Type("ind",[])
let bt_o = Type("bool", [])
let bt_x = TVar(("X",0), ["tvar"])
let bt_type = Type("type",[])

let conjunction_ty = abstr_type bt_o (abstr_type bt_o bt_o)
let forall_ty = abstr_type (abstr_type bt_x bt_o) bt_o

let neg = Const("neg", Type("fun", [bt_o; bt_o]))
let forall = Const("forall", forall_ty)
let disjunction = Const("or", conjunction_ty)
let equality = Const( "eq", abstr_type bt_x  (abstr_type bt_x bt_o))
let l_true = Const("true", bt_o)
let l_false = Const("false", bt_o)


(* defined logical symbols: *)
let exists =  Const("exists", forall_ty)
let exists_def = Const("foo", bt_o)
(*   Abstr("P",abstr_type bt_x bt_o, *)
(*     Appl(neg,Appl(forall, *)
(*       Abstr( "X",bt_x, *)
(*         Appl( neg,Appl("P","X")))))) *)
let exists_ty = abstr_type (abstr_type bt_x bt_o) bt_o

let conjunction = Const("and", conjunction_ty)
let conjunction_def =Const("foo", bt_o)
(*   Abstr("X", bt_o, *)
(*     Abstr("Y", bt_o, *)
(*       Appl( neg, *)
(*         Appl( *)
(*           Appl(disjunction, *)
(*             Appl(neg, "X")), *)
(*           Appl(neg, "Y"))))) *)

let implies = Const("implies", conjunction_ty) (* => *)
let implies_def =Const("foo", bt_o)
(*   Abstr("X", bt_o, *)
(*     Abstr("Y", bt_o, *)
(*       Appl( *)
(*         Appl(disjunction, *)
(*           Appl(neg, "X")), *)
(*         "Y"))) *)
let implies_ty = abstr_type bt_o (abstr_type bt_o bt_o)

let i_f_ty = abstr_type bt_o (abstr_type bt_o bt_o)
let i_f = Const("if", i_f_ty) (* <= *)
let i_f_def = Const("foo", bt_o)
(*   Abstr("X", bt_o, *)
(*     Abstr("Y", bt_o, *)
(*       Appl( *)
(*         Appl(disjunction, *)
(*           "X"), *)
(*         Appl(neg, "Y")))) *)

let iff_ty = abstr_type bt_o (abstr_type bt_o bt_o)
let iff = Const("iff", iff_ty) (* <=> *)
let iff_def = Const("foo", bt_o)
(*   Abstr("X", bt_o, *)
(*     Abstr("Y", bt_o, *)
(*       Appl( *)
(*         Appl(conjunction, *)
(*           Appl( *)
(*             Appl(implies, "X"), *)
(*             "Y")), *)
(*         Appl( *)
(*           Appl(implies, "Y"), *)
(*           "X")))) *)

let nequals_ty = abstr_type bt_x (abstr_type bt_x bt_o)
let nequals = Const("nequals", nequals_ty) (* != *)
let nequals_def = Const("foo", bt_o)
(*   Abstr("X", bt_x, *)
(*     Abstr("Y", bt_x, *)
(*       Appl(neg, *)
(*         Appl( *)
(*            Appl(equality, "X"), *)
(*           "Y")))) *)




let hashtbl_of_list ls =
  let ht = Hashtbl.create (length ls) in
  let _ = iter (fun (k,v) -> Hashtbl.add ht k v) ls in
  ht

let list_of_hashtbl ht =
  Hashtbl.fold (fun k v acc -> (k,v)::acc) ht []


let new_signature () = {
  basetypes = [bt_i; bt_o; bt_type];
  typevars  = hashtbl_of_list [("X", TVar(("X",0), ["tvar"]))];
  tvarcount = 0;
  fixedlogs = hashtbl_of_list [
      ("neg",         neg);
      ("false",       l_false) ;
      ("true",        l_true) ;
      ("forall",      forall) ;
      ("or",          disjunction) ;
      ("=",            equality) ];
  defineds  = hashtbl_of_list [
      ("exists",      (exists, exists_def,       exists_ty));
      ("conjunction", (conjunction, conjunction_def,  conjunction_ty));
      ("implies",     (implies, implies_def,      implies_ty));
      ("if",          (i_f, i_f_def,    i_f_ty))
	(*   (iff,         (iff_def,         Some iff_ty)); *)
	(*   (nequals,     (nequals_def,     Some nequals_ty)) *)
    ];
  uis       = Hashtbl.create 10
}
  
  
let copy_signature sigma = {
  basetypes = sigma.basetypes;
  typevars  = sigma.typevars;
  tvarcount = sigma.tvarcount; 
  fixedlogs = sigma.fixedlogs;
  defineds  = sigma.defineds;
  uis       = sigma.uis
}


let add_type_var sigma name =
  let tvar = TVar((name, sigma.tvarcount), ["tvar"]) in 
  (Hashtbl.add sigma.typevars name tvar);
  sigma.tvarcount <- sigma.tvarcount + 1; 
  tvar
    
let get_type_var sigma name =
  try (
     Hashtbl.find sigma.typevars name
   )
  with Not_found -> failwith "get_type_var: unknown type variable"
      
      
exception ADD_EX of string 
    
let add_defined_symbol sigma sym def =
  match sym with
    Const(name, bt_x) ->
      (Hashtbl.add sigma.defineds name (sym, def ,  bt_x))
(*   | _ -> print_string "can only add definitions for constants and not for "^(Term.to_string sym)^"\n" *)
												    
let defined_symbol_set_type sigma sym ty =
  match sym with
    Const(name, _) ->
      try ( 
	let (sym,def,oldtyp) = Hashtbl.find sigma.defineds name in
	Hashtbl.replace sigma.defineds name (sym, def,  ty))
      with 
	Not_found -> (Hashtbl.add sigma.defineds name (Const(name, ty),Const(name, ty) ,  ty))
(* failwith "defined_symbol_set_type: unknown symbol" *)
	
let add_uninterpreted_symbol sigma name ty =
  Hashtbl.add sigma.uis name ty


let get_defined_symbol sigma name =
  try (
    match Hashtbl.find sigma.defineds name with
      (_,t,_) -> t)
  with Not_found -> failwith "get_defined_symbol: unknown symbol"


let all_fixed_basetypes sigma =
  sigma.basetypes


let all_type_vars sigma =
  list_of_hashtbl sigma.typevars


let all_fixed_logical_symbols sigma =
  list_of_hashtbl sigma.fixedlogs


let all_defined_symbols sigma =
  list_of_hashtbl sigma.defineds


let all_uninterpreted_symbols sigma =
  list_of_hashtbl sigma.uis

let rec my_find n vs i = 
  match vs with
    [] -> -1
  | (name,_)::vr -> 
      if (name = n) then 
	i
      else 
	my_find n vr (i+1)

let rec fix_types_aux sigma term bound_vars  = 
  match term with 
      Const(name, tp) as t -> 
	(try (let (_,_,tyo) = Hashtbl.find sigma.defineds name in
		Const(name, tyo))
	  with Not_found -> Const(name, tp))
    | Free(s, tp) as t -> t
    | Var((name,i), _) as t  ->   
	let index = my_find name bound_vars 0 in 
	  if (index = -1) then t 
	  else  Bound(index)
    | Bound(n) as t -> t 
    | Typ(tp) as t -> t
    | Appl(t1,t2) ->
	Appl( (fix_types_aux sigma t1 bound_vars), (fix_types_aux sigma t2 bound_vars))
    | Abstr(x,ty,t1) -> 
	Abstr(x,ty, (fix_types_aux sigma t1 ((x,ty)::bound_vars)))
	  
let fix_types sigma term = 
  fix_types_aux sigma term []
    
    
let class_of_symbol sigma name =
  if name = "i" or name = "o" then
    Some FixedBT
  else if Hashtbl.mem sigma.fixedlogs name then
    Some FixedLogicalSymbol
  else if Hashtbl.mem sigma.defineds name then
    Some DefinedSymbol
  else if Hashtbl.mem sigma.uis name then
    Some UISymbol
  else if Hashtbl.mem sigma.typevars name then
    Some Typevar
  else None
    

let type_of_symbol sigma name =
  try (
    let (_,_,tyo) = Hashtbl.find sigma.defineds name in
    match tyo with
    | ty -> ty
   )
  with Not_found ->
    try Hashtbl.find sigma.uis name
    with Not_found ->
      failwith "type_of_symbol: unknown symbol"
	

      
