(* ========================================================================= *)
(* The Signature                                                             *)
(* ========================================================================= *)

(** Module Signature implements HOL signatures.
    @author Arnaud
    @since 07-06-06*)

open Hol_type
open Term


type ('a,'b) hashtbl = ('a,'b) Hashtbl.t

type signature = {
    basetypes : typ list;
    mutable typevars  : (string, typ) hashtbl;
    mutable tvarcount : int ;
    fixedlogs : (string, term) hashtbl;
(*                name,  symbol, definition, type  *)
    defineds  : (string, term * term * typ) hashtbl;
    uis       : (string, typ) hashtbl;
  }
(** abstract type of HOL signatures *)

(** classes of symbols in the signature *)
type symbol_class = FixedBT | Typevar | FixedLogicalSymbol | DefinedSymbol | UISymbol


(** {6 Fixed basetypes and logical constants} *)

(** the two fixed base types: *)

val bt_i : typ
val bt_o : typ
val bt_x : typ
val bt_type : typ


(** the primitive logical constants: *)

val neg : term
val forall : term
val disjunction : term
val equality : term

val l_true : Term.term
val l_false : Term.term

(** defined logical symbols: *)

val exists : term 
val conjunction : term
val implies : term (* => *)
val i_f : term (* <= *)
val iff : term
val nequals : term  
(* tbc... *)


(** {6 Signature construction} *)

val new_signature : unit -> signature
(** Returns a fresh signature containing base types, logical constants and
    the type variable "X". *)

val copy_signature : signature -> signature
(** [copy_signature sigma] returns a copy of [sigma]. *)

val add_type_var : signature -> string -> Hol_type.typ

val get_type_var : signature -> string -> Hol_type.typ

val add_defined_symbol :  signature -> Term.term -> Term.term -> unit

val defined_symbol_set_type : signature -> Term.term -> Hol_type.typ -> unit

val add_uninterpreted_symbol : signature -> string -> typ -> unit


(** {6 Access functions} *)

val get_defined_symbol : signature -> string -> Term.term

val all_fixed_basetypes : signature -> typ list

val all_type_vars : signature ->  (string * typ) list

val all_fixed_logical_symbols : signature -> (string * term) list

val all_defined_symbols : signature -> (string * (Term.term * Term.term * Hol_type.typ)) list

val all_uninterpreted_symbols : signature -> (string * typ) list

val fix_types : signature -> term -> term 

val class_of_symbol : signature -> string -> symbol_class option
(** Returns None if symbol is not present in the signature,
    don't know if this is really needed *)

val type_of_symbol : signature -> string -> typ
(** Assumes the symbol is defined or uninterpreted and present in the signature *)

