(* ========================================================================= *)
(* User readable terms                                                       *)
(* ========================================================================= *)

open Hol_type

(** the user-readable terms have a simple structure which should not
    be hidden in an ADT.
    The insertion into the termset is easier to implement if the term
    structure is public. *)
    
(*Terms in Isabelle.  Bound variables are indicated by depth number.
  Free variables, (scheme) variables and constants have names.
  An term is "closed" if every bound variable of level "lev"
  is enclosed by at least "lev" abstractions.

  It is possible to create meaningless terms containing loose bound vars
  or type mismatches.  But such terms are not allowed in rules. *)


type term =
    Const of string * typ |
    Free of string * typ |
    Var of indexname * typ |
    Bound of int |
    Typ of typ |
    Abstr of string * typ * term |
    Appl of term * term

(* type term = *)
(*     Symbol of string  *)
(*   | Const of string * hol_type  *)
(*   | Var of string  * hol_type  *)
(*   | Bound of int  *)
(*   | Typ of hol_type  *)
(*   | Appl of term * term  *)
(*   | Abstr of term * hol_type * term *)


let term_for_type t =
  Typ(t)

let is_symbol = function
    Const(_,_) -> true
  | Var(_,_) -> true
  | Free(_,_) -> true
  | Bound(_) -> true
  | Typ(_) -> true
  | _ -> false

exception TYPE of string * term

(* from Isabelle Pure/term.ML  *)
(*Compute the type of the term, checking that combinations are well-typed
   Ts = [T0,T1,...] holds types of bound variables 0, 1, ...*)
(* 	(match t with  *)
(* 	    TVar(_,_) ->  *)
(* 	      let real_type = (Signature.type_of_symbol !symSig name) in  *)
(* 		(print_string ("type of Const: "^name^" :: "^(Hol_type.to_string real_type)^"\n")); *)
(* 		real_type *)
(* 	  | tp -> tp) *)
  
let print_types ts = 
  (List.iter (fun t -> (print_string (", "^(Hol_type.to_string t)))) ts);
  (print_string "\n")

let rec type_match t1 t2 = 
  (print_string ("type match "^(Hol_type.to_string t1)^" *TO* "^(Hol_type.to_string t2)^"\n"));
  match (t1, t2) with 
      (Type(n1, [a1;a2]), Type(n2, [b1;b2])) -> 
	(n1 = n2) && (type_match a1 b1) && (type_match a2 b2)
    | (Type(n1, []), Type(n2, [])) -> 	(n1 = n2)
    | (TFree(n1,_), TFree(n2,_)) -> n1 = n2
    | (TVar((n1, i1), _), t2) -> true
    | (_,_) -> false
	
let rec type_of1 ts te = 
  match (ts, te) with  
      (ts, Const(name,t)) -> t	
    | (ts, Free  (_,t)) -> t
    | (ts, Var(_,t)) -> t
    | (ts, Bound(i)) -> 
	(try 
	    (List.nth ts i)
	  with Failure(_) -> raise (TYPE ("type_of: bound variable ", Bound(i))))
    | (ts, Typ(t)) -> Hol_type.bt_type
    | (ts, Abstr(x,t,body)) -> 
(* 	(print_string ("type check lambda: "^(to_string (Abstr(x,t,body)))^"\n")); *)
	let bodyt = (type_of1 (t::ts) body) in 
	  Type("fun", [t; bodyt])
    | (ts, Appl(f, u)) ->
(* 	(print_string ("type check appl: "^(to_string (Appl(f,u)))^"\n")); *)
	(print_string ("bound types : "));
	(print_types ts);
	let tu = (type_of1 ts u) in 
	let tf = (type_of1 ts f) in
	  (print_string ("applying symbol "^(Hol_type.to_string tf)^" *TO* "^(Hol_type.to_string tu)^"\n"));
	  match tf with 
              Type("fun", [t1;t2]) -> 
		if t1=tu then t2  
		else if (type_match t1 tu) then t2 
		else raise (TYPE ("type_of: type mismatch in application ", Appl(f, u)))
		  (* ^(Hol_type.string_of t1)^" "^(Hol_type.string_of tu)^", term "^(Term.string_of Appl(f,u))) *)
	    | _ -> raise (TYPE ("type_of: type mismatch in application ", Appl(f,u)))
		(* ^(Hol_type.string_of t1)^" "^(Hol_type.string_of tu)^", term "^(Term.string_of Appl(f,u)))	     *)
		
let type_of t : typ = (type_of1 [] t)
  
(* let get_symbol = function *)
(*     Symbol s -> s *)
(*   | _ -> failwith "not a symbol" *)
  
let is_appl = function
    Appl _ -> true
  | _ -> false
  
let is_abstr = function
    Abstr _ -> true
  | _ -> false  


let rec bn bound = function
    Const(s, tp) as t ->
      if List.mem_assoc s bound then
        List.assoc s bound
      else t
  | Free(s, tp) as t ->
      if List.mem_assoc s bound then
        List.assoc s bound
      else t 
  | Var((s,i), _) as t ->
      if List.mem_assoc s bound then
        List.assoc s bound
      else t
  | Bound(n) -> Bound(n) 
  | Typ(t) -> Typ(t)
  | Appl(Abstr(x,_,t1),t2) ->
      bn ((x, bn bound t2)::bound) t1
  | Appl(t1,t2) -> Appl(bn bound t1, bn bound t2)
  | Abstr(x,ty,t1) -> Abstr(x,ty,bn bound t1)

let beta_normalize = bn []


let multi_abstr args body =
  List.fold_right (fun (x,ty) acc -> Abstr(x, ty, acc)) args body

let multi_quantified quantifier args body =
  List.fold_right (fun (x,ty) acc -> Appl(quantifier,Abstr(x, ty, acc))) args body


(** pretty printer *)
let rec to_string = function
  | Const(s, tp) -> s^"_"^(Hol_type.to_string tp)
  | Var((s,i), tp) -> s^"_"^(string_of_int i)^"_"^(Hol_type.to_string tp)
  | Free(s, tp) -> s^"_"^(Hol_type.to_string tp)
  | Bound(n) -> "bound_"^(string_of_int n)
  | Typ(tp) -> Hol_type.to_string tp
  | Appl(t1,t2) -> (add_pars t1)^" "^(add_pars t2)
  | Abstr(x,ty,t) -> "lambda ["^x^" : "^(Hol_type.to_string ty)^"]: "^(to_string t)
and add_pars t =
  if is_symbol t then to_string t
  else "("^(to_string t)^")"
