(* ========================================================================= *)
(* Sets of terms in shared representation                                    *)
(* ========================================================================= *)

open Hol_type
open Term
open Signature
open Darray
open List

type id = int
type ('a,'b) hashtbl = ('a,'b) Hashtbl.t

type nodestruct =
  Symbol_node of string |
  Appl_node of id * id |
  Abstr_node of hol_type * id |
  Bound_node of hol_type * int

type scopevar =
  Boundvar of int * hol_type |
  Appliedvar of term * ((string * scopevar) list)
  
type termnode = {
  name : id;
  structure : nodestruct;
  mutable parents : int;
}

type termset = {
 nodes          : termnode darray;
 term2id        : (nodestruct,id) hashtbl;
 appl_with_arg  : (id, (id,id) hashtbl) hashtbl;
 appl_with_func : (id, (id,id) hashtbl) hashtbl;
 abstr_with     : (id*hol_type, id) hashtbl;
 nodetypes      : (id,hol_type) hashtbl
}


let new_termset () =
  (* arbitrary initial value *)
  let i = 10 in
  {
    nodes = da_make i {name=0; structure=Symbol_node ""; parents = 0};
    term2id = Hashtbl.create i;
    appl_with_arg = Hashtbl.create i;
    appl_with_func = Hashtbl.create i;
    abstr_with = Hashtbl.create i;
    nodetypes = Hashtbl.create i;
  }

let appl_with_argument ts func arg =
  Hashtbl.find (Hashtbl.find ts.appl_with_arg func) arg

let appl_with_function ts arg func =
  Hashtbl.find (Hashtbl.find ts.appl_with_func arg) func

let abstr_with_type ts body typ =
  Hashtbl.find ts.abstr_with (body,typ)

let appl_exists ts func arg =
  try (
    Hashtbl.mem (Hashtbl.find ts.appl_with_arg func) arg
  ) with _ -> false

let abstr_exists ts body typ =
  Hashtbl.mem ts.abstr_with (body,typ)

let node_exists ts id =
  try (
    ignore (da_get ts.nodes id);
    true
  ) with _ -> false

(** create a new node *)
let make_node name structure =
  { name = name;
    structure = structure;
    parents = 0 }

(** get node by id *)
let get_node ts id =
  da_get ts.nodes id

(** set node by id *)
let set_node ts id node =
  da_set ts.nodes id node

(** increase node's parents counter by one *)
let add_parent ts id =
  let node = get_node ts id in
  let _ = node.parents <- node.parents + 1 in
  set_node ts id node

(** decrease node's parents counter by one *)
let remove_parent ts id =
  let node = get_node ts id in
  assert(node.parents > 0);
  let _ = node.parents <- node.parents - 1 in
  set_node ts id node

(** remove a node from the term graph *)
let rec remove_node ts id =
  let node = get_node ts id in
  let _ = match node.structure with
      Abstr_node (_,id1) ->
        if (get_node ts id1).parents = 1 then
          remove_node ts id1
        else remove_parent ts id1        
    | Appl_node(id1,id2) -> (
        if (get_node ts id1).parents = 1 then
          remove_node ts id1
        else remove_parent ts id1;
        if (get_node ts id1).parents = 1 then
          remove_node ts id1
        else remove_parent ts id1)
    | _ -> ()
  in da_remove ts.nodes id


(* ------------------------------------------------------------------------- *)
(* val insert : termset -> nodestruct -> id                                  *)
(*                                                                           *)
(* [insert ts term nodestruct] assumes [nodestruct] corresponds to [term],   *)
(* determines a fresh [id] for [nodestruct], inserts [nodestruct] into [ts]  *)
(* and updates all hashtables.                                               *)
(* ------------------------------------------------------------------------- *)

let insert ts nodestruct =
  match nodestruct with
    Abstr_node (typ,id1) ->
      if abstr_exists ts id1 typ
      then abstr_with_type ts id1 typ
      else
      let id = da_next_free ts.nodes in
      let _ = da_insert ts.nodes (make_node id nodestruct) in
      let _ = add_parent ts id1 in
      let _ =  Hashtbl.add ts.abstr_with (id1,typ) id in
      id     
  | Appl_node (id1,id2) ->
      if appl_exists ts id1 id2
      then appl_with_function ts id2 id1
      else
      let id = da_next_free ts.nodes in
      let _ = da_insert ts.nodes (make_node id nodestruct) in
      let _ = add_parent ts id1 in
      let _ = add_parent ts id2 in
      let appl_with_arg_id1 = ref (Hashtbl.create 10) in
      let _ =
        if Hashtbl.mem ts.appl_with_arg id1 then
          appl_with_arg_id1 := Hashtbl.find ts.appl_with_arg id1 in
      Hashtbl.add !appl_with_arg_id1 id2 id;
      Hashtbl.replace ts.appl_with_arg id1 !appl_with_arg_id1;
      let appl_with_func_id2 = ref (Hashtbl.create 10) in
      let _ =
        if Hashtbl.mem ts.appl_with_func id2 then
          appl_with_func_id2 := Hashtbl.find ts.appl_with_func id2 in
      Hashtbl.add !appl_with_func_id2 id1 id;
      Hashtbl.replace ts.appl_with_func id2 !appl_with_func_id2;
      id 
  | _ ->
      if Hashtbl.mem ts.term2id nodestruct
      then Hashtbl.find ts.term2id nodestruct
      else
      let id = da_next_free ts.nodes in
      let _ = da_insert ts.nodes (make_node id nodestruct) in
      Hashtbl.add ts.term2id nodestruct id;
      id


(* ---------------------------------------------------------------------------------------------- *)
(* val create' : termset -> term -> int -> (string*(int*hol_type)) list -> (string*id) list -> id *)
(*                                                                                                *)
(* [create' ts t d bound applied] expects                                                         *)
(*   [ts] a termset                                                                               *)
(*   [t] a term                                                                                   *)
(*   [d] the current scope                                                                        *)
(*   [bound] a list mapping symbols to pairs of deBruijn-indices and types                        *)
(*   [applied] a list mapping symbols to IDs for beta-reduction                                   *)
(* ---------------------------------------------------------------------------------------------- *)

let rec create' ts t d scope argstack =
  let rec make_appl id1 d argstack1 =
        (match argstack1 with
           ((Appliedvar(t1,scope_t1))::rest) ->
             let id2 = create' ts t1 d scope_t1 [] in
             let id3 = insert ts (Appl_node (id1,id2)) in
             make_appl id3 d rest
         | _ -> id1)
  in
  match t with
    Symbol s ->
      if mem_assoc s scope
      then
      (match assoc s scope with
         Boundvar (ds,typs) ->
	   let id1 = insert ts (Bound_node (typs, d-ds)) in
           make_appl id1 d argstack
       | Appliedvar (t1,scope_t1) -> create' ts t1 d scope_t1 argstack)
      else let id1 = insert ts (Symbol_node s) in
        make_appl id1 d argstack
  | Var s -> -1
  | Const(s,t) -> -1
  | Type t -> -1
  | Appl (t1,t2) -> create' ts t1 d scope (Appliedvar(t2,scope)::argstack)
  | Abstr (x,typx,t1) ->
     assert(is_symbol x);
     (match argstack with
        [] -> let id1 = create' ts t1 (d+1) ((get_symbol x, Boundvar (d+1,typx))::scope) [] in
              insert ts (Abstr_node (typx,id1))
      | (a::arest) -> create' ts t1 d ((get_symbol x, a)::scope) arest)

let create ts t = create' ts t (-1) [] []


let rec retrieve' ts id d bound =
  assert(node_exists ts id);
  match (da_get ts.nodes id).structure with
    Symbol_node s -> Symbol s
  | Appl_node (id1,id2) -> Appl(retrieve' ts id1 d bound, retrieve' ts id2 d bound)
  | Abstr_node (ty,id1) ->
          (* deBruijn -> varname may be done by custom function *)
          let var_name = "x" ^ (string_of_int (d+1)) in
          Abstr(Symbol var_name, ty, retrieve' ts id1 (d+1) ((d+1,var_name)::bound))
  | Bound_node (_,idx) ->
          assert(mem_assoc idx bound);
          Symbol(assoc (d-idx) bound)

let retrieve ts id = retrieve' ts id (-1) []


(* lazily computes types of nodes,
   assuming the nodes represent well-typed terms *)
let rec node_type sigma ts id =
  assert(node_exists ts id);
  try Hashtbl.find ts.nodetypes id
  with Not_found ->
    let update id' ty =
      let _ = Hashtbl.add ts.nodetypes id' ty in ty
    in
    match (da_get ts.nodes id).structure with
        Symbol_node s ->
            update id (type_of_symbol sigma s)
      | Appl_node (id1,id2) ->
            (*let ty1,ty2 = node_type sigma ts id1, node_type sigma ts id2 in (* ty2 doesn't really need to be computed *)*)
            let ty1 = node_type sigma ts id1 in
            update id (result_type ty1)
      | Abstr_node (ty,id1) ->
            let ty1 = node_type sigma ts id1 in
            update id (abstr_type ty ty1)
      | Bound_node (ty,_) ->
            update id ty


let eta_expand sigma ts id =
  let arg_types = all_arg_types (node_type sigma ts id) in
  fold_left (fun id' ty ->
      let applied_var = insert ts (Bound_node (ty,0)) in
      let appl_node = insert ts (Appl_node (id', applied_var)) in
      insert ts (Abstr_node (ty, appl_node)))
  id arg_types
    
    
let to_string ts =
  let ty id =
    try (
      Hol_type.to_string (Hashtbl.find ts.nodetypes id))
    with Not_found -> "?"
  in
  let ns = function
      Symbol_node s -> "symbol "^s
    | Appl_node (id1,id2) -> "appl("^(string_of_int id1)^","^(string_of_int id2)^")"
    | Abstr_node (ty,id) -> "abstr("^(Hol_type.to_string ty)^","^(string_of_int id)^")"
    | Bound_node (ty,idx) -> "bound("^(Hol_type.to_string ty)^","^(string_of_int idx)^")"
  in
  let nodearray = da2array ts.nodes in
  let s = ref "" in
  for i = 0 to Array.length nodearray -1 do
    let node = nodearray.(i) in
    (*s := !s ^ (string_of_int node.name) ^ ": "^(ns node.structure)^ "   " ^(ty i)^ "\n"*)
    (*s := !s ^ (Printf.sprintf "%2i: %-16s  %s\n" node.name (ns node.structure) (ty i))*)
    s := !s ^ (Printf.sprintf "%2i: %-16s  %s       parents: %i\n" node.name (ns node.structure) (ty i) node.parents)
  done;
  !s


