 (*  Author: Jia Meng, Cambridge University Computer Laboratory
    ID: $Id: clause.ml,v 1.6 2006/08/21 13:23:20 jzimmer Exp $
    Copyright 2004 University of Cambridge

ML data structure for storing/printing FOL clauses and arity clauses.
Typed equality is treated differently.
*)

(* module ResClause : RES_CLAUSE  = struct *)

open Hol_type 
open Term 
open Symtab


type clause_id = int
type axiom_name = string
type polarity = bool
type cclass = string 
type sort  = cclass list



(* Added for typed equality *)
  let special_equal = ref false (* by default,equality does not carry type information *)
  let eq_typ_wrapper = "typeinfo" (* default string *)
      
  let schematic_var_prefix = "V_"
  let fixed_var_prefix = "v_"
      
  let tvar_prefix = "T_"
  let tfree_prefix = "t_"
      
  let clause_prefix = "cls_" 
  let arclause_prefix = "clsarity_" 
  let clrelclause_prefix = "clsrel_"
      
  let const_prefix = "c_"
  let tconst_prefix = "tc_" 
  let class_prefix = "class_" 
(* TODO is @ the equivalent for union? *)
  let union_all xss = (List.fold_left (@) [] xss)
      
(*Provide readable names for the more common symbolic functions*)
(* let const_trans_table = *)
(*   Symtab.make [("op =", "equal"), *)
(* 	       ("Orderings.less_eq", "lessequals"), *)
(* 	       ("Orderings.less", "less"), *)
(* 	       ("op &", "and"), *)
(* 	       ("op |", "or"), *)
(* 	       ("HOL.plus", "plus"), *)
(* 	       ("HOL.minus", "minus"), *)
(* 	       ("HOL.times", "times"), *)
(* 	       ("Divides.op div", "div"), *)
(* 	       ("HOL.divide", "divide"), *)
(* 	       ("op -->", "implies"), *)
(* 	       ("{}", "emptyset"), *)
(* 	       ("op :", "in"), *)
(* 	       ("op Un", "union"), *)
(* 	       ("op Int", "inter"), *)
(* 	       ("List.op @", "append")] *)
  
(* let type_const_trans_table = *)
(*       Symtab.make [("*", "prod"); *)
(* 	  	   ("+", "sum"); *)
(* 		   ("~=>", "map")] *)

(*Escaping of special characters.
  Alphanumeric characters are left unchanged.
  The character _ goes to __
  Characters in the range ASCII space to / go to _A to _P, respectively.
  Other printing characters go to _NNN where NNN is the decimal ASCII code.*)
let isAlphaNum c = 
  ('A' <= c && c <= 'Z') ||
  ('a' <= c && c <= 'z') ||
  ('0' <= c && c <= '9')


let a_minus_space = (Char.code 'A') - (Char.code ' ')

let ascii_of_c c =
  if isAlphaNum c then Char.escaped c
  else if c = '_' then "__"
  else if (' ' <= c && c <= '/')
  then "_" ^ Char.escaped (Char.chr (Char.code c + a_minus_space))
  else "_" ^ (string_of_int (Char.code c))
	       
let ascii_of c = ""^(ascii_of_c c)
(* let ascii_of = String.translate ascii_of_c *)

let rec commas  = function
  | [] -> "" 
  | [s] -> s
  | s::sr -> (s ^ ", " ^ (commas sr))


(* convert a list of strings into one single string surrounded by brackets *)
let paren_pack = function 
    [] -> ""   (*empty argument list*)
  | strings -> "(" ^ (commas strings) ^ ")"

let bracket_pack strings = "[" ^ (commas strings) ^ "]"


(*Remove the initial ' character from a type variable, if it is present*)
let trim_type_var s = s
(*   if (s <> "") && ((String.sub s 0 1) = "'") then String.extract(s,1,NONE) *)
(*   else error ("trim_type: Malformed type variable encountered: " ^ s) *)

let ascii_of_indexname ind =  
  match ind with 
    (v,0) -> v
  | (v,i) -> v ^ "_" ^ string_of_int i
				  
let make_schematic_var v = schematic_var_prefix ^ (ascii_of_indexname v)
(* let make_fixed_var x = fixed_var_prefix ^ (ascii_of x) *)
let make_fixed_var x = fixed_var_prefix^x

let make_schematic_type_var (x,i) =
    tvar_prefix^(ascii_of_indexname (trim_type_var x,i))

let make_fixed_type_var x = tfree_prefix ^ x
(*  (ascii_of (trim_type_var x)) *)

let lookup_const c = c

(*     match Symtab.lookup const_trans_table c of *)
(*         SOME c' => c' *)
(*       | NONE => ascii_of c *)

let lookup_type_const c = c
(*     case Symtab.lookup type_const_trans_table c of *)
(*         SOME c' => c' *)
(*       | NONE => ascii_of c *)

let make_fixed_const = function 
    "op =" -> "="   (*MUST BE "equal" because it's built-in to ATPs*)
  | c      -> const_prefix ^ lookup_const c

let make_fixed_type_const c = tconst_prefix ^ lookup_type_const c

(* let make_type_class clas = class_prefix ^ ascii_of_c clas *)
let make_type_class clas = class_prefix ^  clas

(***** definitions and functions for FOL clauses, for conversion to TPTP or DFG format. *****)

let keep_types = ref true


type kindt = Axiom | Hypothesis | Conjecture

let name_of_kind = function 
    Axiom -> "axiom"
  | Hypothesis -> "hypothesis"
  | Conjecture -> "conjecture"
	
(* Types are classified by sorts. *)
(* type arity = string * sort list * sort *)

(* type typ = *)
(*     Type  of string * typ list | *)
(*     TFree of string * sort | *)
(*     TVar  of indexname * sort *)

(**** Isabelle FOL clauses ****)

type pred_name = string

type typ_var = FOLTVar of indexname | FOLTFree of string

exception FOL_TYPE of string 

(*FIXME: give the constructors more sensible names*)
type fol_type = 
    AtomV of string
  | AtomF of string
  | Comp of string * fol_type list
	
let rec string_of_fol_type = function 
    (AtomV x) -> x
  |  (AtomF x) -> x
  |  (Comp(tcon,tps)) ->
      tcon ^ (paren_pack (List.map string_of_fol_type tps))
      
let mk_fol_type = function 
    ("Var",x,_) -> AtomV(x)
  | ("Fixed",x,_) -> AtomF(x)
  | ("Comp",con,args) -> Comp(con,args)
  | _ -> raise (FOL_TYPE "mk_fol_type")


(*First string is the type class the second is a TVar or TFfree*)
type type_literal = 
    LTVar of string * string
  | LTFree of string * string

type fol_term = UVar of string * fol_type
                  | Fun of string * fol_type list * fol_term list
type predicate = Predicate of pred_name * fol_type list * fol_term list

type literal = Literal of polarity * predicate

let mk_typ_var_sort = function
    TFree(a,s) -> (FOLTFree a,s)
  | TVar(v,s) -> (FOLTVar v,s)

type thmrec = {prop: Term.term}                 (*conclusion*)
type thm = Thm of thmrec

(*A clause has first-order literals and other, type-related literals*)
type clauserec =  {clause_id: clause_id ;
		   axiom_name: axiom_name ;
		   th: thm ;
		   kind: kindt ;
		   literals: literal list ;
		   types_sorts: (typ_var * sort) list}

type clause = Clause of clauserec 


exception CLAUSE of string * Term.term

let isFalse = function
    (Literal(pol, Predicate(pname,_,[]))) ->
      (pol && pname = "c_False") ||
      (not pol && pname = "c_True")
    | _ -> false
	
let isTrue = function 
    (Literal (pol, Predicate(pname,_,[]))) ->
      (pol && pname = "c_True") ||
      (not pol && pname = "c_False")
  |  _ -> false
  
let isTaut cls = 
  match cls with 
    Clause(clsrec) -> List.exists isTrue clsrec.literals  


(*Declarations of the current theory--to allow suppressing types.*)
(* let const_typargs = ref (Library.K [] : (string*typ -> typ list)) *)

(* let num_typargs(s,T) = if !keep_types then length (!const_typargs (s,T)) else 0 *)

(*Initialize the type suppression mechanism with the current theory before
    producing any clauses!*)
(* let init thy = (const_typargs := Sign.const_typargs thy) *)
    

let rec unzip  = function 
    [] -> ([],[])
  | [(a, b)]  -> ([a], [b])
  | (a,b)::ar -> 
      let (xs,ys) = (unzip ar) in 
      (a::xs, b::ys)
	
(*Flatten a type to a fol_type while accumulating sort constraints on the TFrees and
  TVars it contains.*)    
let rec type_of = function 
    (Type(a, ts)) ->
      let (folTyps, ts) = types_of ts in
      let t = make_fixed_type_const a
      in (Comp(t,folTyps), ts)
  |  (TFree (a,s)) -> (AtomF((make_fixed_type_var a)), [(FOLTFree a, s)])
  |  (TVar (v, s)) -> (AtomV((make_schematic_type_var v)), [(FOLTVar v, s)])
and types_of ts =
  let (folTyps,ts) = unzip (List.map type_of ts)
  in (folTyps, union_all ts) 
    

(*TODO*)
let const_tpa = function 
    (str,ty) -> [ty]

	
let const_typargs = ref const_tpa

let const_types_of (c,t) = types_of (!const_typargs (c,t))

(* Any variables created via the METAHYPS tactical should be treated as
   universal vars, although it is represented as "Free(...)" by Isabelle *)
let isMeta x = 
  match (String.sub x 0 9) with 
     "METAHYP1_" -> true 
  | _ -> false


let pred_name_type = function 
    Const(c,t) -> (make_fixed_const c, const_types_of (c,t))
   | Free(x,t) -> if (isMeta x) then
       raise (CLAUSE ("Predicate Not First Order 1", Free(x,t)))
   else (make_fixed_var x, ([],[]))
  |  Var(_) as v -> raise (CLAUSE ("Predicate Not First Order 2 ", v))
  |  t        -> raise (CLAUSE ("Predicate input unexpected ", t))


(* For typed equality *)
(* here "arg_typ" is the type of "="'s argument's type, not the type of the equality *)
(* Find type of equality arg *)
let eq_arg_type = function 
    (Type("fun",[t;_])) ->
      let (folT,_) = type_of t
      in  folT  
  | t -> raise (FOL_TYPE "eq_art_type")
	
let fun_name_type term args = 
  match term, args with 
    (Const(c,t)), args -> (make_fixed_const c, const_types_of (c,t))
  | (Free(x,t)), args  -> 
      if isMeta x then raise (CLAUSE ("Function Not First Order", Free(x,t)))
       else (make_fixed_var x, ([],[]))
  | f, args -> raise (CLAUSE ("Function Not First Order 1", f))

(*Convert a term to a fol_term while accumulating sort constraints on the TFrees and
  TVars it contains.*)
let strip_comb u : term * term list =
  let rec stripc  = function 
      (Appl(f,t), ts) -> stripc (f, t::ts)
    | x ->  x
  in
  stripc(u,[])  
    
let rec term_of = function 
    (Var(ind_nm,t)) ->
      let (folType,ts) = type_of t
      in (UVar(make_schematic_var ind_nm, folType), ts) 
  |  (Free(x,t)) ->
      let (folType, ts) = type_of t
      in
	  if isMeta x then (UVar(make_schematic_var(x,0),folType), ts)
	  else (Fun(make_fixed_var x, [], []), ts)  (*Frees don't need types!*)
      
  |  app ->
      let (f,args) = strip_comb app in 
      let (funName,(contys,ts1)) = fun_name_type f args in 
      let (args',ts2) = terms_of args
      in
	  (Fun(funName,contys,args'), union_all (ts1::ts2))
and terms_of ts = unzip (List.map term_of ts)

(*Create a predicate letue, again accumulating sort constraints.*)    
let pred_of (term, args) = 
  match (term, args) with 
    (Const("op =", typ), args) ->
      let arg_typ = eq_arg_type typ  in 
      let (args',ts) = terms_of args in
      let equal_name = make_fixed_const "op ="
      in
      (Predicate(equal_name,[arg_typ],args'),
       union_all ts)
	
  | (pred,args) ->
      let (pname, (predType,ts1)) = pred_name_type pred in 
      let (args',ts2) = terms_of args
      in
	  (Predicate(pname,predType,args'), union_all (ts1::ts2))
      

(*Treatment of literals, possibly negated*)
let rec predicate_of (term, pol) = 
  match (term, pol) with  
    (Appl(Const("neg",_), p), polarity) -> predicate_of (p, (not polarity))
  | (term,polarity) -> ((pred_of (strip_comb term)), polarity)

     
  
let rec literals_of_term1 args term = 
  match (args, term) with
    (args, Appl(Const("Trueprop",_), p)) -> literals_of_term1 args p
  | (args, Appl(Appl(Const("op |",_), p), q)) -> 
      literals_of_term1 (literals_of_term1 args p) q	
  | ((lits,ts), p) ->
      let ((pred, ts'),polarity) = predicate_of (p,true) in
      let lits' = Literal(polarity,pred)::lits
      in
      (lits', ts @ ts')
	
let literals_of_term = literals_of_term1 ([],[])


(* let rec list_ord as bs = function  *)
(*     (_, ([],[])) -> EQUAL *)
(*   | (_, ([],_)) -> LESS *)
(*   | (_, (_,[])) -> GREATER *)
(*   | (ord, (x::xs, y::ys)) -> *)
(*       (match ord(x,y) with EQUAL -> (list_ord ord (xs,ys)) *)
(*       | xy_ord -> xy_ord) *)
		     
(*Make literals for sorted type variables.  FIXME: can it use map?*) 
let rec sorts_on_typs (t, ss) = 
  match (t,ss) with  
    (_, [])   -> ([]) 
  |  (v, "HOL.type"::s) ->
      sorts_on_typs (v,s)                (*IGNORE sort "type"*)
  |  ((FOLTVar indx), s::ss) ->
      LTVar(make_type_class s, make_schematic_type_var indx) :: 
      sorts_on_typs ((FOLTVar indx), ss)
  |  ((FOLTFree x), s::ss) ->
      LTFree(make_type_class s, make_fixed_type_var x) :: 
      sorts_on_typs ((FOLTFree x), ss)


let pred_of_sort = function 
    (LTVar (s,ty)) -> (s,1)
|   (LTFree (s,ty)) -> (s,1)

 (* from Isabelle's Pure/library.ML *)
let member eq list x =
  let rec memb = function 
      [] -> false
    | (y :: ys) -> eq (x, y) || memb ys
  in memb list 


let string_eq (x, y) = (x=y)
let filter_out f = List.filter (fun x -> (not (f x)))

let insert eq x xs = if member eq xs x then xs else x :: xs
let remove eq x xs = if member eq xs x then filter_out (fun y -> eq (x, y)) xs else xs
let ins_string x xs = (insert string_eq  x xs)


let rec union_string xs ys =
  match (xs,ys) with 
    (_,[]) -> xs
  | ([],ys) -> ys
  | ((x::xs), ys) ->  union_string xs (ins_string x ys)


(*Given a list of sorted type variables, return two separate lists.
  The first is for TVars, the second for TFrees.*)
let rec add_typs_aux tss = 
  match tss with  
    [] -> ([],[])
  | ((FOLTVar(indx),s)::tss) ->
      let vs = sorts_on_typs (FOLTVar(indx), s) in
      let (vss,fss) = (add_typs_aux tss)
      in
      (union_string vs vss, fss)
      
  | ((FOLTFree(x),s)::tss) ->
      let fs = sorts_on_typs (FOLTFree(x), s) in
      let (vss,fss) = (add_typs_aux tss)
      in
      (vss, union_string fs fss)
	

(** Too general means, positive equality literal with a variable X as one operand,
  when X does not occur properly in the other operand. This rules out clearly
  inconsistent clauses such as V=a|V=b, though it by no means guarantees soundness. **)

let rec occurs a term = 
  match a, term with 
    a, (UVar(b,_)) -> a=b
  | a, (Fun (_,_,ts)) -> List.exists (occurs a) ts

(*Is the first operand a variable that does not properly occur in the second operand?*)
let too_general_terms (t1, t2) = 
  match (t1, t2) with 
    (UVar _, UVar _) -> false
  | (Fun _, _) -> false
  | (UVar (a,_), t) -> not (occurs a t)
	
let too_general_lit = function 
    (Literal(true, Predicate("equal",_,[x;y]))) ->
      (too_general_terms (x,y)) || (too_general_terms (y,x))
  | _ -> false

(*   thy_ref: theory_ref;   *)
(*   der: bool * Pt.proof;       *)
(*   maxidx: int,                 *)
(*   shyps: sort list,           *)
(*   hyps: term list,             *)
(*   tpairs: (term * term) list,  *)


let prop_of thm =
  match thm with Thm(thmrec) -> thmrec.prop

(** make axiom and conjecture clauses. **)

exception MAKE_CLAUSE of string

let make_clause (clause_id, axiom_name, thm, kind) =
  let term = prop_of thm
  in
  let (lits,types_sorts) = literals_of_term term 
  in
  if List.for_all isFalse lits  then
    raise (MAKE_CLAUSE "Problem too trivial for resolution (empty clause)")
  else
    if (kind=Axiom && (List.for_all too_general_lit lits))
    then
      raise (MAKE_CLAUSE "Omitting axiom, equalities are too general")
    else let new_clause = Clause({clause_id = clause_id; 
				  axiom_name = axiom_name;
				  th = thm;
				  kind = kind;
				  literals = lits;
				  types_sorts = types_sorts})
    in 
    new_clause 
      
  
let rec get_tvar_strs xs = 
  match xs with  
    [] -> []
  | (FOLTVar(indx),s)::tss ->
      (make_schematic_type_var indx)::(get_tvar_strs tss)
  | (FOLTFree(x),s)::tss -> get_tvar_strs tss

(* TODO *) 
let check_is_fol_term t = true 

exception TERM of string * term list

(* check if a clause is first-order before making a conjecture clause*)
let make_conjecture_clause n thm =
  let t = prop_of thm in
  let _ =
    try 
      check_is_fol_term t
    with 
      TERM("check_is_fol_term",_) -> raise (CLAUSE ("Goal is not FOL",t))
  in
  make_clause(n, "conjecture", thm, Conjecture)
    
    

let rec make_conjecture_clauses_aux n bs = 
  match (n, bs) with 
    (_, []) -> []
  | (n, (t::ts)) -> (make_conjecture_clause n t)::(make_conjecture_clauses_aux (n+1) ts)
						    
let make_conjecture_clauses = make_conjecture_clauses_aux 0

(* let is_fol_term t = *)
(*   not (List.exists (has_bool o fastype_of) (term_vars t)  ||  not (Term.is_first_order ["all","All","Ex"] t) || *)
(*   has_bool_arg_const t ||   *)
(*     has_meta_conn t) *)
  
(*before converting an axiom clause to "clause" format, check if it is FOL*)
let make_axiom_clause thm (ax_name,cls_id) =
(*   if is_fol_term (prop_of thm)  *)
(*   then *)
  try 
    (Some (make_clause(cls_id, ax_name, thm, Axiom)))
  with  
    MAKE_CLAUSE(_) -> None
(*   else (Output.debug ("Omitting " ^ ax_name ^ ": Axiom is not FOL") NONE) *)
      
let rec make_axiom_clauses = function 
    [] -> []
  | ((thm,(name,id))::thms) ->
      match (make_axiom_clause thm (name,id)) with 
          Some cls -> if isTaut cls then (make_axiom_clauses thms)
                      else (name,cls) :: (make_axiom_clauses thms)
        | None -> make_axiom_clauses thms

(**** Isabelle arities ****)

(* exception ARCLAUSE of string *)
 
type tcons = string 

type arLit = TConsLit of bool * (cclass * tcons * string list)
               | TVarLit of bool * (cclass * string)
 
type arityClauseRec = {aclause_id: clause_id;
	  	       aaxiom_name: axiom_name;
		       akind: kindt;
		       aconclLit: arLit;
		       apremLits: arLit list}

type arityClause =  ArityClause of arityClauseRec

let get_axiomName (Clause cls) = cls.axiom_name
let get_axiomName (ArityClause cls) = cls.aaxiom_name

let get_kind (Clause cls) = cls.kind
let get_kind (ArityClause cls) = cls.akind

let rec gen_TVars = function 
    0 -> []
  | n -> ("T_" ^ string_of_int n) :: gen_TVars (n-1)

let rec pack_sort = function 
    (_,[])  -> []
  | (tvar, "HOL.type"::srt) -> pack_sort(tvar, srt)   (*IGNORE sort "type"*)
  | (tvar, cls::srt) ->   (make_type_class cls, tvar) :: pack_sort(tvar, srt)
    
let make_TVarLit (b, (cls,str)) = TVarLit(b, (cls,str))
let make_TConsLit (b, (cls,tcons,tvars)) = 
      TConsLit(b, (make_type_class cls, make_fixed_type_const tcons, tvars))

let rec zip (xs, ys) = 
  match (xs, ys) with 
    (x::xr, y::yr) -> (x,y)::(zip (xr, yr))
  | ([], []) -> []
  | (_, _) -> []

(* TODO is this pair?*)
let pair a = 
  let pairc b = (a,b) 
  in pairc

(*Arity of type constructor tcon :: (arg1,...,argN)res*)
let make_axiom_arity_clause (tcons, n, (res,args)) =
  let nargs = List.length args in 
  let tvars = gen_TVars nargs in
  let tvars_srts = (zip (tvars,args)) in
  let tvars_srts' = union_all(List.map pack_sort tvars_srts) in
  let false_tvars_srts' = List.map (pair false) tvars_srts'
  in
      ArityClause {aclause_id = n; 
		   akind = Axiom;
                   aaxiom_name = lookup_type_const tcons;
                   aconclLit = make_TConsLit(true, (res,tcons,tvars));
                   apremLits = List.map make_TVarLit false_tvars_srts'}
   


(**** Isabelle class relations ****)
type classrelClauseRec = {caxiom_name: axiom_name;
			  subclass: cclass;
			  superclass: cclass}

type classrelClause =  ClassrelClause of classrelClauseRec

(* let make_axiom_classrelClause n subclass superclass = *)
(*   ClassrelClause {axiom_name = (clrelclause_prefix ^ ascii_of subclass ^  *)
(*                   "_" ^ string_of_int n); *)
(*                   subclass = (make_type_class subclass);  *)
(*                   superclass = (make_type_class superclass)} *)

(* let classrelClauses_of_aux = function  *)
(*     (n, sub, []) ->  [] *)
(*   | (n, sub, ("HOL.type"::sups)) -> (\*Should be ignored*\) *)
(*       (classrelClauses_of_aux n sub sups) *)
(*   | (n, sub, (sup::sups)) -> *)
(*       (make_axiom_classrelClause n sub sup) :: (classrelClauses_of_aux (n+1) sub sups) *)

(* let classrelClauses_of (sub,sups) = classrelClauses_of_aux 0 sub sups *)

(* let classrel_clauses_thy = *)
(*   let cls = o.classes in  *)
(*   maps classrelClauses_of o Graph.dest o Sorts.rep_algebra o Sign.classes_of *)


(** Isabelle arities **)

(* let rec arity_clause = function  *)
(*     (_, (tcons, [])) -> [] *)
(*   | (n, (tcons, ("HOL.type",_)::ars)) -> (\*ignore*\) *)
(*       arity_clause n (tcons,ars) *)
(*   | (n, (tcons, ar::ars)) -> *)
(*       (make_axiom_arity_clause (tcons,n,ar)) ::  *)
(*       (arity_clause (n+1) (tcons,ars)) *)

(* let multi_arity_clause = function  *)
(*     [] -> [] *)
(*   | ((tcons,ars) :: tc_arlists) -> *)
(*       (\*Reversal ensures that older entries always get the same axiom name*\) *)
(*       arity_clause 0 (tcons, rev ars)  @   *)
(*       multi_arity_clause tc_arlists  *)
	
(* let arity_clause_thy thy = *)
(*   let arities = thy |> Sign.classes_of *)
(*     |> Sorts.rep_algebra |> #arities |> Symtab.dest *)
(*     |> map (apsnd (map (fn (c, (_, Ss)) => (c, Ss)))) *)
(*   in multi_arity_clause (rev arities)  *)


(**** Find occurrences of predicates in clauses ****)

(*FIXME: multiple-arity checking doesn't work, as update_new is the wrong 
  function (it flags repeated declarations of a function, even with the same arity)*)

(* let update_many (tab, keypairs) = List.fold_left (uncurry Symtab.update) tab keypairs *)

(* let add_predicate_preds (Predicate(pname,tys,tms), preds) =  *)
(*   if pname = "equal" then preds (\*equality is built-in and requires no declaration*\) *)
(*   else Symtab.update (pname, length tys + length tms) preds *)

(* let add_literal_preds (Literal(_,pred), preds) = add_predicate_preds (pred,preds) *)

(* let add_type_sort_preds = function  *)
(*     ((FOLTVar indx,s), preds) ->  update_many (preds, List.map pred_of_sort (sorts_on_typs (FOLTVar indx, s))) *)
(*   | ((FOLTFree x,s), preds) -> *)
(*       update_many (preds, List.map pred_of_sort (sorts_on_typs (FOLTFree x, s))) *)
      
(* let add_clause_preds cls preds = *)
(*   try  *)
(*     List.fold_left add_literal_preds (List.fold_left add_type_sort_preds preds cls.types_sorts) cls.literals *)
(*   with Symtab.DUP a -> raise (ERROR "predicate " ^ a ^ " has multiple arities") *)

(* let add_classrelClause_preds cls preds =  *)
(*   (Symtab.update (cls.subclass,1) (Symtab.update (cls.superclass,1) preds)) *)

(* let add_arityClause_preds acls  preds = *)
(*   let TConsLit(_, (tclass, _, _)) = acls.conclLit *)
(*   in Symtab.update (tclass,1) preds   *)

(* let preds_of_clauses clauses clsrel_clauses arity_clauses =  *)
(*   Symtab.dest *)
(*     (foldl add_classrelClause_preds  *)
(*       (foldl add_arityClause_preds *)
(*         (foldl add_clause_preds Symtab.empty clauses) *)
(*         arity_clauses) *)
(*       clsrel_clauses) *)

(* (\*** Find occurrences of functions in clauses ***\) *)

(* let add_foltype_funcs = function  *)
(*     (AtomV _, funcs) -> funcs *)
(*   |  (AtomF a, funcs) -> Symtab.update (a,0) funcs *)
(*   |  (Comp(a,tys), funcs) ->  *)
(*       List.fold_left add_foltype_funcs (Symtab.update (a, length tys) funcs) tys *)

(* let add_folterm_funcs = function  *)
(*     (UVar(_,ty), funcs) -> add_foltype_funcs (ty, funcs) *)
(*   | (Fun(a,tys,tms), funcs) -> *)
(*       List.fold_left add_foltype_funcs  *)
(* 	(List.fold_left add_folterm_funcs (Symtab.update (a, length tys + length tms) funcs)  *)
(* 	   tms)  *)
(* 	tys *)
	
(* let add_predicate_funcs (Predicate(_,tys,tms), funcs) =  *)
(*     List.fold_left add_foltype_funcs (List.fold_left add_folterm_funcs funcs tms) tys *)

(* let add_literal_funcs (Literal(_,pred), funcs) = add_predicate_funcs (pred,funcs) *)

(* (\*TFrees are recorded as constants*\) *)
(* let add_type_sort_funcs = function  *)
(*     ((FOLTVar _, _), funcs) -> funcs *)
(*   | ((FOLTFree a, _), funcs) -> *)
(*       Symtab.update (make_fixed_type_var a, 0) funcs *)

(* let add_arityClause_funcs acls funcs = *)
(*   let TConsLit(_, (_, tcons, tvars)) = acls.conclLit *)
(*   in  Symtab.update (tcons, length tvars) funcs   *)

(* let add_clause_funcs cls funcs = *)
(*   try  *)
(*     List.fold_left add_literal_funcs (List.fold_left add_type_sort_funcs funcs cls.types_sorts) *)
(*       cls.literals *)
(*   with Symtab.DUP a -> raise (ERROR "function " ^ a ^ " has multiple arities") *)
      
(* let funcs_of_clauses clauses arity_clauses =  *)
(*   Symtab.dest (List.fold_left add_arityClause_funcs  *)
(*                      (List.fold_left add_clause_funcs Symtab.empty clauses) *)
(*                      arity_clauses) *)


(**** String-oriented operations ****)

let wrap_eq_type typ t = eq_typ_wrapper ^"(" ^ t ^ "," ^ typ ^ ")"

(*Only need to wrap equality's arguments with "typeinfo" if the output clauses are typed 
 and if we specifically ask for types to be included.   *)
let rec string_of_equality (typ,terms) =
  let [tstr1;tstr2] = match List.map string_of_term terms with 
    [x;y] -> [x;y] 
  | _ -> ["foo";"bar"] in 
  let typ' = string_of_fol_type typ
  in
  if (!keep_types && !special_equal)
  then "equal(" ^ (wrap_eq_type typ' tstr1) ^ "," ^ 
    (wrap_eq_type typ' tstr2) ^ ")"
  else "equal(" ^ tstr1 ^ "," ^ tstr2 ^ ")"
and string_of_term = function 
    (UVar(x,_)) -> x
  | (Fun("equal",[typ],terms)) -> string_of_equality (typ,terms)
  | (Fun(name,typs,[])) -> name ^ (paren_pack (List.map string_of_fol_type typs))
  | (Fun(name,typs,terms)) -> 
      let terms_as_strings = List.map string_of_term terms in
      let typs' = if !keep_types then List.map string_of_fol_type typs else []
      in  name ^ (paren_pack (terms_as_strings @ typs'))  

(* before output the string of the predicate, check if the predicate corresponds to an equality or not. *)
let string_of_predicate = function 
    (Predicate("equal",[typ],terms)) -> string_of_equality (typ,terms)
  | (Predicate(name,typs,terms)) ->
      let terms_as_strings = List.map string_of_term terms in
      let typs' = if !keep_types then List.map string_of_fol_type typs else []
      in  name ^ (paren_pack (terms_as_strings @ typs'))  

let string_of_clausename (cls_id,ax_name) = 
    clause_prefix ^  ax_name ^ "_" ^ string_of_int cls_id

let string_of_type_clsname (cls_id,ax_name,idx) = 
    string_of_clausename (cls_id,ax_name) ^ "_tcs" ^ (string_of_int idx)

(*Write a list of strings to a file*)
let writeln_strs out strs   = List.iter (fun s -> output_string out s) strs
(*   List.app (fn s => TextIO.output (os,s)) *)

    
(**** Producing DFG files ****)

(*Attach sign in DFG syntax: false means negate.*)
(* let dfg_sign true s = s *)
(*   | dfg_sign false s = "not(" ^ s ^ ")"   *)

(* let dfg_literal (Literal(pol,pred)) = dfg_sign pol (string_of_predicate pred) *)

(* let dfg_of_typeLit (LTVar (s,ty)) = "not(" ^ s ^ "(" ^ ty ^ "))" *)
(*   | dfg_of_typeLit (LTFree (s,ty)) = s ^ "(" ^ ty ^ ")" *)
 
(*Enclose the clause body by quantifiers, if necessary *)
(* let dfg_forall [] body = body   *)
(*   | dfg_forall vars body = "forall([" ^ commas vars ^ "],\n" ^ body ^ ")" *)

(* let gen_dfg_cls (cls_id, ax_name, knd, lits, vars) =  *)
(*     "clause( %(" ^ knd ^ ")\n" ^  *)
(*     dfg_forall vars ("or(" ^ lits ^ ")") ^ ",\n" ^  *)
(*     string_of_clausename (cls_id,ax_name) ^  ").\n\n" *)

(* let dfg_clause_aux (Clause{literals, types_sorts, ...}) =  *)
(*   let lits = List.map dfg_literal literals *)
(*       let (tvar_lits,tfree_lits) = add_typs_aux types_sorts *)
(*       let tvar_lits_strs =  *)
(* 	  if !keep_types then List.map dfg_of_typeLit tvar_lits else [] *)
(*       let tfree_lits = *)
(*           if !keep_types then List.map dfg_of_typeLit tfree_lits else [] *)
(*   in *)
(*       (tvar_lits_strs @ lits, tfree_lits) *)
   

(* let dfg_folterms (Literal(pol,pred)) =  *)
(*   let Predicate (_, _, folterms) = pred *)
(*   in  folterms   *)

(* let get_uvars (UVar(a,typ)) = [a]  *)
(*   | get_uvars (Fun (_,typ,tlist)) = union_all(List.map get_uvars tlist) *)

(* let dfg_vars (Clause {literals,...}) = *)
(*   union_all (List.map get_uvars (List.concat (List.map dfg_folterms literals))) *)

(* let clause2dfg (cls as Clause{axiom_name,clause_id,kind,types_sorts,...}) = *)
(*     let (lits,tfree_lits) = dfg_clause_aux cls  *)
(*            "lits" includes the typing assumptions (TVars) *)
(*         let vars = dfg_vars cls *)
(*         let tvars = get_tvar_strs types_sorts *)
(* 	let cls_str = gen_dfg_cls(clause_id, axiom_name, name_of_kind kind,  *)
(* 	                           commas lits, tvars@vars)  *)
(*     in (cls_str, tfree_lits)  *)

(* let string_of_arity (name, num) =  "(" ^ name ^ "," ^ string_of_int num ^ ")" *)

(* let string_of_preds [] = "" *)
(*   | string_of_preds preds = "predicates[" ^ commas(List.map string_of_arity preds) ^ "].\n" *)

(* let string_of_funcs [] = "" *)
(*   | string_of_funcs funcs = "functions[" ^ commas(List.map string_of_arity funcs) ^ "].\n"  *)

(* let string_of_symbols predstr funcstr =  *)
(*   "list_of_symbols.\n" ^ predstr  ^ funcstr  ^ "_of_list.\n\n" *)

(* let string_of_start name = "begin_problem(" ^ name ^ ").\n\n" *)

(* let string_of_descrip name =  *)
(*   "list_of_descriptions.\nname({*" ^ name ^  *)
(*   "*}).\nauthor({*Isabelle*}).\nstatus(unknown).\ndescription({*auto-generated*}).\n_of_list.\n\n" *)

(* let dfg_tfree_clause tfree_lit = *)
(*   "clause( %(conjecture)\n" ^ "or( " ^ tfree_lit ^ "),\n" ^ "tfree_tcs" ^ ").\n\n" *)

  let string_of_arClauseID x = 
    match x with 
      (ArityClause cls) -> 
	arclause_prefix^cls.aaxiom_name^"_"^(string_of_int cls.aclause_id) 

(* let dfg_of_arLit (TConsLit(pol,(c,t,args))) = *)
(*       dfg_sign pol (c ^ "(" ^ t ^ paren_pack args ^ ")") *)
(*   | dfg_of_arLit (TVarLit(pol,(c,str))) = *)
(*       dfg_sign pol (c ^ "(" ^ str ^ ")") *)
    
(* let dfg_classrelLits sub sup =  "not(" ^ sub ^ "(T)), " ^ sup ^ "(T)" *)

(* let dfg_classrelClause (ClassrelClause {axiom_name,subclass,superclass,...}) = *)
(*   "clause(forall([T],\nor( " ^ dfg_classrelLits subclass superclass ^ ")),\n" ^ *)
(*   axiom_name ^ ").\n\n" *)
      
(* let dfg_arity_clause (arcls as ArityClause{kind,conclLit,premLits,...}) =  *)
(*   let arcls_id = string_of_arClauseID arcls *)
(*       let knd = name_of_kind kind *)
(*       let TConsLit(_, (_,_,tvars)) = conclLit *)
(*       let lits = List.map dfg_of_arLit (conclLit :: premLits) *)
(*   in *)
(*       "clause( %(" ^ knd ^ ")\n" ^  *)
(*       dfg_forall tvars ("or( " ^ commas lits ^ ")") ^ ",\n" ^ *)
(*       arcls_id ^ ").\n\n" *)
  

(* write out a subgoal in DFG format to the file "xxxx_N" *)
(* let dfg_write_file thms filename (ax_tuples,classrel_clauses,arity_clauses) =  *)
(*   let  *)
(*     let _ = Output.debug ("Preparing to write the DFG file " ^ filename) *)
(*     let conjectures = make_conjecture_clauses thms *)
(*     let (clnames,axclauses) = ListPair.unzip (make_axiom_clauses ax_tuples) *)
(*     let (dfg_clss, tfree_litss) = ListPair.unzip (List.map clause2dfg conjectures) *)
(*     let clss = conjectures @ axclauses *)
(*     let funcs = funcs_of_clauses clss arity_clauses *)
(*     and preds = preds_of_clauses clss classrel_clauses arity_clauses *)
(*     and probname = Path.pack (Path.base (Path.unpack filename)) *)
(*     let (axstrs, _) = ListPair.unzip (List.map clause2dfg axclauses) *)
(*     let tfree_clss = List.map dfg_tfree_clause (union_all tfree_litss)  *)
(*     let out = TextIO.openOut filename *)
(*   in *)
(*     TextIO.output (out, string_of_start probname)  *)
(*     TextIO.output (out, string_of_descrip probname)  *)
(*     TextIO.output (out, string_of_symbols (string_of_funcs funcs) (string_of_preds preds))  *)
(*     TextIO.output (out, "list_of_clauses(axioms,cnf).\n") *)
(*     writeln_strs out axstrs *)
(*     List.app (curry TextIO.output out o dfg_classrelClause) classrel_clauses *)
(*     List.app (curry TextIO.output out o dfg_arity_clause) arity_clauses *)
(*     TextIO.output (out, "_of_list.\n\nlist_of_clauses(conjectures,cnf).\n") *)
(*     writeln_strs out tfree_clss *)
(*     writeln_strs out dfg_clss *)
(*     TextIO.output (out, "_of_list.\n\n_problem.\n") *)
(*     TextIO.closeOut out *)
(*     clnames *)
  


(**** Produce TPTP files ****)

(*Attach sign in TPTP syntax: false means negate.*)
let tptp_sign b s  = 
  match (b,s) with  
    (true, s) -> "++" ^ s
  | (false, s) ->  "--" ^ s

let tptp_literal (Literal(pol,pred)) = 
      (if pol then "++" else "--") ^ string_of_predicate pred

let tptp_of_typeLit = function 
    (LTVar (s,ty)) -> "--" ^ s ^ "(" ^ ty ^ ")"
  | (LTFree (s,ty)) -> "++" ^ s ^ "(" ^ ty ^ ")"
 
let gen_tptp_cls (cls_id,ax_name,knd,lits) = 
    "cnf(" ^ string_of_clausename (cls_id,ax_name) ^ "," ^ 
    knd ^ "," ^ lits ^ ").\n"

let gen_tptp_type_cls (cls_id,ax_name,knd,tfree_lit,idx) = 
    "cnf(" ^ string_of_type_clsname (cls_id,ax_name,idx) ^ "," ^ 
    knd ^ ",[" ^ tfree_lit ^ "]).\n"

let tptp_type_lits x = 
  match x with  
    (Clause cls) ->  
      let lits = List.map tptp_literal cls.literals in 
      let (tvar_lits,tfree_lits) = add_typs_aux cls.types_sorts in 
      let tvar_lits_strs =
	if !keep_types then List.map tptp_of_typeLit tvar_lits else []
      in
      let tfree_lits =
	if !keep_types then List.map tptp_of_typeLit tfree_lits else []
      in
      ((List.append tvar_lits_strs lits), tfree_lits)
	
(* {clause_id = clause_id;  *)
(* 	    axiom_name = axiom_name; *)
(* 	    th = thm; *)
(* 	    kind = kind; *)
(* 	    literals = lits; *)
(* 	    types_sorts = types_sorts}) *)
  
let clause2tptp cls = 
  match cls with 
    (Clause cl) ->  
      let (lits,tfree_lits) = (tptp_type_lits (Clause(cl)))in
      (*"lits" includes the typing assumptions (TVars)*)
      let knd = (name_of_kind cl.kind) in
      let cls_str = gen_tptp_cls(cl.clause_id, cl.axiom_name, knd, bracket_pack lits) 
      in
      (cls_str,tfree_lits) 
    

let tptp_tfree_clause tfree_lit =
    "cnf(" ^ "tfree_tcs," ^ "conjecture" ^ ",[" ^ tfree_lit ^ "]).\n"

let tptp_of_arLit = function 
    (TConsLit(b,(c,t,args))) ->  tptp_sign b (c ^ "(" ^ t ^ paren_pack args ^ ")")
  |  (TVarLit(b,(c,str))) ->     tptp_sign b (c ^ "(" ^ str ^ ")")
    
let tptp_arity_clause cl = function 
    (ArityClause arcls) ->
      let arcls_id = (string_of_arClauseID (ArityClause arcls)) in
      let knd = (name_of_kind arcls.akind) in
      let lits = List.map tptp_of_arLit (arcls.aconclLit :: arcls.apremLits)
      in
      "cnf("^arcls_id^","^(bracket_pack lits)^").\n"
  

let tptp_classrelLits sub sup = 
    let tvar = "(T)"
    in 
	"[--" ^ sub ^ tvar ^ ",++" ^ sup ^ tvar ^ "]"


let tptp_classrelClause clrelcls =
  "cnf(" ^ clrelcls.caxiom_name ^ ",axiom," ^ tptp_classrelLits clrelcls.subclass clrelcls.superclass ^ ").\n" 

(* write out a subgoal as tptp clauses to the file "xxxx_N"*)
let tptp_write_file thms filename (ax_tuples,classrel_clauses,arity_clauses) =
  let _ = print_string ("Preparing to write the TPTP file " ^ filename) in 
  let clss = make_conjecture_clauses thms in 
  let (clnames,axclauses) = unzip (make_axiom_clauses ax_tuples) in 
  let (tptp_clss,tfree_litss) = unzip (List.map clause2tptp clss) in 
(*   let tfree_clss = List.map tptp_tfree_clause (List.fold_left (op union_string) [] tfree_litss) in  *)
  let tfree_clss = (List.map tptp_tfree_clause (List.fold_left union_string [] tfree_litss)) in  
  let out = open_out filename in 
(*  let out = TextIO.openOut filename in*)
  

(*   List.app (curry TextIO.output (out o (#1 o clause2tptp))) axclauses; *)
  writeln_strs out tfree_clss;
  writeln_strs out tptp_clss;
(*   List.app (curry TextIO.output out o tptp_classrelClause) classrel_clauses; *)
(*   List.app (curry TextIO.output out o tptp_arity_clause) arity_clauses; *)
  close_out out ;
  clnames
    
(*     (curry TextIO.output out o o1 clause2tptp) axclauses *)
(*     (writeln_strs out tfree_clss) *)
(*     (writeln_strs out tptp_clss) *)
(*     (List.app (curry TextIO.output out o tptp_classrelClause) classrel_clauses) *)
(*     (List.app (curry TextIO.output out o tptp_arity_clause) arity_clauses) *)
(*     (TextIO.closeOut out) *)
(*     clnames *)
    

(* end;; *)
