%{

open Hol_type
open Term
open Signature

let parse_error s =
  print_endline s;
  flush stdout

let terms = ref []
let conjs = ref []
let axioms = ref []
let bound = ref []
let varcount = ref 0

let sigma = new_signature ()
(* TODO: define non-primitive logical symbols in signature *)

let reset_vars = 
  varcount := !varcount + 1;
  bound := []

let add_term t =
  (reset_vars) ;
  terms := t :: !terms

let add_conj t =
  (reset_vars) ;
  conjs := t :: !conjs

let add_axiom t =
  (reset_vars) ;
  axioms := t :: !axioms

let add_def (s,t) =
  (reset_vars) ;
  (add_defined_symbol sigma s t)

let rec my_find n vs i = 
  match vs with
    [] -> Free(n, Signature.bt_x)
  | (name,t)::vr -> 
      if (name = n) then 
	Bound(i)
      else 
	my_find n vr (i+1)
	  
let get_var n = 
  my_find n !bound 0
    

let add_var_decls ts =   
  bound := !bound @ ts ;
  List.iter (fun (n,t) -> (print_string ("new var "^n^": "^(Hol_type.to_string t)^"\n"))) ts

let my_add_type_var sigma name =  
  let tvar = (add_type_var sigma name) in 
  (** pretty printer *)
  print_string ("type vars: ");
  List.iter (fun (n,t) -> print_string ((Hol_type.to_string t)^", "))
    (all_type_vars sigma) ;
  print_string ("\n"); 
  tvar

let fix_type s t = 
  match s with 
    Const(name, ty) -> Const(name, t)
  | r -> r


%}

%start input
%type <Term.term list * Term.term list * Term.term list * Signature.signature> input
%token AMPERSAND
%token AT_SIGN
%token CARET
%token CNF
%token COLON
%token COMMA
%token EQUALS
%token EXCLAMATION
%token FOF
%token GETS
%token GREATER
%token HOF
%token IF
%token IFF
%token IMPLIES
%token INCLUDE
%token INPUT_CLAUSE
%token INPUT_FORMULA
%token DEFINITION
%token CONJECTURE
%token AXIOM
%token LAMBDA
%token LBRKT
%token LPAREN
%token MAP_TO
%token MMINUS
%token NAMPERSAND
%token NEQUALS
%token NIFF
%token NVLINE
%token PERIOD
%token PPLUS
%token QUESTION
%token RPAREN
%token RBRKT
%token TILDE
%token TOK_FALSE
%token TOK_TRUE
%token TOK_I
%token TOK_O
%token TOK_TYPE
%token VLINE
%token EOF
%token <int> Unsigned_integer
%token <int> Signed_integer
%token <int> Decimal_part
%token <float> Real
%token <string> Atomic_system_word
%token <string> Distinct_object
%token <string> Single_quoted
%token <string> Upper_word
%token <string> Lower_word
%token Unrecognized
%%

/* 
*/

input : /* (Term.term list * Term.term list * Term.term list * Signature.signature) */
  | null { (!terms,!axioms,!conjs,sigma) }
  | input hof_annotated { (!terms,!axioms,!conjs,sigma) }
;


hof_annotated : /* unit */
  | HOF LPAREN name COMMA DEFINITION COMMA definition RPAREN PERIOD { add_def $7 }
  | HOF LPAREN name COMMA CONJECTURE COMMA term RPAREN PERIOD { add_conj $7 }
  | HOF LPAREN name COMMA AXIOM COMMA term RPAREN PERIOD { add_axiom $7 }
  | HOF LPAREN name COMMA formula_role COMMA term RPAREN PERIOD { add_term $7 }
  | definition PERIOD { add_def $1 }
  | term PERIOD { add_term $1 }
;

/* hof_formula :  formula
;
*/
 
formula_role : /* string */
  | atomic_word { $1 } 
;

name : /* string */
  | atomic_word { $1 }
;

definition : /* (Term.term * Term.term) */
  | defined_element GETS term { ($1,$3) }
  | LPAREN defined_element GETS term RPAREN { ($2,$4) }
;

defined_element : /* Term.term */
  | constant { $1 }
  | typed_constant { $1 }
;

typed_constant : /* Term.term */
  | LPAREN constant COLON type_expr RPAREN { (fix_type $2 $4) } 
;

constant : /* Term.term */
  | TILDE { Signature.neg }
  | VLINE { Signature.disjunction }
  | EQUALS { Signature.equality }
  | atomic_word   { Const($1, Signature.bt_x) }
;
  

term : /* Term.term */
  | apply_term { $1 }
  | lambda_term { $1 }
  | defined_atom { $1 }
  | LPAREN term RPAREN { $2 }
  | variable { $1 }
  | constant { $1 }
  | quantified_term { $1 }
  | unary_connective term { Appl($1,$2) }
  | term binary_connective term { Appl(Appl($2,$1),$3) }
;

binary_connective : /* Term.term */
  | IFF { Signature.iff } 
  | IMPLIES { Signature.implies }
  | IF { Signature.i_f }
  | VLINE { Signature.disjunction }
  | AMPERSAND { Signature.conjunction }
  | EQUALS { Signature.equality }
  | NEQUALS { Signature.nequals }
;

defined_atom : /* Term.term */
  | TOK_TRUE { Signature.l_true }
  | TOK_FALSE { Signature.l_false }
;

unary_connective : /* Term.term */
  | TILDE { Signature.neg }
;

quantified_term : /* Term.term */
  | quantifier LBRKT variable_decls RBRKT COLON term { (multi_quantified $1 $3 $6) }
;

quantifier : /* Term.term */
  | EXCLAMATION { Signature.forall }
  | QUESTION { Signature.exists }
;

apply_term : /* Term.term */
  | term AT_SIGN term { Appl($1,$3) }
  | term AT_SIGN type_term { Appl($1,$3) }
;

lambda_term : /* Term.term */
  | lambda_symbol LBRKT variable_decls RBRKT COLON term { (multi_abstr $3 $6) }
;

lambda_symbol :
  | LAMBDA { }
  | CARET { }
;

type_term : /* Term.term */
  | LPAREN unitary_type map_arrow type_expr RPAREN { term_for_type (abstr_type $2 $4) }
  | defined_type { term_for_type $1 }
;

unitary_type : /* Hol_type.hol_type */
  | atomic_type { $1 }
  | LPAREN unitary_type map_arrow type_expr RPAREN { abstr_type $2 $4 }
;

variable_decls : /* (string * Hol_type.hol_type) list */
  | variable_decl {  [$1] }
  | variable_decl COMMA variable_decls { $1::$3 }
;

variable_decl : /* (string * Hol_type.hol_type) */
  | variable_name COLON type_expr { ($1, $3) }
  | variable_name COLON TOK_TYPE { ($1, my_add_type_var sigma $1)  }
;

type_expr : /* Hol_type.hol_type */
  | atomic_type { $1 }
  | type_expr map_arrow type_expr { abstr_type $1 $3 }
  | LPAREN type_expr RPAREN { $2 }
;


atomic_type : /* Hol_type.hol_type */
  | defined_type { $1 }
  | type_variable { $1}
;

defined_type : /* Hol_type.hol_type */
  | TOK_O { Signature.bt_o }
  | TOK_I { Signature.bt_i }
  | TOK_TYPE { Signature.bt_type }
;

map_arrow :
  | MAP_TO { }
  | GREATER  { }
;

type_variable : /* Hol_type.hol_type */
  | Upper_word { typevar $1 }
;

atomic_word : /* string */
  | Lower_word { print_string("atomic word "^$1^"\n"); $1 }
;

variable_name : /* string */
  | Upper_word { print_string("Variable decl "^$1^"\n"); $1 }
;

variable : /* Term.term */
  | Upper_word { print_string("Variable "^$1^"\n"); Var(($1,!varcount), bt_x) }
;


null : {};
